/*---------------------------------------------------------------------------

	File:
		centiped.cpp
	
	Abstract:
		Implements the "Centipede" game driver functions.

	Revisions:
		18-Feb-97 EAM Created

	Copyright (C) 1997 Edward Massey

	Centipede hardware information (including Pokey emulation by Ron Fries)
	adopted from the excellent Centipede emulator by Ivan Mackintosh.

	TODO:  Use the generic graphics decoding routines.  Decode the sprites
	       with the MY_GFX_DECODE struct similiar to way Galaxian does it.

	TODO:  Add support for Millipede

	TODO:  Color fix (see Ivan's e-mail message)

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "game.h"
#include "game1.h"
#include "../m6502/exec6502.h"

static IExec6502 *_p6502;

extern "C"
{
void TRACE(LPCTSTR pszFormat, ...);
#include "../main/pokey.h"
};

static IMAGE *_pimgChars;
static BYTE _rgbSprites[512][16][16];
static BYTE _rgbSpriteMask[512][16][16];

inline static void DrawSprite(LPBYTE rgbBuffer, LONG lPitch, int x, int y,
	int n, int color)
{
	DrawSprite((LPBYTE)_rgbSprites, (LPBYTE)_rgbSpriteMask,
		n, color, x, y, rgbBuffer, lPitch);
}

#define CODE_START	(0x2000)
#define	HISCORE_BYTES	(0x2F)
#define RAM_HISCORES	(0x2)
#define	ROM_HISCORES	(0x3A69)
#define DIPSW1		(0x800)
#define DIPSW2		(0x801)
#define SPRITE_OFFSET	(0x7C0)
#define SPRITE_XPOS	(0x7D0)
#define SPRITE_YPOS	(0x7E0)
#define SPRITE_COL	(0x7F0)
#define SCREEN_MEM	(0x400)

static const TCHAR _szPath[] = "centi";

RAGE_ROMTABLE_BEGIN(_rgRoms)
	RAGE_ROMTABLE_ENTRY("136001.307", 0x2000, 0x0800)
	RAGE_ROMTABLE_ENTRY("136001.308", 0x2800, 0x0800)
	RAGE_ROMTABLE_ENTRY("136001.309", 0x3000, 0x0800)
	RAGE_ROMTABLE_ENTRY("136001.310", 0x3800, 0x0800)

	RAGE_ROMTABLE_ENTRY("136001.310", 0xF800, 0x0800)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(_rgGfxRoms)
	RAGE_ROMTABLE_ENTRY("136001.211", 0x0000, 0x0800)
	RAGE_ROMTABLE_ENTRY("136001.212", 0x0800, 0x0800)
RAGE_ROMTABLE_END

static const DECODE_GRAPHICS dgsChars =
{
	0x00000000,
	{
		{ 8, 8 },
		256,
		8 * 8,
		2,
		{ 0, 8 * 8 * 256 },
		2,
		{ 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8 },
		{ 7, 6, 5, 4, 3, 2, 1, 0 }
	},

	1,
	
	1, 
	
	{
		0
	},
	0, 
	NULL
};

static BYTE M_RDMEM(WORD A);
static void M_WRMEM(WORD A, BYTE);

static const SIZE _sizBuffer = { 256 + 16, 256 + 16};
static const RECT _rcDisplay = { 0, 0, 256 - 16, 256 };

static const EXEC6502_MEM_DESCRIPTOR rgmd[] = 
{
	{ 0x0000, 0x1FFF, 0, NULL,    NULL }, // direct access before ROMs
	{ 0x0800, 0x0800, 0, M_RDMEM, M_WRMEM },
	{ 0x0801, 0x0801, 0, M_RDMEM, M_WRMEM },
	{ 0x0C00, 0x0C00, 0, M_RDMEM, M_WRMEM },
	{ 0x0C01, 0x0C01, 0, M_RDMEM, M_WRMEM },
	{ 0x0C02, 0x0C02, 0, M_RDMEM, M_WRMEM },
	{ 0x0C03, 0x0C03, 0, M_RDMEM, M_WRMEM },
	{ 0x1000, 0x1008, 0, M_RDMEM, M_WRMEM },
	{ 0x100A, 0x100A, 0, M_RDMEM, M_WRMEM },
	{ 0x100F, 0x100F, 0, M_RDMEM, M_WRMEM },
	{ 0x2000, 0xFFFF, 0, NULL,    M_WRMEM },
	{ 0, 0, 0, (EXEC6502_MEMREAD_PROC)0xFFFFFFFF, 0 },
};

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

enum 
{
    black, red, purple, cyan, yellow, blue, green, white
};

static const int COLOR_CODES = 16; // number of color schemes

static const PALETTEENTRY _rgpe[] =
{
	{ 0x00, 0x00, 0x00 }, // black
	{ 0xFF, 0x00, 0x00 }, // red
	{ 0xB8, 0x02, 0xB8 }, // purple
	{ 0x00, 0xFF, 0xFF }, // cyan
	{ 0xFF, 0xFF, 0x00 }, // yellow
	{ 0x24, 0x24, 0xFF }, // blue
	{ 0x00, 0xFF, 0x00 }, // green
	{ 0xFF, 0xFF, 0xFF }, // white
};

static const BYTE _rgbColorTable[4 * COLOR_CODES] =
{
	black, red,  green, white, // mushroom
	0, 1, 1, 1, 
	0, 2, 2, 2,
	0, 3, 3, 3,
	0, 4, 4, 4,
	0, 5, 5, 5,
	0, 6, 6, 6,
	0, 7, 7, 7,
	0, 1, 1, 1,
	black, red, white, green, // player's ship/spider/centipede head?
	0, 3, 3, 3,
	0, 4, 4, 4,
	0, 5, 5, 5,
	black, red, white, green, // player's ship/spider/centipede head?
	0, 7, 7, 7,
	0, 7, 7, 7,
};

static void GenSprites()
{
	int i, j, k, i2, k2;
	
	for (i = 0; i < 255; i++)
	{
		i2 = i & 0x3E;

		if (i & 1)
			i2 += 128;

		for (j = 0; j < 16; j++)
		{
			for (k = 0; k < 16; k++)
			{
				int nVal;

				if (j < 8)
				{
					if (k < 8)
					{
//						if (i & 0x80)
//							nVal = _rgbChars[i2][j][7 - k];
//						else
							nVal = _pimgChars->rgbData[i2 * 64 + j * 8 + k];
					}
					else
					{
//						if (i & 0x80)
//							nVal = _rgbChars[i2 + 1][j][7 - (k - 8)];
//						else
							nVal = _pimgChars->rgbData[(i2 + 1) * 64 + j * 8 + (k - 8)];
					}
				}
				else
				{
					nVal = 0;
				}
					
				_rgbSprites[i][j][k]    = nVal;
				_rgbSpriteMask[i][j][k] = nVal ? 0x00 : 0xFF;
			}
		}
	}
}

static HRESULT Init(GAME_LOAD_INFO &gli)
{
	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	BuildCompositePalette(_rgpe, DIM(_rgpe), _rgbColorTable,
		COLOR_CODES * 4, gli.ppe, 256, gli.cpe);

	if (readroms(RAM, _rgRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

#ifndef MILLIPEDE
	// hack to be removed eventually
	RAM[0x38a8] = 0xea;
	RAM[0x38a9] = 0xea;
	RAM[0x38ae] = 0xea;
	RAM[0x38af] = 0xea;
#endif

	BYTE tmpstorage[0x1000];

	if (readroms(tmpstorage, _rgGfxRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//
	//
	
	MyGfxDecode3(dgsChars, tmpstorage, _pimgChars);

#if 1
	GenSprites();
#else
	//
	//
	//
	
	static const MY_GFX_DECODE mgdSprites =
	{
		{ 16, 8 },
		128,
		16 * 8,
		2,
		128 * 16 * 8,
		2,
		{ 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8,
				8*8, 9*8, 10*8, 11*8, 12*8, 13*8, 14*8, 15*8 },
		{ 7, 6, 5, 4, 3, 2, 1, 0 },
	};

	static const int rgiMirror[] = {
		0 * mgdSprites.cImage,
		1 * mgdSprites.cImage,
		2 * mgdSprites.cImage,
		3 * mgdSprites.cImage
	};

	static const int cImageInc = 1;

	MyGfxDecode2(tmpstorage, mgdSprites, cImageInc, rgiMirror,
		(LPBYTE)_rgbSprites, (LPBYTE)_rgbSpriteMask);
#endif

	//
	//
	//
	
	Pokey_sound_init(FREQ_17_APPROX, 11025); // 15960);

	//
	//
	//

	EXEC6502_CREATE cs;

	ZERO_INIT(cs);

	cs.rgbMemory = RAM;
	cs.pmd = rgmd;

	HRESULT hr;
	
	hr = EclCreate6502(cs, _p6502);

	if (FAILED(hr))
		return hr;

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

int _cRead = 0, _cWrite = 0, _cReadDirect = 0, _cWriteDirect = 0;

void CentiRead(WORD A)
{
#ifdef _DEBUG

	static int _iTrace = 0;
	static const int _rgcTrace[] = { -1, -1 };

	_iTrace++;

//	if (A < 0x2000 || A >= 0x4000)
	{
		if ((_iTrace >= _rgcTrace[0]) && (_iTrace <= _rgcTrace[1]))
		{
			TRACE("%04X RAM[%04X]\n", _iTrace, A);
		}
	}

	if (_iTrace == 0x01D2)
		TRACE("\n");
#endif
}

void CentiWrite(WORD A, BYTE V)
{
#ifdef _DEBUG
	if (A == 0x00D7 || A == 0x00D8)
		TRACE("write to 0x1800!!\n");

	static int _iTrace = 0;
	static const int _rgcTrace[] = { 0x0600, 0x1000 };

	_iTrace++;
	
	if (A <= 0x100 || A > 0x1FF)
	{
		if ((_iTrace >= _rgcTrace[0]) && (_iTrace <= _rgcTrace[1]))
		{
			TRACE("%04X RAM[%04X]=%02X\n", 0 /*_iTrace*/, A, V);
		}
	}

	if (_iTrace == 0x0609)
		TRACE("\n");
#endif
}

static BYTE M_RDMEM(WORD A)
{
/*
	_cRead++;

	CentiRead(A);
*/

	if (A == 0x0800) // SW1
	{
		BYTE b = 0;

		return b;
	}

	if (A == 0x0801) // SW2
	{
		BYTE b = RAM[A];

		return b;
	}

	if (A == 0x0C00) // trackball
	{
		BYTE b = RAM[A];

		return b;
	}

	if (A == 0x0C01)
	{
		BYTE b = 0xFF;

		if (osd_key_pressed(OSD_KEY_3))
			b ^= 0x40;

		if (osd_key_pressed(OSD_KEY_1))
			b ^= 0x01;

		if (osd_key_pressed(OSD_KEY_CONTROL))
			b ^= 0x04;

		return b;
	}

	if (A == 0x0C02) // trackball
	{
		BYTE b = RAM[A];

		return b;
	}

	if (A == 0x0C03)
	{
		BYTE b = 0xFF;

		if (osd_key_pressed(OSD_KEY_UP))
			b ^= 0x10;
	
		if (osd_key_pressed(OSD_KEY_DOWN))
			b ^= 0x20;
	
		if (osd_key_pressed(OSD_KEY_LEFT))
			b ^= 0x40;
		
		if (osd_key_pressed(OSD_KEY_RIGHT))
			b ^= 0x80;

		return b;
	}

	if (A == 0x100a)
		return rand();

//	_cReadDirect++;

	return(RAM[A]);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void M_WRMEM(WORD A, BYTE V)
{
/*
	_cWrite++;

	CentiWrite(A, V);
*/

	if (A == 0x2000)
		return; // reset watchdog timer

	if (A == 0x2400)
		return; // clear trackball counters

	if (A == 0x1800)
	{
		RAM[A] = V;
		return; // irq ack
	}

	// don't allow code to overwrite itself
	if (A >= CODE_START)
		return;

//	if (_fSoundActive)
	{
		if ((A >= 0x1000) && (A <= 0x1008))
		{
			Update_pokey_sound(0xC200+A, V);
			RAM[A] = V;
			return;
		}
	}

	if (A == 0x100F)
	{
		if (V == 0)
			RAM[0x0C00] = 0x20;
		else
			RAM[0x0C00] = 0x60;

		RAM[A] = V;
		return;
	}

//	_cWriteDirect++;

	RAM[A] = V;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static int Exec()
{
	int cCycles;
	
	cCycles = _p6502->Exec(1000000 / 60); // 1 MHz, 60 fps...  16667 cycles/frame
	cCycles += _p6502->Irq(); 

	return cCycles;
}

static void Term()
{
	_p6502->Release();

	FreeImage(_pimgChars);

//	TRACE("reads:  %i total, %i direct\n", _cRead, _cReadDirect);
//	TRACE("writes: %i total, %i direct\n", _cWrite, _cWriteDirect);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void UpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
	Pokey_process(rgbBuffer, cbBuffer);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void RenderScene(LPBYTE rgbBuffer, LONG lPitch)
{

//	1404 (low nibble) is playfield color RAM
//  140C (low nibble) is sprite color RAM

	int i, j;
	
	const BYTE *pBase = &RAM[0x0400];

	for (j = 0; j < 30; j++)
	{
		for (i = 31; i >= 0; i--)
		{
			_pimgChars->DrawChar(rgbBuffer, lPitch, 8 * j, 8 * i, *pBase + 64, 0);
			pBase++;
		}
	}

	for (i = 0; i < 16; i++)
	{
		int ypos = RAM[SPRITE_YPOS + i];

		if (ypos >= 0xF8)
			continue;

		ypos = 247 - ypos;

		const int xpos = 238 - RAM[SPRITE_XPOS + i];
		const int offset = RAM[SPRITE_OFFSET + i];
		const int color = 0; // RAM[SPRITE_COL + i] & 0x0F;

		DrawSprite(rgbBuffer, lPitch, xpos, ypos, offset, color);
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

#define DECLARE_GAME_DESCRIPTOR(name,title,ctx) \
	static HRESULT Init(GAME_LOAD_INFO &gli); \
	/*const*/ GAME_DESCRIPTOR (name) = { (title), (ctx), Init, Exec, RenderScene, UpdateSound, Term };

DECLARE_GAME_DESCRIPTOR(_gmdCentipede,TEXT("Centipede"),NULL);

