/*---------------------------------------------------------------------------

	File:
		congo.cpp
	
	Abstract:
		Implements the "Congo Bongo" game driver functions.

	Revisions:
		14-Mar-97 EAM Created

	Copyright (C) 1997 Edward Massey

	Hardware information was adopted from Ville Laitinen's Congo Bongo driver
	for MAME, by Nicola Salmoria.

	TODO:  Don't bother drawing chars that are in the first or last two cols.
	TODO:  Don't draw sprites that are too far off to the right.
	TODO:  Move to interleaved sprite data, should make a big difference here
	TODO:  Background image, overlay topmost chars
	
---------------------------------------------------------------------------*/

#include "pch.h"
#include <stdio.h>
#include "misc.h"
#include "z80api.h"
#include "game.h"
#include "game1.h"

static BYTE _rgbChars[256][8][8];
static BYTE _rgbChars2[1024][8][8];
static BYTE _rgbSprites[512][32][32];
static BYTE _rgbSpriteMask[512][32][32];

static BYTE _rgbBackground[0x8000];

#include "blit.h"

/*---------------------------------------------------------------------------

sprite colors 

00 - monkey
04 - monkey (the one you kick off)
0F - land slide
05 - player
19 - gorilla
14 - fire in gorilla's hand
14, 94 - coconuts

1F, 07, 08, 09 - snooring

---------------------------------------------------------------------------*/

static const PALETTEENTRY _rgpe[] =
{
	// 00
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 64,  64,  64  },
	{ 0,   128, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 01 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 02 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 03 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 04 - monkey
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 64,  64,  64   },
	{ 0,   128, 0   },
	{ 0xDB, 0x92, 0x49 }, // body
	{ 0,   0,   255 },
	{ 0xDB, 0x92, 0x49 }, // body (dark)
	{ 255, 255, 255 },

	// 05 
	{ 0,   0,   0   },
	{ 255, 255, 255 }, // suit, hat
	{ 0,   0,   0   }, // belt, shadow, eyes
	{ 252, 184, 184 }, // skin
	{ 252, 184, 184 }, // skin (highlight)
	{ 0,   0,   0   }, // hair
	{ 255, 0,   0   }, // nose
	{ 192, 192, 192 }, // hat, suit

	// 06 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 07 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 08 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 09 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 0A 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 0B 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 0C 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 0D 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 0E 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 0F 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 10 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 11 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 12 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 13 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 14 - coconut
	{ 0,   0,   0   },
	{ 255, 0,   0   }, // 3
	{ 64,  64,  64   },// hitting ground (furthest out)
	{ 0,   128, 0   }, // 2
	{ 0xDB, 0x92, 0x49 }, // main
	{ 0,   0,   0 }, // outline
	{ 255, 0,   255 }, // shadow
	{ 255, 255, 255 }, // outline (when hitting ground)

	// 15 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 16 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 17 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 18 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 19 - gorilla
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   128, 0   }, // outline
	{ 0xDB, 0x92, 0x49 }, // body
	{ 0xDB, 0x92, 0x49 }, // body (shadowed)
	{ 255, 0,   255 }, // lips, outline of eyes
	{ 255, 255, 255 }, // eyes, teeth

	// 1A 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 1B 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 1C 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 1D 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 1E 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 1F 
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },
};

static HRESULT Init(GAME_LOAD_INFO &gli)
{
	static const TCHAR szPath[] = "congo";

	static const SIZE _sizBuffer = { 256 + 32, 256 + 32};
	static const RECT _rcDisplay = { 16, 0, 256 - 16, 256};
//	static const RECT _rcDisplay = { 0, 0, 256 + 32, 256 + 32};

	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	for (int i = 0; i < DIM(_rgpe); i++)
		gli.ppe[i] = _rgpe[i];

	gli.cpe = DIM(_rgpe);

	//
	//
	//
	
	RAGE_ROMTABLE_BEGIN(ROMs)
		RAGE_ROMTABLE_ENTRY("congo1.bin",  0x0000, 0x2000)
		RAGE_ROMTABLE_ENTRY("congo2.bin",  0x2000, 0x2000)
		RAGE_ROMTABLE_ENTRY("congo3.bin",  0x4000, 0x2000)
		RAGE_ROMTABLE_ENTRY("congo4.bin",  0x6000, 0x2000)
	RAGE_ROMTABLE_END

	if (readroms(RAM, ROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//
	//
	
	BYTE tmpstorage[0x13000];

	RAGE_ROMTABLE_BEGIN(GFXROMs)
		RAGE_ROMTABLE_ENTRY("congo5.bin",  0x0000, 0x1000) // charset 1
		RAGE_ROMTABLE_ENTRY("congo16.bin", 0x1000, 0x2000) // sprites
		RAGE_ROMTABLE_ENTRY("congo15.bin", 0x3000, 0x2000)
		RAGE_ROMTABLE_ENTRY("congo12.bin", 0x5000, 0x2000)
		RAGE_ROMTABLE_ENTRY("congo14.bin", 0x7000, 0x2000)
		RAGE_ROMTABLE_ENTRY("congo11.bin", 0x9000, 0x2000)
		RAGE_ROMTABLE_ENTRY("congo13.bin", 0xb000, 0x2000)
		RAGE_ROMTABLE_ENTRY("congo10.bin", 0xd000, 0x2000) // charset 2
		RAGE_ROMTABLE_ENTRY("congo9.bin",  0xf000, 0x2000)
		RAGE_ROMTABLE_ENTRY("congo8.bin",  0x11000, 0x2000)
	RAGE_ROMTABLE_END

	if (readroms(tmpstorage, GFXROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	static const MY_GFX_DECODE mgdChars1 =
	{
		{ 8, 8 },
		256,
		8 * 8,
		2,
		8 * 8 * 256,
		2,
		{ 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
		{ 0, 1, 2, 3, 4, 5, 6, 7 },
	};

	MyGfxDecode(tmpstorage, mgdChars1, (LPBYTE)_rgbChars);

	static const MY_GFX_DECODE mgdChars2 =
	{
		{ 8, 8 },
		1024,
		8 * 8,
		3,
		8 * 8 * 1024,
		3,
		{ 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8 },
		{ 0, 1, 2, 3, 4, 5, 6, 7 },
	};

	MyGfxDecode(tmpstorage + 0xD000, mgdChars2, (LPBYTE)_rgbChars2);

	static const MY_GFX_DECODE mgdSprites =
	{
		{ 32, 32 },
		128,
		128 * 8,
		3,
		2 * 1024 * 8 * 8,
		3,
		{ 103*8, 102*8, 101*8, 100*8, 99*8, 98*8, 97*8, 96*8,
			71*8, 70*8, 69*8, 68*8, 67*8, 66*8, 65*8, 64*8,
			39*8, 38*8, 37*8, 36*8, 35*8, 34*8, 33*8, 32*8,
			7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
		{ 0, 1, 2, 3, 4, 5, 6, 7,
			8*8+0, 8*8+1, 8*8+2, 8*8+3, 8*8+4, 8*8+5, 8*8+6, 8*8+7,
			16*8+0, 16*8+1, 16*8+2, 16*8+3, 16*8+4, 16*8+5, 16*8+6, 16*8+7,
			24*8+0, 24*8+1, 24*8+2, 24*8+3, 24*8+4, 24*8+5, 24*8+6, 24*8+7 },
	};

	static const int rgiMirror[] = {
		0 * mgdSprites.cImage,
		2 * mgdSprites.cImage,
		1 * mgdSprites.cImage,
		3 * mgdSprites.cImage
	};

	static const int cImageInc = 1;

	MyGfxDecode2(tmpstorage + 0x1000, mgdSprites, cImageInc, rgiMirror,
		(LPBYTE)_rgbSprites, (LPBYTE)_rgbSpriteMask);

	RAGE_ROMTABLE_BEGIN(BKGROMs)
		RAGE_ROMTABLE_ENTRY("congo6.bin", 0x0000, 0x2000)
		RAGE_ROMTABLE_ENTRY("congo6.bin", 0x2000, 0x2000)
		RAGE_ROMTABLE_ENTRY("congo7.bin", 0x4000, 0x2000)
		RAGE_ROMTABLE_ENTRY("congo7.bin", 0x6000, 0x2000)
	RAGE_ROMTABLE_END

	if (readroms(_rgbBackground, BKGROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE M_RDMEM(WORD A)
{
	switch (A)
	{
		case 0xC000:
		{
			BYTE b = 0x00;

			if (osd_key_pressed(OSD_KEY_RIGHT))
				b ^= 0x01;

			if (osd_key_pressed(OSD_KEY_LEFT))
				b ^= 0x02;

			if (osd_key_pressed(OSD_KEY_UP))
				b ^= 0x04;

			if (osd_key_pressed(OSD_KEY_DOWN))
				b ^= 0x08;

			if (osd_key_pressed(OSD_KEY_CONTROL))
				b ^= 0x10;

			return b;
		}

		case 0xC001:
		{
			BYTE b = 0x00;

			return b;
		}

		case 0xC002:
		{
			BYTE b = 0x7F;

			return b;
		}

		case 0xC003:
		{
			BYTE b = 0x3F;

			return b;
		}

		case 0xC008:
		{
			BYTE b = 0x00;

			if (osd_key_pressed(OSD_KEY_1))
				b ^= 0x04;

			if (osd_key_pressed(OSD_KEY_2))
				b ^= 0x08;

			if (osd_key_pressed(OSD_KEY_3))
				b ^= 0x80;

			return b;
		}

		default:
			return RAM[A];
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void M_WRMEM(WORD A, BYTE V)
{
	if (A < 0x8000)
		return;

	RAM[A] = V;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static int interrupt_vector = 0xff;

static int Interrupt(void)
{
	if (!CheckMessages())
		CPURunning = 0;

	Throttle(60); // FRAMES_PER_SECOND);

	if (RAM[0xc01f] == 0)
		return IGNORE_INT;
	
	return interrupt_vector;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

// TODO:  call higher-level Exec function that handles the common parts
//        of Interrupt, pass it our CPU speed and frame rate

static HRESULT Run(LPVOID pvContext)
{
	EXEC_Z80_STRUCT es;
	
	ZERO_INIT(es);
	
	es.nCallbackFreq = 50000; // CALLBACK_FREQ;
	es.pfnInterrupt = Interrupt;
	es.pfnReadMemory = M_RDMEM;
	es.pfnWriteMemory = M_WRMEM;
	es.rgbMemory = RAM;
	
	return ExecZ80(es);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

inline static void DrawChar2(LPBYTE rgbBuffer, LONG lPitch, int X, int Y,
	int offset_value, int color)
{
	LPDWORD pDest = (LPDWORD)&rgbBuffer[lPitch * Y + X];
	LPCDWORD pSrc = (LPCDWORD)&_rgbChars2[offset_value][0][0];
	color |= (color << 8) | (color << 16) | (color << 24);
	lPitch = (lPitch / 4) - 1;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
}

inline static void DrawSprite2(LPBYTE rgbBuffer, LONG lPitch, int x, int y,
	int n, int color)
{
	//
	//
	//
	
	color |= (color << 8) | (color << 16) | (color << 24);

	//
	//
	//
	
	LPDWORD pDest = (LPDWORD)&rgbBuffer[lPitch * y + x];
	LPCDWORD pSrc = (LPCDWORD)&_rgbSprites[n][0][0];
	LPCDWORD pMask = (LPCDWORD)&_rgbSpriteMask[n][0][0];
	
	//
	//  Draw sprite, with transparency, and color scheme adjusted
	//  to offscreen buffer, 32-bits at a time
	//
	//  CONSIDER:  unroll the loop
	//
	
	lPitch = (lPitch / 4) - 8;
	
	for (int i = 0; i < 8; i++)
	{
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		pDest += lPitch;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		pDest += lPitch;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		pDest += lPitch;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		pDest += lPitch;

	}
}

static const int _rgiHex[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 17, 18, 19, 20, 21, 22 };

static int base = 0;
static int keystate = 0;

static HRESULT RenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	int i, j;

#if 0
	const BOOL fShowI = GetAsyncKeyState('I') < 0;
	const BOOL fShow0 = GetAsyncKeyState('0') < 0;

	if (GetAsyncKeyState('C') < 0)
	{
		if (GetAsyncKeyState('1') < 0)
		{
			for (i = 0; i < 256; i++)
			{
				DrawChar(rgbBuffer, lPitch, 8 * (i % 32), 8 * (i / 32), i, 0);
			}
		}

		if (GetAsyncKeyState('2') < 0)
		{
			for (i = 0; i < 1024; i++)
			{
				DrawChar2(rgbBuffer, lPitch, 8 * (i % 32), 8 * (i / 32), i, 0);
			}
		}

		return S_OK;
	}

	base = RAM[0xc028] + 256 * RAM[0xc029];

	base = base / 8;
	base *= 32;

	for (i = 0; i < 32; i++)
	{
		for (j = 0; j < 32; j++)
		{
			int k = i * 32 + j + base;
			
			DrawChar2(rgbBuffer, lPitch, 8 * i, 8 * j,
				_rgbBackground[k] + 256 * ((_rgbBackground[k + 0x4000]) & 3), 0);
		}
	}

	DrawChar(rgbBuffer, lPitch, 16, 0, _rgiHex[(base & 0xf000) >> 12], 0);
	DrawChar(rgbBuffer, lPitch, 24, 0, _rgiHex[(base & 0x0f00) >> 8], 0);
	DrawChar(rgbBuffer, lPitch, 32, 0, _rgiHex[(base & 0x00f0) >> 4], 0);
	DrawChar(rgbBuffer, lPitch, 40, 0, _rgiHex[(base & 0x000f) >> 0], 0);
#endif

	for (i = 0; i < 32; i++)
	{
		for (j = 2; j < 30; j++)
		{
			const int c = RAM[0xA000 + i + j * 32];
			const int clr = RAM[0xA400 + i + j * 32];

			// skip blanks in playfield
//			if (c != 0x60)
				DrawChar((LPBYTE)_rgbChars, c, 0, 8 * (31 - j), 8 * i, rgbBuffer, lPitch);
		}
	}

	// note: assumes byte packing...

	struct SPRITE
	{
		BYTE d1, d2, x, i, c, y;
		BYTE d3[0x20 - 6];
	};

	const SPRITE *rgs = (SPRITE*)&RAM[0x8400];

	for (i = 0x1E; i >= 0; i--)
	{
		if (rgs[i].x != 0xFF)
		{
#if 0
			if (fShowI)
			{
				DrawChar(rgbBuffer, lPitch, rgs[i].x, rgs[i].y,
					_rgiHex[((rgs[i].c) & 0xF0) >> 4], 0);

				DrawChar(rgbBuffer, lPitch, rgs[i].x + 8, rgs[i].y,
					_rgiHex[((rgs[i].c) & 0x0F)], 0);
			}
			else
			{
				if (fShow0)
				{
					DrawSprite2(rgbBuffer, lPitch, rgs[i].x - 15,
						((rgs[i].y + 16) & 0xFF) - 32, rgs[i].i & 0x7F, 0);
				}
				else
				{
//					DrawSprite2(rgbBuffer, lPitch, rgs[i].x - 15,
//						((rgs[i].y + 16) & 0xFF) - 32, rgs[i].i & 0x7F, rgs[i].c & 0x1F);

					DrawSprite2(rgbBuffer, lPitch, rgs[i].x,
						rgs[i].y, rgs[i].i + 2 * (rgs[i].c & 0x80), (rgs[i].c & 0x1F) << 3);
				}
			}
#else
			DrawSprite2(rgbBuffer, lPitch, rgs[i].x, rgs[i].y,
				rgs[i].i & 0x7F, 0);
#endif
		}
	}

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdCongoBongo = 
{
	TEXT("Congo Bongo"),
	Init,
	Run,
	RenderScene,
	NULL,
	NULL,
	NULL,
	NULL
};

