/*---------------------------------------------------------------------------

	File:
		dkong.cpp
	
	Abstract:
		Implements the "Donkey Kong" game driver functions.  Handles all the
		game's hardware emulation except CPU, loading ROMs.

	Revisions:
		10-Feb-97 EAM Created

	Copyright (C) 1997 Edward Massey

	Acknowledgements:
		Thanks to Gary Shepherdson for the Donkey Kong technical info.
		Donkey Kong and Donkey Kong Jr. color tables, Donkey Kong Jr. ROM
			load addresses adopted from MAME, by Nicola Salmoria.
		Sound samples and hardware info from Ron Fries

	TODO:  no need to draw the first and last two columns of characters!!!
	TODO:  sound for jr.
	TODO:  move static info (ROMs, decode info) to top of file
	TODO:  scope symbols in DKONG_GAME
	TODO:  dynamically allocate sample buffers
	TODO:  get static info from script
	TODO:  left shift graphics 2 bits during decode
	TODO:  optimize Z80 performance
	TODO:  interleave sprite data/mask
	TODO:  add color decoding capability to graphics module
	TODO:  make separate ROM tables for character and sprite ROMs
	TODO:  make jr's ROM loading table entries into macros

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "game.h"
#include "game1.h"
#include "../z80/execz80.h"

#define base 0x7400

static IMAGE *_pimgSprites;
static IMAGE *_pimgChars;

/*---------------------------------------------------------------------------

color palette layout...

16x4 colors for sprites
16 colors for text

---------------------------------------------------------------------------*/

static const PALETTEENTRY _rgpe[] = 
{
	{ 0x00, 0x00, 0x00 }, /* BLACK */
	{ 0xff, 0xff, 0xff }, /* WHITE */
	{ 0xff, 0x00, 0x00 }, /* RED */
	{ 0xff, 0x00, 0xff }, /* PURPLE */
	{ 0x00, 0xff, 0xff }, /* CYAN */
	{ 0xff, 0xff, 0x80 }, /* LTORANGE */
	{ 0xdb, 0x00, 0x00 }, /* DKRED */
	{ 0x00, 0x00, 0xff }, /* BLUE */
	{ 0xff, 0xff, 0x00 }, /* YELLOW */
	{ 239,  0x03, 239  }, /* PINK */
	{ 0x03, 180,  239  }, /* LTBLUE */
	{ 0xff, 131,  0x03 }, /* ORANGE */
	{ 0x00, 0xff, 0x00 }, /* GREEN */
	{ 167,  0x03, 0x03 }, /* DKBROWN */
	{ 0xff, 183,  115  }, /* LTBROWN */
	{ 0x00, 0x46, 0x00 }, /* DKGREEN */
};

enum { BLACK,WHITE,RED,PURPLE,CYAN,LTORANGE,DKRED,BLUE,YELLOW,PINK,
		LTBLUE,ORANGE,GREEN,DKBROWN,LTBROWN,DKGREEN };

static const int _rgiDkongCharColors[][4] =
{
	BLACK,0,0,WHITE,        /* 0-9 */
	BLACK,0,0,WHITE,        /* 0-9 */
	BLACK,0,0,WHITE,        /* 0-9 */
	BLACK,0,0,WHITE,        /* 0-9 */
	BLACK,0,0,RED,  /* A-Z */
	BLACK,0,0,RED,  /* A-Z */
	BLACK,0,0,RED,  /* A-Z */
	BLACK,0,0,RED,  /* A-Z */
	BLACK,0,0,RED,  /* A-Z */
	BLACK,0,0,RED,  /* A-Z */
	BLACK,0,0,RED,  /* A-Z */
	BLACK,0,0,RED,  /* A-Z */
	BLACK,0,0,RED,  /* A-Z */
	BLACK,0,0,RED,  /* A-Z */
	BLACK,0,0,RED,  /* A-Z */
	BLACK,0,0,RED,  /* A-Z */
	0,0,0,RED,      /* RUB END */
	0,0,0,RED,      /* RUB END */
	0,5,6,RED,      /* (C), ITC */
	0,5,6,RED,      /* (C), ITC */
	0,LTBROWN,DKRED,WHITE, /* Kong in intermission */
	0,LTBROWN,DKRED,WHITE, /* Kong in intermission */
	0,LTBROWN,DKRED,WHITE, /* Kong in intermission */
	0,LTBROWN,DKRED,WHITE, /* Kong in intermission */
	0,LTBROWN,DKRED,WHITE, /* Kong in intermission */
	0,LTBROWN,DKRED,WHITE, /* Kong in intermission */
	BLACK,RED,PURPLE,CYAN,  /* BONUS */
	BLACK,RED,PURPLE,CYAN,  /* BONUS */
	BLACK,RED,PURPLE,CYAN,  /* BONUS */
	BLACK,RED,PURPLE,CYAN,  /* BONUS */
	BLACK,RED,PURPLE,CYAN,  /* BONUS */
	BLACK,RED,PURPLE,CYAN,  /* BONUS */
	BLACK,0,0,RED,          /* 0-9 in intermission */
	BLACK,0,0,RED,          /* 0-9 in intermission */
	BLACK,0,0,RED,          /* 0-9 in intermission */
	BLACK,RED,PURPLE,CYAN,  /* 0-9 in intermission, BONUS */
	0,0,0,0,        /* unused */
	0,0,0,0,        /* unused */
	0,0,0,0,        /* unused */
	0,0,0,RED,      /* (TM) */
	0,0,0,0,        /* unused */
	0,0,0,0,        /* unused */
	0,0,0,0,        /* unused */
	0,0,0,0,        /* unused */
	0,CYAN,BLUE,YELLOW, /* ziqqurat level */
	0,CYAN,BLUE,YELLOW, /* ziqqurat level */
	0,LTBLUE,0,0,   /* ziqqurat level */
	0,0,0,0,        /* unused */
	BLACK,RED,PURPLE,CYAN,  /* barrel level girder, ladder */
	BLACK,RED,PURPLE,CYAN,  /* barrel level girder*/
	0,0,0,0,        /* unused */
	0,0,0,0,        /* unused */
	BLACK,RED,PURPLE,CYAN,  /* barrel level girder */
	BLACK,RED,PURPLE,CYAN,  /* barrel level girder */
	0,0,0,0,        /* unused */
	0,0,0,CYAN,     /* HELP! */
	BLACK,RED,PURPLE,CYAN,  /* barrel level girder */
	BLACK,RED,PURPLE,CYAN,  /* barrel level girder */
	0,0,0,0,        /* unused */
	0,0,0,CYAN,     /* HELP! */
	BLACK,RED,PURPLE,CYAN,  /* barrel level girder */
	BLACK,RED,PURPLE,CYAN,  /* barrel level girder */
	0,0,0,0,        /* unused RED */
	0,RED,LTORANGE,BLUE,    /* lives */
};

static const int _rgiDkongSpriteColors[] =
{
	BLACK,LTBLUE,BLUE,ORANGE,       /* Fireball (When Mario has hammer) */
							/* Rotating ends on conveyors */
							/* Springy things (lift screen) */
	BLACK,ORANGE,YELLOW,WHITE,      /* Fireball (normal) */
							/* Flames (on top of oil tank) */
	BLACK,RED,LTORANGE,BLUE,        /* Mario */
	BLACK,RED,PURPLE,WHITE,         /* -Moving Ladder (conveyor screen) */
							/* Moving Lift */
	BLACK,0,0,0,                    /* unused */
	BLACK,0,0,0,                    /* unused */
	BLACK,0,0,0,                    /* unused */
	BLACK,LTBROWN,DKBROWN,WHITE,    /* Kong (Head), Hammer, Scores (100,200,500,800 etc) */
	BLACK,LTBROWN,DKBROWN,ORANGE,   /* Kong (body) */
	BLACK,ORANGE,WHITE,PINK,        /* girl (Head), Heart (when screen completed) */
	BLACK,WHITE,BLUE,PINK,          /* Girl (lower half), Umbrella, Purse, hat */
	BLACK,ORANGE,BLUE,LTORANGE,     /* Rolling Barrel (type 1), Standing Barrel (near Kong) */
	BLACK,WHITE,CYAN,BLUE,          /* Oil tank, Rolling Barrel (type 2), Explosion (barrel hit withhammer) */
	BLACK,0,0,0,                    /* unused */
	BLACK,LTBROWN,YELLOW,BLUE,      /* -Pies (Conveyor screen) */
	BLACK,YELLOW,ORANGE,BLACK  /* -Thing at top/bottom of lifts, Clipping sprite (all black) */
};

static const int _rgiDkongjrCharColors[][4] =
{
	/* chars (first bank) */
	BLACK,BLACK,BLACK,WHITE,       /* 0-9 */
	BLACK,BLACK,BLACK,WHITE,       /* 0-9 */
	BLACK,BLACK,BLACK,WHITE,       /* 0-9 */
	BLACK,BLACK,BLACK,WHITE,       /* 0-9 */
	BLACK,BLACK,BLACK,RED,         /* A-Z */
	BLACK,BLACK,BLACK,RED,         /* A-Z */
	BLACK,BLACK,BLACK,RED,         /* A-Z */
	BLACK,BLACK,BLACK,RED,         /* A-Z */

	BLACK,BLACK,BLACK,RED,         /* A-Z */
	BLACK,BLACK,BLACK,RED,         /* A-Z */
	BLACK,BLACK,BLACK,RED,         /* A-Z */
	BLACK,BLACK,BLACK,RED,         /* A-Z */
	BLACK,BLACK,BLACK,RED,         /* A-Z */
	BLACK,BLACK,BLACK,RED,         /* A-Z */
	BLACK,BLACK,BLACK,RED,         /* A-Z */
	BLACK,BLACK,BLACK,RED,         /* A-Z */

	BLACK,BLACK,BLACK,WHITE,       /* RUB END */
	BLACK,BLACK,BLACK,WHITE,       /* RUB END */
	BLACK,BLACK,BLACK,WHITE,       /* RUB END */
	BLACK,BLACK,BLACK,WHITE,       /* RUB END */
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,

	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	BLACK,BLACK,RED,BLUE,          /* Bonus */
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	BLACK,BLACK,RED,BLUE,          /* Bonus Box */

	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	BLACK,BLACK,RED,BLUE,          /* Bonus Box */
	BLACK,BLACK,RED,BLUE,          /* Bonus Box */
	BLACK,BLACK,RED,BLUE,          /* Bonus Box */
	0,0,0,0,
	0,0,0,0,

	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	BLACK,BLACK,ORANGE,BLACK,      /* Screen1: Islands Bottom */
	BLACK,LTBLUE,BLUE,RED,         /* Screen4: Locks */
	BLACK,LTBLUE,BLUE,WHITE,       /* Screen1: Water */
	BLACK,BLACK,BLUE,BLACK,        /* Screen2: Rope Cage-Lock */
	BLACK,BLACK,BLUE,BLACK,        /* Screen2: Rope Cage-Lock */
	BLACK,GREEN,RED,LTBROWN,       /* Screen1: Vines */
	BLACK,GREEN,BLUE,LTBLUE,       /* Rope & Chains */
	BLACK,CYAN,BLUE,LTBLUE,        /* Screen2: Wall */
	BLACK,LTBLUE,BLUE,BLUE,        /* Screen4: Mario's Platform */
	BLACK,LTBLUE,BLUE,BLUE,        /* Screen4: Mario's Platform */
	BLACK,LTBLUE,BLUE,BLUE,        /* Screen4: Mario's Platform */
	0,0,0,0,                       /* Unused */
	0,0,0,0,                       /* Unused */
	BLACK,GREEN,DKGREEN,WHITE,     /* Screen1: Islands Top */
	BLACK,LTBLUE,BLUE,WHITE,       /* Screen1&2: Top Floor Bottom Part */
	BLACK,LTBLUE,BLUE,BLUE,        /* Screen4: Bottom Platform */
	BLACK,LTBLUE,BLUE,BLUE,        /* Screen4: Bottom Platform */
	BLACK,LTBLUE,ORANGE,YELLOW,    /* Screen1,2,4: Platform */
	BLACK,LTBLUE,BLUE,BLUE,        /* Screen4: Bottom Platform */
	BLACK,LTBLUE,BLUE,BLUE,        /* Screen4: Bottom Platform */
	BLACK,DKGREEN,DKBROWN,LTBROWN, /* Lives */

	/* chars (second bank) */
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	BLACK,LTORANGE,RED,BLACK,      /* Logo: Donkey Kong */
	BLACK,LTORANGE,RED,BLACK,      /* Logo: Donkey Kong */
	BLACK,LTORANGE,RED,BLACK,      /* Logo: Donkey Kong */
	BLACK,LTORANGE,RED,BLACK,      /* Logo: Donkey Kong */
	BLACK,LTORANGE,RED,BLACK,      /* Logo: Donkey Kong */
	BLACK,LTORANGE,RED,BLACK,      /* Logo: Donkey Kong */
	BLACK,LTORANGE,RED,BLACK,      /* Logo: Donkey Kong */
	BLACK,LTORANGE,RED,BLACK,      /* Logo: Donkey Kong */
	BLACK,LTORANGE,RED,BLACK,      /* Logo: Donkey Kong */
	BLACK,LTORANGE,RED,BLACK,      /* Logo: Donkey Kong */
	BLACK,LTORANGE,RED,BLACK,      /* Logo: Donkey Kong */
	BLACK,LTORANGE,RED,BLACK,      /* Logo: Donkey Kong */
	BLACK,LTORANGE,RED,BLACK,      /* Logo: Donkey Kong */
	BLACK,LTORANGE,RED,BLUE,       /* Logo: Junior */
	BLACK,LTORANGE,RED,BLUE,       /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	BLACK,BLACK,RED,BLUE,          /* Logo: Junior */
	0,0,0,0,
	BLACK,GREEN,BLACK,BLACK,       /* Logo: (c) 198 */
	BLACK,GREEN,BLACK,BLACK,       /*       2 Nin   */
	BLACK,GREEN,BLACK,BLACK,       /*       tendo   */
	BLACK,GREEN,BLACK,BLACK,       /*       of Am   */
	BLACK,GREEN,BLACK,BLACK,       /*       erica   */
	BLACK,GREEN,BLACK,BLACK,       /*       Inc.    */
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
};

static const int _rgiDkongjrSpriteColors[] =
{
	BLACK,LTORANGE,BLUE,RED,       /* Mario */
	BLACK,BLACK,BLACK,WHITE,       /* Bonus Score */
	BLACK,GREEN,ORANGE,WHITE,      /* Screen2: Moving Platform */
	BLACK,1,2,3,                   /* ?? */
	BLACK,YELLOW,BLUE,WHITE,       /* Cursor */
	BLACK,4,5,6,                   /* ?? */
	BLACK,7,8,9,                   /* ?? */
	BLACK,WHITE,DKBROWN,LTBROWN,   /* Kong Jr */
	BLACK,BLUE,BLUE,LTBLUE,        /* Cage */
	BLACK,ORANGE,DKBROWN,LTBROWN,  /* Kong */
	BLACK,1,7,9,                   /* ?? */
	BLACK,GREEN,ORANGE,YELLOW,     /* Banana & Pear */
	BLACK,GREEN,RED,WHITE,         /* Apple */
	BLACK,CYAN,BLUE,WHITE,         /* Blue Creature */
	BLACK,YELLOW,RED,WHITE,        /* Red Creature */
	BLACK,YELLOW,PURPLE,WHITE      /* Key */
};

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

//
//
//

static const TCHAR _szPath[] = "dkong";

RAGE_ROMTABLE_BEGIN(ROMs)
	RAGE_ROMTABLE_ENTRY("dk.5e", 0x0000, 0x1000)
	RAGE_ROMTABLE_ENTRY("dk.5c", 0x1000, 0x1000)
	RAGE_ROMTABLE_ENTRY("dk.5b", 0x2000, 0x1000)
	RAGE_ROMTABLE_ENTRY("dk.5a", 0x3000, 0x1000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(GFXROMs)
	RAGE_ROMTABLE_ENTRY("dk.3n", 0x0000, 0x0800)
	RAGE_ROMTABLE_ENTRY("dk.3p", 0x0800, 0x0800)
	
	RAGE_ROMTABLE_ENTRY("dk.7c", 0x1000, 0x0800)
	RAGE_ROMTABLE_ENTRY("dk.7d", 0x1800, 0x0800)
	RAGE_ROMTABLE_ENTRY("dk.7e", 0x2000, 0x0800)
	RAGE_ROMTABLE_ENTRY("dk.7f", 0x2800, 0x0800)
RAGE_ROMTABLE_END

static const TCHAR _szJrPath[] = "dkongjr";

static const RomModule2 _rgDKJROMS[] = {
	{ "dkj.5b", 0x0000, 0x1000, 0x0000 },
	{ "dkj.5b", 0x1000, 0x1000, 0x3000 },
	{ "dkj.5c", 0x0000, 0x0800, 0x2000 },
	{ "dkj.5c", 0x0800, 0x0800, 0x4800 },
	{ "dkj.5c", 0x1000, 0x0800, 0x1000 },
	{ "dkj.5c", 0x1800, 0x0800, 0x5800 },
	{ "dkj.5e", 0x0000, 0x0800, 0x4000 },
	{ "dkj.5e", 0x0800, 0x0800, 0x2800 },
	{ "dkj.5e", 0x1000, 0x0800, 0x5000 },
	{ "dkj.5e", 0x1800, 0x0800, 0x1800 },
	{ NULL }
};

RAGE_ROMTABLE_BEGIN(JrGFXROMs)
	RAGE_ROMTABLE_ENTRY("dkj.3n", 0x0000, 0x1000)
	RAGE_ROMTABLE_ENTRY("dkj.3p", 0x1000, 0x1000)
	
	RAGE_ROMTABLE_ENTRY("dkj.7c", 0x2000, 0x0800)
	RAGE_ROMTABLE_ENTRY("dkj.7d", 0x2800, 0x0800)
	RAGE_ROMTABLE_ENTRY("dkj.7e", 0x3000, 0x0800)
	RAGE_ROMTABLE_ENTRY("dkj.7f", 0x3800, 0x0800)
RAGE_ROMTABLE_END

#define UPDATES_PER_SECOND 60
#define emulation_rate 11025
#define buffer_len 69000

struct SAMPLE_INFO
{
	LPCTSTR pszFilename;
	BOOL fLoop;
	int cbSize;
	char *rgbData;
};

static char _rgbSamples[25][buffer_len];

extern void SoundSetEnvelope(int iChnl, const char *pData, int cSamples, BOOL fLoop);

static SAMPLE_INFO _rgsi[] = 
{
	{ "effect00.sam", FALSE, 0, _rgbSamples[0]  }, // walk
	{ "effect01.sam", FALSE, 0, _rgbSamples[1]  }, // jump
	{ "effect02.sam", FALSE, 0, _rgbSamples[2]  }, // boom
	{ "effect03.sam", FALSE, 0, _rgbSamples[3]  }, // coin
	{ "effect04.sam", FALSE, 0, _rgbSamples[4]  }, // drop
	{ "effect05.sam", FALSE, 0, _rgbSamples[5]  }, // prize
	{ NULL,           FALSE, 0, _rgbSamples[6]  },
	{ NULL,           FALSE, 0, _rgbSamples[7]  },

	{ NULL,           FALSE, 0, _rgbSamples[8]  },
	{ "tune01.sam",   FALSE, 0, _rgbSamples[9]  }, // intro
	{ "tune02.sam",   FALSE, 0, _rgbSamples[10] }, // how high
	{ "tune03.sam",   FALSE, 0, _rgbSamples[11] }, // time
	{ "tune04.sam",   FALSE, 0, _rgbSamples[12] }, // hamer
	{ "tune05.sam",   FALSE, 0, _rgbSamples[13] }, // rivet2a
	{ "tune06.sam",   FALSE, 0, _rgbSamples[14] }, // hamhit
	{ "tune07.sam",   FALSE, 0, _rgbSamples[15] }, // lv1end
	{ "tune08.sam",   FALSE, 0, _rgbSamples[16] }, // back1
	{ "tune09.sam",   FALSE, 0, _rgbSamples[17] },
	{ NULL,           FALSE, 0, _rgbSamples[18] }, // back3
	{ "tune0b.sam",   FALSE, 0, _rgbSamples[19] }, // back2
	{ "tune0c.sam",   FALSE, 0, _rgbSamples[20] }, // rivet1a
	{ "tune0d.sam",   FALSE, 0, _rgbSamples[21] },
	{ "tune0e.sam",   FALSE, 0, _rgbSamples[22] }, // rivet1
	{ "tune0f.sam",   FALSE, 0, _rgbSamples[23] }, // gorilla

	{ "interupt.sam", FALSE, 0, _rgbSamples[24] }, // death
};

int dkong_sh_init(LPCTSTR gamename = "dkong")
{
	int i, j;

	for (i = 0; i < DIM(_rgsi); i++)
	{
		if (_rgsi[i].pszFilename == NULL)
			continue;

		TStr<_MAX_PATH> strFile("%s/%s", gamename, _rgsi[i].pszFilename);

		FILE *pf = fopen(strFile, "rb");

		if (pf != NULL)
		{
			_rgsi[i].cbSize = fread(_rgsi[i].rgbData, 1, buffer_len, pf);

			fclose(pf);

/*
			for (j = 0; j < _rgsi[i].cbSize; j++)
				_rgsi[i].rgbData[j] += (char)0x00; // 0x80;
*/
		}
	}

	return 0;
}

static IExecZ80 *_pZ80;

static BYTE M_RDMEM(WORD A);
static void M_WRMEM(WORD A, BYTE);

static HRESULT Init(GAME_LOAD_INFO &gli)
{
	dkong_sh_init();

	int i, j, k;

	static const SIZE _sizBuffer = { 256 + 16, 256 + 16};
	static const RECT _rcDisplay = { 16, 0, 256 - 16, 256 };

	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	//
	//
	//
	
	for (i = 0; i < 16; i++)
		gli.ppe[i] = _rgpe[i];

	for (i = 0; i < 4 * 16; i++)
		gli.ppe[16 + i] = _rgpe[_rgiDkongSpriteColors[i]];

	gli.cpe = 16 + 16 * 4;

	//
	//
	//

	BYTE tmpstorage[0x3000];

	if (readroms(RAM, ROMs, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	if (readroms(tmpstorage, GFXROMs, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//
	//

	static const DECODE_GRAPHICS dgsChars =
	{
		0x00000000,
		{
			{ 8, 8 },
			256,
			8 * 8,
			2,
			{ 0, 8 * 8 * 256 },
			2,
			{ 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
			{ 0, 1, 2, 3, 4, 5, 6, 7 },
		},

		1,
		
		1, 
		
		{
			0
		},
		0, 
		NULL
	};

	MyGfxDecode3(dgsChars, tmpstorage, _pimgChars);

	for (i = 0; i < dgsChars.mgd.cImage; i++)
	{
		for (j = 0; j < dgsChars.mgd.sizImage.cy; j++)
		{
			for (k = 0; k < dgsChars.mgd.sizImage.cx; k++)
			{
				// NOTE:  assumes char data was decoded in LOWER bits of _rgbChars...
				_pimgChars->rgbData[i * 64 + j * 8 + k] = _rgiDkongCharColors[i >> 2][_pimgChars->rgbData[i * 64 + j * 8 + k]];
			}
		}
	}

	//
	//  Decode sprites
	//
	
	static const DECODE_GRAPHICS dgs =
	{
		0x00000001,
		{
			{ 16, 16 },
			128,
			16 * 8,
			2,
			{ 0, 16 * 16 * 128 },
			2,
			{ 15*8, 14*8, 13*8, 12*8, 11*8, 10*8, 9*8, 8*8,
				7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
			{ 0, 1, 2, 3, 4, 5, 6, 7,	/* the two halves of the sprite are separated */
				64*16*16+0, 64*16*16+1, 64*16*16+2, 64*16*16+3, 64*16*16+4, 64*16*16+5, 64*16*16+6, 64*16*16+7 },
		},

		1,
		
		4, 
		
		{
			0 * 128,
			2 * 128,
			1 * 128,
			3 * 128
		},
		0, 
		NULL
	};

	MyGfxDecode3(dgs, tmpstorage + 0x1000, _pimgSprites);

	//
	//
	//
	
	static const EXECZ80_MEM_DESCRIPTOR rgmd[] = 
	{
		{ 0x0000, 0xFFFF, 0, NULL, NULL },
		{ 0x7C00, 0x7C00, 0, M_RDMEM, M_WRMEM },
		{ 0x7C80, 0x7C80, 0, M_RDMEM, M_WRMEM },
		{ 0x7D00, 0x7D07, 0, M_RDMEM, M_WRMEM },
		{ 0x7D80, 0x7D80, 0, M_RDMEM, M_WRMEM },
		{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
	};

	EXECZ80_CREATE cs;

	ZERO_INIT(cs);

	cs.rgbMemory = RAM;
	cs.pfnMemRead = M_RDMEM;
	cs.pfnMemWrite = M_WRMEM;
	cs.pmd = rgmd;
	
	HRESULT hr = EclCreateZ80(cs, _pZ80);

	if (FAILED(hr))
		return hr;

	_pZ80->Reset();

	return S_OK;
}

static HRESULT InitJr(GAME_LOAD_INFO &gli)
{
	static const SIZE _sizBuffer = { 256 + 16, 256 + 16};
	static const RECT _rcDisplay = { 16, 0, 248, 256 };

	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	int i, j, k;

	for (i = 0; i < 16; i++)
		gli.ppe[i] = _rgpe[i];

	for (i = 0; i < 4 * 16; i++)
		gli.ppe[16 + i] = _rgpe[_rgiDkongjrSpriteColors[i]];

	gli.cpe = 16 + 16 * 4;

	if (readroms(RAM, _rgDKJROMS, "dkongjr", gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	BYTE tmpstorage[0x4000];

	if (readroms(tmpstorage, JrGFXROMs, "dkongjr", gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//
	//
	
	static DECODE_GRAPHICS dgsChars =
	{
		0x00000000,
		{
			{ 8, 8 },
			512,
			8 * 8,
			2,
			{ 0, 8 * 8 * 512 },
			2,
			{ 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
			{ 0, 1, 2, 3, 4, 5, 6, 7 },
		},

		1,
		
		1, 
		
		{
			0
		},
		0, 
		NULL
	};

	MyGfxDecode3(dgsChars, tmpstorage, _pimgChars);

	//
	//
	//
	
	for (i = 0; i < dgsChars.mgd.cImage; i++)
	{
		for (j = 0; j < dgsChars.mgd.sizImage.cy; j++)
		{
			for (k = 0; k < dgsChars.mgd.sizImage.cx; k++)
			{
				// NOTE:  assumes char data was decoded in LOWER bits of _rgbChars...
//				_rgbChars[i][j][k] = _rgiDkongjrCharColors[i >> 2][_rgbChars[i][j][k]];
				_pimgChars->rgbData[i * 64 + j * 8 + k] = _rgiDkongjrCharColors[i >> 2][_pimgChars->rgbData[i * 64 + j * 8 + k]];
			}
		}
	}

	//
	//  Decode sprites
	//
	
	static const DECODE_GRAPHICS dgs =
	{
		0x00000001,
		{
			{ 16, 16 },
			128,
			16 * 8,
			2,
			{ 0, 16 * 16 * 128 },
			2,
			{ 15*8, 14*8, 13*8, 12*8, 11*8, 10*8, 9*8, 8*8,
				7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
			{ 0, 1, 2, 3, 4, 5, 6, 7,	/* the two halves of the sprite are separated */
				64*16*16+0, 64*16*16+1, 64*16*16+2, 64*16*16+3, 64*16*16+4, 64*16*16+5, 64*16*16+6, 64*16*16+7 },
		},

		1,
		
		4, 
		
		{
			0 * 128,
			2 * 128,
			1 * 128,
			3 * 128
		},
		0, 
		NULL
	};

	MyGfxDecode3(dgs, tmpstorage + 0x2000, _pimgSprites);

	//
	//
	//
	
	static const EXECZ80_MEM_DESCRIPTOR rgmd[] = 
	{
		{ 0x0000, 0xFFFF, 0, M_RDMEM, M_WRMEM },
		{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
	};

	EXECZ80_CREATE cs;

	ZERO_INIT(cs);

	cs.rgbMemory = RAM;
	cs.pfnMemRead = M_RDMEM;
	cs.pfnMemWrite = M_WRMEM;
	cs.pmd = rgmd;
	
	HRESULT hr = EclCreateZ80(cs, _pZ80);

	if (FAILED(hr))
		return hr;

	_pZ80->Reset();

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE M_RDMEM(WORD A)
{
#ifdef _DEBUG

	static int _iTrace = 0;
	static const int _rgcTrace[] = { 0, 1000 };

	_iTrace++;

//	if (A < 0x2000 || A >= 0x4000)
	{
		if ((_iTrace >= _rgcTrace[0]) && (_iTrace <= _rgcTrace[1]))
		{
//			TRACE("%04X RAM[%04X]\n", _iTrace, A);
		}
	}
#endif

	switch (A)
	{
		case 0x7C00:
		{
			BYTE b = 0x00;

			if (osd_key_pressed(OSD_KEY_RIGHT))
				b ^= 0x01;

			if (osd_key_pressed(OSD_KEY_LEFT))
				b ^= 0x02;

			if (osd_key_pressed(OSD_KEY_UP))
				b ^= 0x04;

			if (osd_key_pressed(OSD_KEY_DOWN))
				b ^= 0x08;

			if (osd_key_pressed(OSD_KEY_CONTROL))
				b ^= 0x10;

			return b;
		}

		case 0x7C80:
		{
			BYTE b = 0x00;

			return b;
		}

		case 0x7d00:
		{
			BYTE b = 0x00;

			if (osd_key_pressed(OSD_KEY_1))
				b ^= 0x04;

			if (osd_key_pressed(OSD_KEY_2))
				b ^= 0x08;

			if (osd_key_pressed(OSD_KEY_3))
				b ^= 0x80;

			return b;
		}

		case 0x7D80:
		{
			BYTE b = 0x87;

			return b;
		}
	}

	return RAM[A];
}

static void M_WRMEM(WORD A, BYTE V)
{
	if (A < 0x4000)
		return;

#ifdef _DEBUG
	static int _iTrace = 0;
	static const int _rgcTrace[] = { -1, -1 };

//	if (_iTrace == 0x110D)
//		TRACE("\n");

	if (A >= 0x7400 && A < 0x7800)
	{
		_iTrace++;

		if ((_iTrace >= _rgcTrace[0]) && (_iTrace <= _rgcTrace[1]))
		{
			TRACE("%04X RAM[%04X]=%02X\n", _iTrace, A, V);
		}
	}
#endif

	switch (A)
	{
		case 0x7C00:
		{
			int offset = A - 0x7C00;
			int data = V;

			static last, hit;

			if (data != 6)
				hit = 0;

			if (last != data)
			{
				last = data;

				switch (data)
				{
					case 6:
						last = 4;
						if (!hit)
						{
							SoundSetEnvelope(3, _rgsi[data + 8].rgbData, _rgsi[data + 8].cbSize, FALSE);

							hit = 1;
						}
						break;

					case 13:
						last = 11;
						SoundSetEnvelope(3, _rgsi[data + 8].rgbData, _rgsi[data + 8].cbSize, FALSE);
						break;

					case 8:		  /* these samples repeat */
					case 9:
					case 10:
					case 11:
					case 4:
					case 3:
					{
						if (_rgsi[data + 8].cbSize != 0)
						{
							SoundSetEnvelope(4, _rgsi[data + 8].rgbData, _rgsi[data + 8].cbSize, TRUE);
						}

						break;
					}

					default:		  /* the rest do not */
					{
						if (data != 0)
						{
							if (_rgsi[data + 8].cbSize != 0)
							{
								SoundSetEnvelope(4, _rgsi[data + 8].rgbData, _rgsi[data + 8].cbSize, FALSE);
							}
						}
				
						break;
					}
				}
			}

			break;
		}

		case 0x7D80:
		{
			int offset = A - 0x7D80;
			int data = V;

			static last_dead;

			if (last_dead != data)
			{
				last_dead = data;

				if (data)
				{
					SoundSetEnvelope(0, _rgsi[24].rgbData, _rgsi[24].cbSize, FALSE);
					SoundSetEnvelope(1, NULL, 0, FALSE);
					SoundSetEnvelope(2, NULL, 0, FALSE);
					SoundSetEnvelope(3, NULL, 0, FALSE);
					SoundSetEnvelope(4, NULL, 0, FALSE);
					SoundSetEnvelope(5, NULL, 0, FALSE);
				}
			}

			break;
		}
	}

	if (A >= 0x7D00 && A <= 0x7D07)
	{
		int offset = A - 0x7D00;
		int data = V;

		static state[8];

		if (state[offset] != data)
		{
			if ((_rgsi[offset].cbSize != 0) && (data))
			{
				SoundSetEnvelope(offset, _rgsi[offset].rgbData, _rgsi[offset].cbSize, FALSE);
			}

			state[offset] = data;
		}
	}

	RAM[A] = V;
}

extern void SoundUpdateMixer(LPBYTE rgbBuffer, LONG cbBuffer);

static void UpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
	SoundUpdateMixer(rgbBuffer, cbBuffer);
}

static int Exec()
{
	return _pZ80->Exec(50000) + ((RAM[0x7d84] == 0) ? 0 : _pZ80->Nmi());
}

static void Term()
{
	_pZ80->Release();
	FreeImage(_pimgSprites);
	FreeImage(_pimgChars);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static const int _rgiHex[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 17, 18, 19, 20, 21, 22 };

static void RenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	int i;

#if 0
	if (GetAsyncKeyState('C') < 0)
	{
		for (i = 0; i < 512; i++)
		{
			DrawChar((LPBYTE)_rgbChars, i, 0, 8 * (i % 32), 8 * (i / 32),
				rgbBuffer, lPitch);
		}
	}
#endif

	//
	//  Draw 32x32 characters
	//

	const int nCharset = 256 * RAM[0x7c80];
	
	for (i = 0 * 32; i < 32 * 30; i++)
	{
		const int x = 8 * (31 - (i / 32));
		const int y = 8 * (i % 32);

		DrawChar(_pimgChars->rgbData, RAM[0x7400 + i] + nCharset,
			x, y, rgbBuffer, lPitch);
	}

	//
	//  Draw sprites
	//

	struct SPRITE
	{
		BYTE x, i, c, y;
	};

	const SPRITE *const ps = (SPRITE*)&RAM[0x6900];

	for (i = 0; i < 96; i++)
	{
		if (ps[i].x == 0)
			continue;

//		DrawSprite(rgbBuffer, lPitch, ps[i].x, ps[i].y,
//			ps[i].i + 2 * (ps[i].c & 0x80), 16 + ((ps[i].c & 0x0F) << 2));

//		DrawSprite(rgbBuffer, lPitch, 
//			ps[i].i + 2 * (ps[i].c & 0x80), 16 + ((ps[i].c & 0x0F) << 2));

		DrawSprite(_pimgSprites->rgbData, _pimgSprites->rgbMask,
			ps[i].i + 2 * (ps[i].c & 0x80), 16 + ((ps[i].c & 0x0F) << 2),
			ps[i].x - 7, ps[i].y - 8, rgbBuffer, lPitch);
	
//		DrawSpriteI(_pimgSprites->rgbData, _pimgSprites->rgbMask,
//			ps[i].i + 2 * (ps[i].c & 0x80), 16 + ((ps[i].c & 0x0F) << 2),
//			ps[i].x - 7, ps[i].y - 8, rgbBuffer, lPitch);
	}
}

/*---------------------------------------------------------------------------

	Define game descriptors

---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdDonkeyKong = 
{
	TEXT("Donkey Kong"),
	NULL,
	Init,
	Exec,
	RenderScene,
	UpdateSound,
	Term
};

GAME_DESCRIPTOR _gmdDonkeyKongJr = 
{
	TEXT("Donkey Kong Jr."),
	NULL,
	InitJr,
	Exec,
	RenderScene,
	NULL,
	Term
};

