/*---------------------------------------------------------------------------

	File:
		Gyruss.cpp
	
	Abstract:
		Implements the "Gyruss" game driver functions.

	Revisions:
		05-May-97 EAM Created

	Copyright (C) 1997 Edward Massey

	Hardware information adopted from MAME.

	2 Z-80s, 1 6809, 5 8910, 1 8309

	TODO:  Decode sprites...  layout using mirroring and bank selection to
		minimize sprite drawing code.
	TODO:  Dynamically allocate sprite buffers.
	TODO:  Interleave sprite mask/data
	TODO:  Left shift sprite data during decode
	TODO:  Get colors correct
	CONSIDER:  Have directdraw color fill the buffer with black
	TODO:  Optimize SprTrans

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "game.h"
#include "game1.h"
#include "../z80/execz80.h"
#include "../main/ay8910.h"

static AY8910 _rgAy8910[5];

extern void SoundUpdateMixer(LPBYTE rgbBuffer, LONG cbBuffer);
extern void SoundSetEnvelope(int iChnl, const char *pData, int cSamples, BOOL fLoop);

extern void sound_command_w(int offset,int data);
extern int pending_commands;
extern int sound_command_r(int offset);

static int gyruss_portA_r(/*int offset*/);

struct SPRITE
{
	BYTE y;
	BYTE i; // sprite image number
	
	union
	{
		BYTE c;
		BYTE attr;
	};
	
	BYTE x;
};

/*
	   offs+0 :  Ypos
	   offs+1 :  Sprite number
	   offs+2 :  Attribute in the form HF-VF-BK-DH-p3-p2-p1-p0
				 where  HF is horizontal flip
						VF is vertical flip
						BK is for bank select
						DH is for double height (if set sprite is 16*16, else is 16*8)
						px is palette weight
	   offs+3 :  Xpos
*/

static IMAGE *_pimgChars;
static IMAGE *_pimgSprites1;
static IMAGE *_pimgSprites2;
static IMAGE *_pimgSprites3;
static IMAGE *_pimgSprites4;
static IMAGE *_pimgSprites5;
static IMAGE *_pimgSprites6;

static SPRITE *_ps = (SPRITE*)&RAM[0xA000];

BYTE *gyruss_6809_drawplanet = &RAM[0xa7fc];
BYTE *gyruss_6809_drawship   = &RAM[0xa7fd];
BYTE *gyruss_nmi_enable      = &RAM[0xC180];
BYTE *gyruss_video_ram       = &RAM[0x8400];
BYTE *gyruss_color_ram       = &RAM[0x8000];

#define YTABLE_START (0x0000)
#define SINTABLE_START (0x0400)
#define COSTABLE_START (0x0600)

static BYTE _rgbROM6809[0x2000];

static BYTE _rgbROMAudio[0x10000];

BYTE *gyruss_ytable          = &_rgbROM6809[0x0000];
BYTE *gyruss_sintable        = &_rgbROM6809[0x0400];
BYTE *gyruss_costable        = &_rgbROM6809[0x0600];

static IExecZ80 *_pZ80;
static IExecZ80 *_pZ80Audio;

struct SAMPLE_INFO
{
	LPCTSTR pszFilename;
	int cbSize;
	char *rgbData;
};

static char _rgbSamples[7][155 * 1025];

static SAMPLE_INFO _rgsiGyruss[] = 
{
	{ "audio01.out" },
	{ "audio02.out" },
	{ "audio03.out" },
	{ "audio04.out" },
	{ "audio05.out" },
	{ "audio06.out" },
	{ "audio07.out" },
};

int gyruss_sh_init()
{
	int i;

	for (i = 0; i < DIM(_rgsiGyruss); i++)
	{
		_rgsiGyruss[i].rgbData = _rgbSamples[i];

		TStr<_MAX_PATH> strFile("gyruss/%s", _rgsiGyruss[i].pszFilename);

		FILE *pf = fopen(strFile, "rb");

		if (pf != NULL)
		{
			_rgsiGyruss[i].cbSize = fread(_rgsiGyruss[i].rgbData, 1, sizeof(_rgbSamples[0]), pf);

			fclose(pf);
		}
	}

	return 0;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

/* these are NOT the original color PROMs */
static const BYTE color_prom[] =
{
	/* char palette */
	0x00, 0x90, 0xB4, 0xD8, 0x00, 0xB4, 0xD8, 0xFC, 0x00, 0xA4, 0x92, 0xFF, 0x00, 0x48, 0x6C, 0x90,
	0x00, 0xD0, 0x88, 0xFF, 0x00, 0x49, 0x92, 0xFF, 0x00, 0x0D, 0x11, 0x1E, 0x90, 0xFC, 0xB4, 0xD8,
	0xD8, 0xB4, 0x90, 0xB4, 0x00, 0x6C, 0x48, 0xB4, 0x00, 0x49, 0x92, 0xFF, 0x00, 0x55, 0x9E, 0x1A,
	0x00, 0x48, 0x6C, 0x90, 0x00, 0xC4, 0xC8, 0xF0, 0x00, 0x40, 0x80, 0xE0, 0x00, 0x6C, 0x90, 0xB4,

	/* Sprite palette */
	0x00, 0x01, 0x02, 0x03, 0xDF, 0xDF, 0x1C, 0xFF, 0x12, 0x13, 0x00, 0xB6, 0x00, 0x01, 0x92, 0xE0,
	0x00, 0x01, 0x02, 0x03, 0xFF, 0xDB, 0x1C, 0xFF, 0x12, 0x1F, 0x01, 0xB6, 0x03, 0x01, 0x92, 0xE0,
	0x00, 0x01, 0x02, 0x03, 0xDF, 0xDB, 0xDC, 0xFF, 0x12, 0x1F, 0x1C, 0xB6, 0x03, 0x01, 0x92, 0xE0,
	0x00, 0x01, 0x02, 0x03, 0xDF, 0xDB, 0x1C, 0xFF, 0x12, 0x1F, 0x1C, 0xB6, 0x00, 0x01, 0x92, 0xE0,
	0x00, 0x13, 0x02, 0x23, 0xFF, 0xBA, 0xBC, 0x0E, 0x92, 0x37, 0x00, 0x90, 0xB6, 0x49, 0x92, 0x0A,
	0x48, 0x6D, 0x02, 0x03, 0x49, 0xB6, 0x1C, 0x09, 0xB6, 0x1F, 0x00, 0x90, 0x92, 0x49, 0x92, 0x06,
	0x00, 0x0B, 0x02, 0x03, 0x7B, 0xB6, 0x95, 0x09, 0x17, 0x1F, 0x00, 0x94, 0x0E, 0x91, 0x96, 0x13,
	0x00, 0xA8, 0x02, 0x03, 0x49, 0xB6, 0x95, 0x09, 0xED, 0x1F, 0x00, 0x94, 0xE8, 0x09, 0x51, 0x1F,
	0x00, 0x1B, 0xD4, 0x96, 0xE8, 0x1B, 0xF4, 0xB6, 0xFC, 0x17, 0xE4, 0x92, 0xFC, 0x13, 0xE8, 0xFF,
	0x00, 0xFF, 0x0B, 0x03, 0x92, 0xE0, 0x17, 0x19, 0x12, 0x1F, 0x0B, 0xBA, 0x96, 0x5C, 0x92, 0xE0,
	0x00, 0x01, 0x02, 0x03, 0x08, 0x10, 0x1C, 0x09, 0x12, 0x1F, 0x00, 0x90, 0xBC, 0x49, 0x92, 0xFF,
	0x00, 0x01, 0x02, 0x03, 0x08, 0x10, 0x1C, 0x09, 0x12, 0x1F, 0x00, 0x90, 0x7C, 0x49, 0x92, 0xFF,
	0x00, 0x01, 0x02, 0x03, 0x08, 0x10, 0x1C, 0x09, 0x12, 0x1F, 0x00, 0x90, 0xFC, 0x49, 0x92, 0xFF,
	0x01, 0x01, 0x01, 0x01, 0x15, 0x01, 0x01, 0x01, 0x1D, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
	0x48, 0x01, 0x02, 0x03, 0x08, 0x10, 0x1C, 0x09, 0x12, 0x1F, 0x00, 0x90, 0xFC, 0x49, 0x92, 0xFF,
	0x00, 0x01, 0x03, 0x03, 0x09, 0x9F, 0x17, 0x57, 0x12, 0xE0, 0x9B, 0x0E, 0xFC, 0xF0, 0x53, 0xC0
};

static const BYTE color_prom2[] =
{
	/* B4 - palette */
	0x00,0x00,0x05,0x06,0x07,0xFC,0x05,0xBD,0xB5,0xFD,0x05,0xB0,0xA5,0xE0,0x00,0xF7,
	0x00,0x00,0xF8,0x07,0x07,0xFD,0xF8,0xFA,0x05,0xDE,0x50,0x51,0x32,0xFD,0x30,0xF7,

	/* B5 - palette */
	0x00,0x3E,0x3E,0x80,0xFE,0x00,0xAC,0xEE,0xAC,0xC0,0x14,0x00,0x28,0x38,0x16,0xBC,
	0x00,0x3E,0x00,0xC0,0xFE,0xC0,0x3E,0x80,0x3E,0xF6,0x00,0x80,0x80,0x00,0x0C,0xBC,

	/* E12 (or F15?) - char lookup table */
	/* The PROM is 256x4, but only the first 128 bytes are used */
	0x0F,0x0D,0x0F,0x0C,0x0A,0x04,0x01,0x0F,0x0B,0x04,0x01,0x0F,0x0C,0x04,0x01,0x0F,
	0x0E,0x04,0x01,0x0F,0x00,0x04,0x01,0x0F,0x0A,0x04,0x05,0x01,0x0B,0x04,0x05,0x01,
	0x0C,0x04,0x05,0x01,0x0E,0x04,0x05,0x01,0x00,0x04,0x05,0x00,0x0A,0x06,0x08,0x02,
	0x0B,0x06,0x08,0x02,0x0C,0x06,0x08,0x02,0x0E,0x06,0x08,0x02,0x00,0x06,0x08,0x02,
	0x00,0x01,0x04,0x0F,0x00,0x04,0x02,0x06,0x00,0x01,0x08,0x04,0x00,0x0D,0x01,0x05,
	0x00,0x02,0x03,0x01,0x00,0x0C,0x0F,0x03,0x00,0x05,0x02,0x08,0x0A,0x01,0x04,0x03,
	0x00,0x06,0x0F,0x02,0x00,0x0F,0x03,0x05,0x00,0x03,0x01,0x0F,0x0A,0x02,0x0D,0x05,
	0x00,0x01,0x0F,0x08,0x0A,0x02,0x0D,0x05,0x0A,0x0B,0x09,0x0F,0x09,0x09,0x09,0x09,

	/* E9 - sprite lookup table */
	0x0F,0x0D,0x0F,0x05,0x00,0x0E,0x06,0x0A,0x00,0x04,0x09,0x01,0x00,0x04,0x09,0x01,
	0x00,0x04,0x09,0x01,0x00,0x0C,0x05,0x01,0x00,0x0E,0x05,0x01,0x00,0x0D,0x05,0x01,
	0x00,0x0B,0x05,0x01,0x00,0x01,0x0F,0x04,0x00,0x01,0x0F,0x04,0x00,0x01,0x0F,0x04,
	0x00,0x08,0x07,0x0C,0x00,0x01,0x0F,0x04,0x00,0x01,0x0F,0x04,0x00,0x0A,0x05,0x01,
	0x00,0x05,0x09,0x01,0x00,0x0B,0x0D,0x05,0x00,0x06,0x05,0x01,0x00,0x0A,0x03,0x01,
	0x00,0x0C,0x03,0x01,0x00,0x0E,0x03,0x01,0x00,0x0D,0x03,0x01,0x00,0x0B,0x03,0x01,
	0x00,0x0E,0x0C,0x0F,0x00,0x05,0x03,0x01,0x00,0x0E,0x06,0x09,0x00,0x04,0x09,0x05,
	0x00,0x09,0x0E,0x06,0x00,0x04,0x0E,0x05,0x00,0x09,0x0E,0x05,0x00,0x0B,0x05,0x01,
	0x00,0x0C,0x07,0x01,0x00,0x05,0x0F,0x09,0x00,0x05,0x04,0x09,0x00,0x0A,0x03,0x01,
	0x00,0x0C,0x02,0x03,0x00,0x0C,0x06,0x03,0x00,0x0C,0x06,0x09,0x00,0x0C,0x06,0x01,
	0x00,0x0E,0x06,0x0C,0x00,0x0C,0x02,0x0F,0x00,0x0C,0x02,0x09,0x00,0x0C,0x02,0x01,
	0x00,0x01,0x08,0x0F,0x00,0x0E,0x06,0x0F,0x00,0x09,0x0A,0x0F,0x00,0x05,0x06,0x0F,
	0x00,0x0B,0x09,0x05,0x00,0x0A,0x06,0x0C,0x00,0x0A,0x06,0x09,0x00,0x0A,0x02,0x09,
	0x00,0x06,0x01,0x0F,0x00,0x04,0x01,0x0F,0x00,0x0A,0x03,0x01,0x00,0x0A,0x02,0x0C,
	0x00,0x05,0x09,0x01,0x00,0x0A,0x02,0x01,0x00,0x0E,0x02,0x09,0x00,0x0E,0x02,0x0C,
	0x00,0x01,0x04,0x0F,0x00,0x01,0x04,0x0F,0x00,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00
};

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

//
//
//

static const TCHAR _szPath[] = "gyruss";

RAGE_ROMTABLE_BEGIN(ROMs)
	RAGE_ROMTABLE_ENTRY("gya-1.bin", 0x0000, 0x2000)
	RAGE_ROMTABLE_ENTRY("gya-2.bin", 0x2000, 0x2000)
	RAGE_ROMTABLE_ENTRY("gya-3.bin", 0x4000, 0x2000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(GFXROMs)
	RAGE_ROMTABLE_ENTRY("gy-6.bin",  0x0000, 0x2000)
	RAGE_ROMTABLE_ENTRY("gy-9.bin",  0x2000, 0x2000)
	RAGE_ROMTABLE_ENTRY("gy-7.bin",  0x4000, 0x2000)
	RAGE_ROMTABLE_ENTRY("gy-10.bin", 0x6000, 0x2000)
	RAGE_ROMTABLE_ENTRY("gy-8.bin",  0x8000, 0x2000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(ROM6809)
	RAGE_ROMTABLE_ENTRY("gy-5.bin",  0x0000, 0x2000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(ROMAudio)
	RAGE_ROMTABLE_ENTRY("gy-11.bin", 0x0000, 0x2000)
	RAGE_ROMTABLE_ENTRY("gy-12.bin", 0x2000, 0x2000)
	RAGE_ROMTABLE_ENTRY("gy-13.bin", 0x4000, 0x1000)
RAGE_ROMTABLE_END

static const DECODE_GRAPHICS dgsChars =
{
	0x00000001,
	{
		{ 8, 8 },
		512,
		16 * 8,
		2,
		{ 0, 4 },
		2,
		{ 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
		{ 0, 1, 2, 3, 8*8+0,8*8+1,8*8+2,8*8+3 },
	},

	1,
	
	4, 
	
	{
		0 * 512,  // 0123 0213 0321 0132 0312
		1 * 512,
		2 * 512,
		3 * 512
	},
	0, 
	NULL
};

static const DECODE_GRAPHICS dgsSprites1 =
{
	0x00000001,
	{
		{ 16, 8 },
		128,
		64 * 8,
		4,
		{  4, 0, 0x2000*8+4, 0x2000*8 },
		4,
		{ 39*8, 38*8, 37*8, 36*8, 35*8, 34*8, 33*8, 32*8, 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
		{ 0, 1, 2, 3,  8*8, 8*8+1, 8*8+2, 8*8+3 },
//			{ 0, 1, 2, 3,  8*8, 8*8+1, 8*8+2, 8*8+3, 16*8+0, 16*8+1, 16*8+2, 16*8+3,  24*8, 24*8+1, 24*8+2, 24*8+3 },
	},

	1,
	
	4, 
	
	{
		0 * 128,
		1 * 128,
		2 * 128,
		3 * 128
	},
	0, 
	NULL
};

static const DECODE_GRAPHICS dgsSprites2 =
{
	0x00000001,
	{
		{ 16, 16 },
		128,
		64 * 8,
		4,
		{  4, 0, 0x2000*8+4, 0x2000*8 },
		4,
		{ 39*8, 38*8, 37*8, 36*8, 35*8, 34*8, 33*8, 32*8, 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
		{ 0, 1, 2, 3,  8*8, 8*8+1, 8*8+2, 8*8+3, 16*8+0, 16*8+1, 16*8+2, 16*8+3,  24*8, 24*8+1, 24*8+2, 24*8+3 },
	},

	1,
	
	4, 
	
	{
		0 * 128,
		1 * 128,
		2 * 128,
		3 * 128
	},
	0, 
	NULL
};

static BYTE M_RDMEM(WORD A);
static void M_WRMEM(WORD A, BYTE);
void gyruss_queuereg_w (WORD offset, BYTE data);
static BYTE M_RDMEMAudio(WORD A);
static void M_WRMEMAudio(WORD A, BYTE);
static BYTE GyrussAudioIn(BYTE A);
static void GyrussAudioOut(BYTE, BYTE);

static const EXECZ80_MEM_DESCRIPTOR rgmd[] = 
{
#if 0
	{ 0x0000, 0xFFFF, 0, M_RDMEM, M_WRMEM },
#else
	{ 0x0000, 0xFFFF, 0, NULL,    NULL    }, // start with direct access
	{ 0xA700, 0xA700, 0, NULL,    M_WRMEM },
	{ 0xA702, 0xA702, 0, NULL,    gyruss_queuereg_w },
	{ 0xC000, 0xC000, 0, M_RDMEM, M_WRMEM },
	{ 0xC080, 0xC080, 0, M_RDMEM, M_WRMEM },
	{ 0xC0A0, 0xC0A0, 0, M_RDMEM, M_WRMEM },
	{ 0xC0C0, 0xC0C0, 0, M_RDMEM, M_WRMEM },
	{ 0xC0E0, 0xC0E0, 0, M_RDMEM, M_WRMEM },
	{ 0xC100, 0xC100, 0, M_RDMEM, M_WRMEM },
#endif
	{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
};

static const EXECZ80_MEM_DESCRIPTOR rgmdAudio[] = 
{
#if 0
	{ 0x0000, 0xFFFF, 0, M_RDMEMAudio, M_WRMEMAudio },
#else
	{ 0x0000, 0xFFFF, 0, NULL,         NULL         }, // start with direct acces
	{ 0x8000, 0x8000, 0, M_RDMEMAudio, NULL         }, // read command from latch
	{ 0x0000, 0x3FFF, 0, NULL,         M_WRMEMAudio }, // ROM
#endif
	{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
};

static const SIZE _sizBuffer = { 256 + 16, 256 + 16};
static const RECT _rcDisplay = { 16, 0, 256 - 16, 256 };

static HRESULT Init(GAME_LOAD_INFO &gli)
{
	gyruss_sh_init();

	int i;

	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	for (i = 0; i < 256; i++)
	{
		int bits;

		bits = (color_prom[i] >> 5) & 0x07;
		gli.ppe[i].peRed = bits * (255 / 7);

		bits = (color_prom[i] >> 2) & 0x07;
		gli.ppe[i].peGreen  = bits * (255 / 7);

		bits = (color_prom[i] >> 0) & 0x03;
		gli.ppe[i].peBlue = bits * (255 / 3);
	}

	gli.cpe = 256;

	//
	//
	//

	BYTE tmpstorage[0xFFFF]; // A000];

	ZERO_INIT(tmpstorage);

	if (readroms(RAM, ROMs, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	if (readroms(_rgbROM6809, ROM6809, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	if (readroms(_rgbROMAudio, ROMAudio, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	if (readroms(tmpstorage, GFXROMs, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//  Decode characters
	//

	MyGfxDecode3(dgsChars, tmpstorage, _pimgChars);

	//
	//  Decode sprites1
	//

	MyGfxDecode3(dgsSprites1, tmpstorage + 0x2000, _pimgSprites1);
	MyGfxDecode3(dgsSprites1, tmpstorage + 0x2010, _pimgSprites2);
	MyGfxDecode3(dgsSprites1, tmpstorage + 0x6000, _pimgSprites3);
	MyGfxDecode3(dgsSprites1, tmpstorage + 0x6010, _pimgSprites4);
	MyGfxDecode3(dgsSprites2, tmpstorage + 0x2000, _pimgSprites5);
	MyGfxDecode3(dgsSprites2, tmpstorage + 0x6000, _pimgSprites6);

	//
	//
	//
	
	EXECZ80_CREATE cs;

	ZERO_INIT(cs);

	cs.rgbMemory   = RAM;
	cs.pfnMemRead  = M_RDMEM;
	cs.pfnMemWrite = M_WRMEM;
	cs.pmd         = rgmd;
	
	HRESULT hr = EclCreateZ80(cs, _pZ80);

	if (FAILED(hr))
		return hr;

	//
	//
	//
	
	ZERO_INIT(cs);

	cs.rgbMemory    = _rgbROMAudio;
	cs.pfnMemRead   = M_RDMEMAudio;
	cs.pfnMemWrite  = M_WRMEMAudio;
	cs.pfnPortRead  = GyrussAudioIn;
	cs.pfnPortWrite = GyrussAudioOut;
	cs.pmd          = rgmdAudio;
	
	hr = EclCreateZ80(cs, _pZ80Audio);

	if (FAILED(hr))
		return hr;

	_rgAy8910[2].m_pfnReadPortA = gyruss_portA_r;

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

/*
	{ 0x0000, 0x5fff, MRA_ROM },
	{ 0x8000, 0x87ff, MRA_RAM },
	{ 0x9000, 0x9fff, MRA_RAM },
	{ 0xa700, 0xa700, MRA_RAM },
	{ 0xa7fc, 0xa7fd, MRA_RAM },
	{ 0xc080, 0xc080, input_port_0_r },
	{ 0xc0a0, 0xc0a0, input_port_1_r },
	{ 0xc0c0, 0xc0c0, input_port_2_r },
	{ 0xc0e0, 0xc0e0, input_port_3_r },
	{ 0xc000, 0xc000, input_port_4_r },
	{ 0xc100, 0xc100, input_port_5_r },
*/

static BYTE M_RDMEM(WORD A)
{
	switch (A)
	{
		case 0xC080:
		{
			BYTE b = 0xFF;

			if (osd_key_pressed(OSD_KEY_3))
				b ^= 0x04;

			if (osd_key_pressed(OSD_KEY_1))
				b ^= 0x08;

			if (osd_key_pressed(OSD_KEY_2))
				b ^= 0x10;

			return b;
		}

		case 0xC0A0:
		{
			BYTE b = 0xFF;

			if (osd_key_pressed(OSD_KEY_LEFT))
				b ^= 0x01;

			if (osd_key_pressed(OSD_KEY_RIGHT))
				b ^= 0x02;

			if (osd_key_pressed(OSD_KEY_UP))
				b ^= 0x04;

			if (osd_key_pressed(OSD_KEY_DOWN))
				b ^= 0x08;

			if (osd_key_pressed(OSD_KEY_CONTROL))
				b ^= 0x10;

			return b;
		}

		case 0xC0C0:
		{
			BYTE b = 0xFF;

			return b;
		}

		case 0xC0E0:
		{
			BYTE b = 0xAF;

			return b;
		}

		case 0xC000:
		{
			BYTE b = 0x3B;

			return b;
		}

		case 0xC100:
		{
			BYTE b = 0xFE;

			return b;
		}
	}

	return RAM[A];
}

/*

optimizations:

avoid the * ro multiply by making tables that are looked up as a
function of theta2 + 1 and ro

mirror (negative) costable and sintable

better yet, just make a 256x256 table of x,y pairs!!!

basically, this routine generates x,y coordinates based
on x,y,attr&0x10

*/

// NOTE:  the following two routines are VERY similar to MAME...

/* convert sprite coordinates from polar to cartesian */
static int SprTrans(SPRITE *ps)
{
	const int ro     = gyruss_ytable[ps->y];
	const int theta2 = 2 * ps->x;

	//
	//
	//
	
	ps->y = (gyruss_costable[theta2+1] * ro) >> 8;

	if (ps->y >= 0x80)
	{
		ps->y = 0;
		return 0;
	}

	if (gyruss_costable[theta2] != 0)	/* negative */
	{
		if (ps->y >= 0x78)	/* avoid wraparound from top to bottom of screen */
		{
			ps->y = 0;
			return 0;
		}
		ps->y = -ps->y;
	}

	//
	//
	//
	
	ps->x = (gyruss_sintable[theta2+1] * ro) >> 8;

	if (ps->x >= 0x80)
 	{
		ps->y = 0;
		return 0;
	}

	if (gyruss_sintable[theta2] != 0)	/* negative */
		ps->x = -ps->x;

	//
	//
	//
	
	/* convert from logical coordinates to screen coordinates */
	if (ps->attr & 0x10)
		ps->y += 0x78;
	else
		ps->y += 0x7C;

	ps->x += 0x78;


    return 1;	/* queue this sprite */
}



	/* Gyruss uses a semaphore system to queue sprites, and when critic
	   region is released, the 6809 processor writes queued sprites onto
	   screen visible area.
	   When a701 = 0 and a702 = 1 gyruss hardware queue sprites.
	   When a701 = 1 and a702 = 0 gyruss hardware draw sprites.

           both Z80 e 6809 are interrupted at the same time by the
           VBLANK interrupt.  If there is some work to do (example
           A7FF is not 0 or FF), 6809 waits for Z80 to store a 1 in
           A701 and draw currently queued sprites
	*/
void gyruss_queuereg_w (WORD offset, BYTE data)
{
	if (data == 1)
	{
        int n;

		/* #0-#3 - ship */

		/* #4-#23 */
        if (*gyruss_6809_drawplanet)	/* planet is on screen */
		{
			SprTrans(&_ps[4]);	/* #4 - polar coordinates - ship */

			_ps[5].y = 0;	/* #5 - unused */

			/* #6-#23 - planet */
        }
		else
		{
			for (n = 4;n < 24;n += 2)	/* 10 double height sprites in polar coordinates - enemies flying */
			{
				SprTrans(&_ps[n]);

				_ps[n+1].y = 0;
			}
		}


		/* #24-#59 */
		for (n = 24;n < 60;n++)	/* 36 sprites in polar coordinates - enemies at center of screen */
			SprTrans(&_ps[n]);


		/* #60-#63 - unused */


		/* #64-#77 */
        if (*gyruss_6809_drawship == 0)
		{
			for (n = 64;n < 78;n++)	/* 14 sprites in polar coordinates - bullets */
				SprTrans(&_ps[n]);
		}
		/* else 14 sprites - player ship being formed */


		/* #78-#93 - stars */
	    for (n = 78;n < 86;n++)
		{
			if (SprTrans(&_ps[n]))
			{
				/* make a mirror copy */
				_ps[n + 8].x = _ps[n].y - 4;
				_ps[n + 8].y = _ps[n].x + 4;
			}
			else
				_ps[n + 8].y = 0;
		}
	}
}

static void M_WRMEM(WORD A, BYTE V)
{
	if (A < 0x6000)
		return;

	switch (A)
	{
		case 0xA702: // queue
			gyruss_queuereg_w(A, V);
			break;

		case 0xA700: // sprite bank
			_ps = (SPRITE*)&RAM[V ? 0xA200 : 0xA000];
			break;

		case 0xC100: // sound command
			sound_command_w(A, V);
			break;
	}

	RAM[A] = V;
}

//static BYTE rgbBufs[5][512];

static void UpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
	int rgnBufferLocal[512];
	int *pn = rgnBufferLocal;
	
	ZeroMemory(rgnBufferLocal, sizeof(rgnBufferLocal));

	_rgAy8910[0].UpdateSound(rgnBufferLocal, DIM(rgnBufferLocal));
	_rgAy8910[1].UpdateSound(rgnBufferLocal, DIM(rgnBufferLocal));
	_rgAy8910[2].UpdateSound(rgnBufferLocal, DIM(rgnBufferLocal));
	_rgAy8910[3].UpdateSound(rgnBufferLocal, DIM(rgnBufferLocal));
	_rgAy8910[4].UpdateSound(rgnBufferLocal, DIM(rgnBufferLocal));

	while (cbBuffer--)
	{
		int nAcc = 128 + *pn++ / 5;

		if (nAcc < 0)
			nAcc = 0;
		else if (nAcc > 255)
			nAcc = 255;

		*rgbBuffer++ = nAcc;
	}

/*
	_rgAy8910[0].UpdateSound(rgbBufs[0], DIM(rgbBufs[0]));
	_rgAy8910[1].UpdateSound(rgbBufs[1], DIM(rgbBufs[0]));
	_rgAy8910[2].UpdateSound(rgbBufs[2], DIM(rgbBufs[0]));
	_rgAy8910[3].UpdateSound(rgbBufs[3], DIM(rgbBufs[0]));
	_rgAy8910[4].UpdateSound(rgbBufs[4], DIM(rgbBufs[0]));

	SoundSetEnvelope(1, rgbBufferLocal, DIM(rgbBufferLocal), FALSE);
	SoundSetEnvelope(2, (char*)rgbBufs[1], DIM(rgbBufs[0]), TRUE);
	SoundSetEnvelope(3, (char*)rgbBufs[2], DIM(rgbBufs[0]), TRUE);
	SoundSetEnvelope(4, (char*)rgbBufs[3], DIM(rgbBufs[0]), TRUE);
	SoundSetEnvelope(5, (char*)rgbBufs[4], DIM(rgbBufs[0]), TRUE);

	SoundUpdateMixer(rgbBuffer, cbBuffer);
*/
}

enum
{
	GM_MENU,
	GM_GRAPHICS,
	GM_SOUNDS,
	GM_PAUSED,
	GM_RUNNING,
	GM_SETTINGS,
	GM_PALETTE
};

static int _iGameMode = GM_RUNNING;
static int _iColor = 0;
static BOOL _fFoobar;

/*
		for (;;)
		{
//			GetActiveGame()->Exec();
//			_mode_info.rgpfnExec[_iGameMode]();
			_pfnExec();

			if (!CheckMessages())
				break;

			Throttle(60);
		}
*/

static int Exec()
{
	int cCycles = 0;
	
	switch (_iGameMode)
	{
		case GM_RUNNING:
		{
			if (GetAsyncKeyState('G') < 0)
			{
				if (_fFoobar)
				{
				}
				else
				{
					_iGameMode = GM_GRAPHICS;
					_fFoobar = TRUE;
				}
			}
			else
			{
				_fFoobar = FALSE;

				cCycles += _pZ80->Exec(51200);
				
				if (*gyruss_nmi_enable != 0)
					cCycles += _pZ80->Nmi();

				BOOL fSoundEnable = FALSE; // TRUE;

				if (fSoundEnable)
				{
					_pZ80Audio->Exec(3579500 / 60);

					if (pending_commands > 0)
						_pZ80Audio->Irq(0xFF);
				}

				if (GetAsyncKeyState('S') < 0)
				{
					for (int n = 0; n <= 7; n++)
					{
						if (GetAsyncKeyState('0' + n) < 0)
						{
							static int current;

							if (n != current)
							{
								current = n;
								SoundSetEnvelope(0, _rgsiGyruss[n].rgbData, _rgsiGyruss[n].cbSize, FALSE);
							}
						}
					}
				}
			}

			break;
		}

		case GM_PAUSED:
		{
			break;
		}

		case GM_SOUNDS:
		{
			break;
		}

		case GM_GRAPHICS:
		{
			if (GetAsyncKeyState('G') < 0)
			{
				if (_fFoobar)
				{
				}
				else
				{
					_iGameMode = GM_RUNNING;
					_fFoobar = TRUE;
				}
			}
			else
			{
				_fFoobar = FALSE;
			}

			break;
		}

		case GM_MENU:
		{
			break;
		}
	}

	return cCycles;
}

static void Term()
{
	if (_pZ80 != NULL)
		_pZ80->Release();
	
	if (_pZ80Audio != NULL)
		_pZ80Audio->Release();

	FreeImage(_pimgChars);
	FreeImage(_pimgSprites1);
	FreeImage(_pimgSprites2);
	FreeImage(_pimgSprites3);
	FreeImage(_pimgSprites4);
	FreeImage(_pimgSprites5);
	FreeImage(_pimgSprites6);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

inline static void DrawSprite16x8(LPCBYTE rgbData, LPCBYTE rgbMask,
	int n, int color, int x, int y, LPBYTE rgbBuffer, LONG lPitch)
{
	//
	//
	//
	
	color |= (color << 8) | (color << 16) | (color << 24);

	//
	//
	//
	
	LPDWORD pDest = (LPDWORD)&rgbBuffer[lPitch * y + x];
	LPCDWORD pSrc = (LPCDWORD)&rgbData[n * 16 * 8];
	LPCDWORD pMask = (LPCDWORD)&rgbMask[n * 16 * 8];
	
	//
	//  Draw sprite, with transparency, and color scheme adjusted
	//  to offscreen buffer, 32-bits at a time
	//
	//  CONSIDER:  unroll the loop
	//
	
	lPitch = (lPitch / 4) - 4;
	
	for (int i = 0; i < 8; i++)
	{
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		pDest += lPitch;
	}
}

static void RenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	int i;

	//
	//  Clear screen
	//
	//  CONSIDER:  have directdraw clear it for us (based on flags/color in gli in Init())
	//
	
	for (i = 0; i < 256; i++)
		ZeroMemory(&rgbBuffer[lPitch * i], 256);

#ifdef _DEBUG
	if (_iGameMode == GM_GRAPHICS)
	{
		IMAGE *pimg = NULL;
		
		if (GetAsyncKeyState('1') < 0)
			pimg = _pimgChars;
		else if (GetAsyncKeyState('2') < 0)
			pimg = _pimgSprites1;
		else if (GetAsyncKeyState('3') < 0)
			pimg = _pimgSprites2;
		else if (GetAsyncKeyState('4') < 0)
			pimg = _pimgSprites3;
		else if (GetAsyncKeyState('5') < 0)
			pimg = _pimgSprites4;
		else if (GetAsyncKeyState('6') < 0)
			pimg = _pimgSprites5;
		else if (GetAsyncKeyState('7') < 0)
			pimg = _pimgSprites6;

		if (pimg != NULL)
		{
			int cx = (_rcDisplay.right - _rcDisplay.left) / pimg->sizImage.cx;
			int cy = (_rcDisplay.bottom - _rcDisplay.top) / pimg->sizImage.cy;
			
			for (i = 0; i < pimg->cImages; i++)
			{
				if (_rcDisplay.top + pimg->sizImage.cy * (i / cy) >= _rcDisplay.bottom - pimg->sizImage.cy)
					break;

				if (pimg->sizImage.cx == 16 && pimg->sizImage.cy == 16 && pimg->dwFlags & DGF_TRANSPARENT)
				{
					DrawSprite(pimg->rgbData, pimg->rgbMask, i, _iColor << 4,
						_rcDisplay.left + pimg->sizImage.cx * (i % cx),
						_rcDisplay.top + pimg->sizImage.cy * (i / cy),
						rgbBuffer, lPitch);
				}
				else if (pimg->sizImage.cx == 16 && pimg->sizImage.cy == 8 && pimg->dwFlags & DGF_TRANSPARENT)
				{
					DrawSprite16x8(pimg->rgbData, pimg->rgbMask, i, _iColor << 4,
						_rcDisplay.left + pimg->sizImage.cx * (i % cx),
						_rcDisplay.top + pimg->sizImage.cy * (i / cy),
						rgbBuffer, lPitch);
				}
				else if (pimg->sizImage.cx == 8 && pimg->sizImage.cy == 8)
				{
					DrawChar(pimg->rgbData, i, 0,
						_rcDisplay.left + pimg->sizImage.cx * (i % cx),
						_rcDisplay.top + pimg->sizImage.cy * (i / cy),
						rgbBuffer, lPitch);
				}
			}
		}

		return;
	}
#endif

	//
	//
	//
	
	const SPRITE *ps = &_ps[94 * 2];
	
	for (int c = (94 - 1) * 2; c > 0; c--, ps--)
	{
		int x, y, c, n;

		if (ps->c & 0x10)	/* double height */
		{
#if 1
			if (ps->y > 2)
			{
				x = ps->x;
				y = ps->y;
				c = ps->c & 0x0f;
				n = (ps->i / 2); //  | ((ps->c & 0xC0) << 1);

				IMAGE *pimg;

				if (ps->c & 0x20)
					pimg = _pimgSprites5;
				else
					pimg = _pimgSprites6;

				DrawSprite(pimg->rgbData, pimg->rgbMask,
					n, 64 + (c << 4), x, y, rgbBuffer, lPitch);
			}
#endif
		}
		else	/* single height */
		{
#if 1
			if (ps->y > 2)
			{
				x = ps->x;
				y = ps->y;
				c = ps->c & 0x0f;
				n = (ps->i / 2); // | ((ps->c & 0xC0) << 1);

				IMAGE *pimg;

#if 0
				if (ps->c & 0x20)
				{
					if (ps->i & 1)
						pimg = _pimgSprites2;
					else
						pimg = _pimgSprites1;
				}
				else
				{
					if (ps->i & 1)
						pimg = _pimgSprites4;
					else
						pimg = _pimgSprites3;
				}

				DrawSprite16x8(pimg->rgbData, pimg->rgbMask,
					n, 64 + (c << 4), x, y, rgbBuffer, lPitch);
#else
				if (ps->c & 0x20)
					pimg = _pimgSprites5;
				else
					pimg = _pimgSprites6;

				DrawSprite(pimg->rgbData, pimg->rgbMask,
					n, 64 + (c << 4), x, y, rgbBuffer, lPitch);
#endif
			}

			const SPRITE *ps2 = ps + 1;

			if (ps2->y > 2)
			{
				x = ps2->x;
				y = ps2->y;
				c = ps2->c & 0x0f;
				n = (ps2->i / 2); // | ((ps2->c & 0xC0) << 1);

				IMAGE *pimg;

#if 0
				if (ps->c & 0x20)
				{
					if (ps->i & 1)
						pimg = _pimgSprites2;
					else
						pimg = _pimgSprites1;
				}
				else
				{
					if (ps->i & 1)
						pimg = _pimgSprites4;
					else
						pimg = _pimgSprites3;
				}

				DrawSprite16x8(pimg->rgbData, pimg->rgbMask,
					n, 64 + (c << 4), x, y, rgbBuffer, lPitch);
#else
				if (ps->c & 0x20)
					pimg = _pimgSprites5;
				else
					pimg = _pimgSprites6;

				DrawSprite(pimg->rgbData, pimg->rgbMask,
					n, 64 + (c << 4), x, y, rgbBuffer, lPitch);
#endif
			}
#endif
		}
	}

	//
	//  Draw character graphics (overlayed transparently)
	//

	for (i = 2 * 32; i < 32 * 30; i++)
	{
		int img = gyruss_video_ram[i] + 8 * (gyruss_color_ram[i] & 0x20);
		
		if (img == 0x83) // skip spaces
			continue;

		img = gyruss_video_ram[i] + 8 * (gyruss_color_ram[i] & 0xE0);

		const int x = 8 * (31 - (i / 32));
		const int y = 8 * (i % 32);

		DrawCharT(_pimgChars->rgbData, _pimgChars->rgbMask,
			img, (gyruss_color_ram[i] & 0x0F) << 2, x, y, rgbBuffer, lPitch);
	}
}

static int gyruss_portA_r()
{
	int clockticks,clock;

	LARGE_INTEGER x, y;
	
	if (QueryPerformanceCounter(&x) && QueryPerformanceFrequency(&y))
	{
		clockticks = (int)(x.QuadPart * 2L * 1572750L / y.QuadPart);
	}

#define TIMER_RATE (570)

//	clockticks = (Z80_IPeriod - cpu_geticount());
	clock = clockticks / TIMER_RATE;

	/* to speed up the emulation, detect when the program is looping waiting */
	/* for the timer, and skip the necessary CPU cycles in that case */
#if 0
	if (cpu_getreturnpc() == 0x0101)
	{
		/* wait until clock & 0x04 == 0 */
		if ((clock & 0x04) != 0)
		{
			clock = clock + 0x04;
			clockticks = clock * TIMER_RATE;
			cpu_seticount(Z80_IPeriod - clockticks);
		}
	}
	else if (cpu_getreturnpc() == 0x0108)
	{
		/* wait until clock & 0x04 != 0 */
		if ((clock & 0x04) == 0)
		{
			clock = (clock + 0x04) & ~0x03;
			clockticks = clock * TIMER_RATE;
			cpu_seticount(Z80_IPeriod - clockticks);
		}
	}
#endif

	return clock;
}

static BYTE GyrussAudioIn(BYTE A)
{
	switch (A)
	{
		case 0x01: // AY8910_read_port_0_r
			return _rgAy8910[0].ReadRegLatch();

		case 0x05: // AY8910_read_port_1_r
			return _rgAy8910[1].ReadRegLatch();

		case 0x09: // AY8910_read_port_2_r
			return _rgAy8910[2].ReadRegLatch();

		case 0x0d: // AY8910_read_port_3_r
			return _rgAy8910[3].ReadRegLatch();

		case 0x11: // AY8910_read_port_4_r
			return _rgAy8910[4].ReadRegLatch();
	}

	return 0;
}

static void GyrussAudioOut(BYTE A, BYTE B)
{
	switch (A)
	{
		case 0x00: // AY8910_control_port_0_w
			_rgAy8910[0].WriteLatch(B);
			break;

		case 0x02: // AY8910_write_port_0_w
			_rgAy8910[0].WriteRegLatch(B);
			break;

		case 0x04: // AY8910_control_port_1_w
			_rgAy8910[1].WriteLatch(B);
			break;

		case 0x06: // AY8910_write_port_1_w
			_rgAy8910[1].WriteRegLatch(B);
			break;

		case 0x08: // AY8910_control_port_2_w
			_rgAy8910[2].WriteLatch(B);
			break;

		case 0x0a: // AY8910_write_port_2_w
			_rgAy8910[2].WriteRegLatch(B);
			break;

		case 0x0c: // AY8910_control_port_3_w
			_rgAy8910[3].WriteLatch(B);
			break;

		case 0x0e: // AY8910_write_port_3_w
			_rgAy8910[3].WriteRegLatch(B);
			break;

		case 0x10: // AY8910_control_port_4_w
			_rgAy8910[4].WriteLatch(B);
			break;

		case 0x12: // AY8910_write_port_4_w
			_rgAy8910[4].WriteRegLatch(B);
			break;

		case 0x14: // gyruss_sh_soundfx_on_w
		{
			if (B)
			{
				B--;
				
				if (B < DIM(_rgsiGyruss))
				{
					SoundSetEnvelope(0, _rgsiGyruss[B].rgbData, _rgsiGyruss[B].cbSize, FALSE);
				}
			}
			else
			{
				SoundSetEnvelope(0, NULL, 0, FALSE);
			}

			break;
		}

		case 0x18: // gyruss_sh_soundfx_data_w
			break;
	}
}

static BYTE M_RDMEMAudio(WORD A)
{
	switch (A)
	{
		case 0x8000: // sound_command_r
			return sound_command_r(A);
			break;

	}
	
	return RAM[A];
}

static void M_WRMEMAudio(WORD A, BYTE V)
{
	if (A < 0x4000)
		return;

	RAM[A] = V;
}

/*---------------------------------------------------------------------------

	Define game descriptors

---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdGyruss = 
{
	TEXT("Gyruss"),
	NULL,
	Init,
	Exec,
	RenderScene,
	UpdateSound,
	Term
};

