/*---------------------------------------------------------------------------

	File:
		invaders.cpp
	
	Abstract:
		Implements the "Space Invaders" game driver functions.

	Revisions:
		16-Feb-97 EAM Created

	Copyright (C) 1997 Edward Massey

	Hardware information adopted Michael Strutt's Space Invaders emulator,
	available at the Repository.

	TODO:  colors

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "game.h"
#include "game1.h"
#include "../z80/execz80.h"

// machine state...
static BYTE _rgbImage[256][256];
static IExecZ80 *_pZ80;
static BYTE _bBitmap, _bBitmapPrevious, _bShiftValue;
static BYTE bSound1 = 0;
static BYTE bSound2 = 0;

static BYTE M_RDMEM(WORD A);
static void M_WRMEM(WORD A, BYTE V);
static BYTE DoIn(BYTE bPort);
static void DoOut(byte Port,byte Value);

extern void SoundUpdateMixer(LPBYTE rgbBuffer, LONG cbBuffer);
extern void SoundSetEnvelope(int iChnl, const char *pData, int cSamples, BOOL fLoop);

static const TCHAR szPath[] = "invaders";

RAGE_ROMTABLE_BEGIN(ROMs0)
	RAGE_ROMTABLE_ENTRY("invaders.h", 0x0000, 0x0800)
	RAGE_ROMTABLE_ENTRY("invaders.g", 0x0800, 0x0800)
	RAGE_ROMTABLE_ENTRY("invaders.f", 0x1000, 0x0800)
	RAGE_ROMTABLE_ENTRY("invaders.e", 0x1800, 0x0800)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(ROMs1)
	RAGE_ROMTABLE_ENTRY("invdelux.h", 0x0000, 0x0800)
	RAGE_ROMTABLE_ENTRY("invdelux.g", 0x0800, 0x0800)
	RAGE_ROMTABLE_ENTRY("invdelux.f", 0x1000, 0x0800)
	RAGE_ROMTABLE_ENTRY("invdelux.e", 0x1800, 0x0800)
	RAGE_ROMTABLE_ENTRY("invdelux.d", 0x4000, 0x0800)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(ROMs2)
	RAGE_ROMTABLE_ENTRY("lrescue.1", 0x0000, 0x0800)
	RAGE_ROMTABLE_ENTRY("lrescue.2", 0x0800, 0x0800)
	RAGE_ROMTABLE_ENTRY("lrescue.3", 0x1000, 0x0800)
	RAGE_ROMTABLE_ENTRY("lrescue.4", 0x1800, 0x0800)
	RAGE_ROMTABLE_ENTRY("lrescue.5", 0x4000, 0x0800)
	RAGE_ROMTABLE_ENTRY("lrescue.6", 0x4800, 0x0800)
RAGE_ROMTABLE_END

static const SIZE _sizBuffer = { 256, 256 };
static const RECT _rcDisplay = { 0, 0, 256 - 32, 256 };

static const EXECZ80_MEM_DESCRIPTOR rgmd[] = 
{
	{ 0x0000, 0xFFFF, 0, NULL, NULL },
	{ 0x2400, 0x43FF, 0, NULL, M_WRMEM },
	{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
};

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

enum 
{
    black, red, purple, cyan, yellow, blue, green, white
};

static const int COLOR_CODES = 16; // number of color schemes

static const PALETTEENTRY _rgpe[] =
{
	{ 0x00, 0x00, 0x00 }, // black
	{ 0xFF, 0x00, 0x00 }, // red
	{ 0xB8, 0x02, 0xB8 }, // purple
	{ 0x00, 0xFF, 0xFF }, // cyan
	{ 0xFF, 0xFF, 0x00 }, // yellow
	{ 0x24, 0x24, 0xFF }, // blue
	{ 0x00, 0x90, 0x00 }, // green
	{ 0xFF, 0xFF, 0xFF }, // white
};

static const BYTE _rgbColorTable[4 * COLOR_CODES] =
{
    black, white, white, white, /* background, white text */
    black, red, blue, yellow,   /* flag ship */
    black, blue, red, yellow,   /* red bug */
    black, blue, purple, red,   /* purple bug */
    black, blue, green, red,    /* green bug */
    black, red, red, red,       /* red text */
    black, white, red, cyan,    /* player ship, cyan text */
    black, yellow, red, purple, /* explosion, purple text */
};

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

struct SAMPLE_INFO
{
	LPCTSTR pszFilename;
	int cbSize;
	char *rgbData;
};

static SAMPLE_INFO _rgsiInvaders[] = 
{
	{ "0.raw" },
	{ "1.raw" },
	{ "2.raw" },
	{ "3.raw" },
	{ "4.raw" },
	{ "5.raw" },
	{ "6.raw" },
	{ "7.raw" },
	{ "8.raw" },
};

static char _rgbSamples[9][32 * 1024];

static int LoadSamples()
{
	int i, j;

	for (i = 0; i < DIM(_rgsiInvaders); i++)
	{
		_rgsiInvaders[i].rgbData = _rgbSamples[i];

		TStr<_MAX_PATH> strFile("invaders/%s", _rgsiInvaders[i].pszFilename);

		FILE *pf = fopen(strFile, "rb");

		if (pf != NULL)
		{
			_rgsiInvaders[i].cbSize = fread(_rgsiInvaders[i].rgbData, 1, sizeof(_rgbSamples[0]), pf);

			fclose(pf);
		}
	}

	return 0;
}

static HRESULT Init(GAME_LOAD_INFO &gli)
{
	LoadSamples();

	ZERO_INIT(_rgbImage);

	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	BuildCompositePalette(_rgpe, DIM(_rgpe), _rgbColorTable,
		COLOR_CODES * 4, gli.ppe, 256, gli.cpe);

	if (readroms(RAM, (RomModule*)gli.pvContext, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	EXECZ80_CREATE cs;

	ZERO_INIT(cs);

	cs.rgbMemory    = RAM;
	cs.pfnMemRead   = M_RDMEM;
	cs.pfnMemWrite  = M_WRMEM;
	cs.pfnPortRead  = DoIn;
	cs.pfnPortWrite = DoOut;
	cs.pmd          = rgmd;

	HRESULT hr = EclCreateZ80(cs, _pZ80);

	if (FAILED(hr))
		return hr;

	_pZ80->Reset();

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE M_RDMEM(WORD A)
{
	return RAM[A];
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void M_WRMEM(WORD A, BYTE V)
{
	if (A >= 0x2400 && A <= 0x2400 + (256 * 256 / 8))
	{
		if (RAM[A] == V)
			return;

		RAM[A] = V;

		int offset = A - 0x2400;
		int x = offset / 32;
		int y = 256 - 8 * (offset % 32);

		_rgbImage[y--][x] = ((V & 0x01) != 0);
		_rgbImage[y--][x] = ((V & 0x02) != 0);
		_rgbImage[y--][x] = ((V & 0x04) != 0);
		_rgbImage[y--][x] = ((V & 0x08) != 0);
		_rgbImage[y--][x] = ((V & 0x10) != 0);
		_rgbImage[y--][x] = ((V & 0x20) != 0);
		_rgbImage[y--][x] = ((V & 0x40) != 0);
		_rgbImage[y--][x] = ((V & 0x80) != 0);
	}

	RAM[A] = V;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE DoIn(BYTE bPort)
{
	switch(bPort)
	{
		case 0x01:
		{
			BYTE bVal = 0x81;

			if (osd_key_pressed(OSD_KEY_3))
				bVal ^= 0x01;

			if (osd_key_pressed(OSD_KEY_2))
				bVal ^= 0x02;

			if (osd_key_pressed(OSD_KEY_1))
				bVal ^= 0x04;

			if (osd_key_pressed(OSD_KEY_CONTROL))
				bVal ^= 0x10;

			if (osd_key_pressed(OSD_KEY_LEFT))
				bVal ^= 0x20;

			if (osd_key_pressed(OSD_KEY_RIGHT))
				bVal ^= 0x40;
		
			return bVal;
		}
		
		case 0x02:
		{
			BYTE bVal = 0x00;

			if (osd_key_pressed(OSD_KEY_T))
				bVal ^= 0x04;

			if (osd_key_pressed(OSD_KEY_P))
				bVal ^= 0x08;

			if (osd_key_pressed(OSD_KEY_LEFT))
				bVal ^= 0x20;

			if (osd_key_pressed(OSD_KEY_RIGHT))
				bVal ^= 0x40;
			
			if (osd_key_pressed(OSD_KEY_CONTROL))
				bVal ^= 0x10;
			
			return bVal;
		}

		case 0x03:
		{
			WORD wValue;
			
			wValue = _bBitmap << 8;
			wValue |= _bBitmapPrevious;
			wValue <<= _bShiftValue;
			
			return (BYTE)(wValue >> 8);
		}

		default:
			return(0x00);
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static inline void InvaderSound(int ch, int i, BOOL fLoop)
{
	if (i == -1)
		SoundSetEnvelope(ch, NULL, 0, FALSE);
	else
		SoundSetEnvelope(ch, _rgsiInvaders[i].rgbData, _rgsiInvaders[i].cbSize, fLoop);
}

struct INVADER_SOUND
{
	int nChannel;
	int i; // 	int cbSize; char *rgbData;
	BOOL fLoop;
	BYTE bMask;
};

static const INVADER_SOUND _rgSound3[] = 
{
	{ 0, 0, TRUE,  0x01 },
	{ 1, 1, FALSE, 0x02 },
	{ 2, 2, FALSE, 0x04 },
	{ 3, 3, FALSE, 0x08 },
};

static const INVADER_SOUND _rgSound5[] = 
{
	{ 4, 4, FALSE, 0x01 },
	{ 4, 5, FALSE, 0x02 },
	{ 4, 6, FALSE, 0x04 },
	{ 4, 7, FALSE, 0x08 },
	{ 5, 8, FALSE, 0x10 },
};

static void DoOut(byte Port,byte Value)
{
	switch(Port) {
		case 2:
			_bShiftValue=Value;
			break;
		case 3:
			if(Value&0x01 && ~bSound1&0x01) InvaderSound(0, 0, TRUE);
			if(~Value&0x01 && bSound1&0x01) InvaderSound(0, -1, FALSE);
			if(Value&0x02 && ~bSound1&0x02) InvaderSound(1, 1, FALSE);
			if(~Value&0x02 && bSound1&0x02) InvaderSound(1, -1, FALSE);
			if(Value&0x04 && ~bSound1&0x04) InvaderSound(2, 2, FALSE);
			if(~Value&0x04 && bSound1&0x04) InvaderSound(2, -1, FALSE);
			if(Value&0x08 && ~bSound1&0x08) InvaderSound(3, 3, FALSE);
			if(~Value&0x08 && bSound1&0x08) InvaderSound(3, -1, FALSE);

			bSound1=Value;
			break;
		case 4:
			_bBitmapPrevious=_bBitmap;
			_bBitmap=Value;
			break;
		case 5:
			if(Value&0x01 && ~bSound2&0x01) InvaderSound(4, 4, FALSE);
			if(Value&0x02 && ~bSound2&0x02) InvaderSound(4, 5, FALSE);
			if(Value&0x04 && ~bSound2&0x04) InvaderSound(4, 6, FALSE);
			if(Value&0x08 && ~bSound2&0x08) InvaderSound(4, 7, FALSE);
			if(Value&0x10 && ~bSound2&0x10) InvaderSound(5, 8, FALSE);
			if(~Value&0x10 && bSound2&0x10) InvaderSound(5, -1, FALSE);
			bSound2=Value;
			break;
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void UpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
	SoundUpdateMixer(rgbBuffer, cbBuffer);
}

/*---------------------------------------------------------------------------

	Exec

	Execute one frame of emulation.  Space Invaders uses an 8080 so we have
	to handle IRQ and NMI slightly different than a Z80.

	Note that I'm making two passes of execution here.

---------------------------------------------------------------------------*/

static int Exec()
{
	int cCycles = 0;
	
	cCycles += _pZ80->Exec(2000000 / 60);
	cCycles += _pZ80->Irq(0xCF); // NOTE:  opcode for rst ???
	cCycles += _pZ80->Exec(2000000 / 60);
	cCycles += _pZ80->Irq(0xD7); // NOTE:  opcode for rst ???

#if 0
//	static int bNMI;

	for (int i = 0; i < 2; i++)
	{
		cCycles += _pZ80->Exec(2000000 / 60);

//		EXECZ80_CONTEXT ctx;
		
//		_pZ80->GetContext(ctx);

#if 1
//		ctx.IFF1 = 1;

//		_pZ80->SetContext(ctx);
		
		if(i & 1)
			cCycles += _pZ80->Irq(0xD7); // NOTE:  opcode for rst ???
		else	
			cCycles += _pZ80->Irq(0xCF); // NOTE:  opcode for rst ???
#else
//		if (ctx.IFF1)
		{
			WORD wAddr = (i & 1) ? 0x10 : 0x08;
			
			RAM[--ctx.SP] = HIBYTE(wAddr);
			RAM[--ctx.SP] = LOBYTE(wAddr);
			ctx.PC = wAddr;

			_pZ80->SetContext(ctx);
		}
#endif
	}
#endif

#ifdef _DEBUG
	if (GetAsyncKeyState('S') < 0)
	{
		for (int n = 0; n <= 8; n++)
		{
			if (GetAsyncKeyState('0' + n) < 0)
				InvaderSound(0, n, FALSE);
		}
	}
#endif

	return cCycles;
}

static void Term()
{
	if (_pZ80 != NULL)
		_pZ80->Release();
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void RenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	for (int i = 0; i < 256; i++)
		memcpy(&rgbBuffer[lPitch * i], &_rgbImage[i][0], 256);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdInvaders = 
{
	TEXT("Space Invaders"),
	(LPVOID)ROMs0,
	Init,
	Exec,
	RenderScene,
	UpdateSound,
	Term
};

GAME_DESCRIPTOR _gmdInvadersDeluxe = 
{
	TEXT("Space Invaders Deluxe"),
	(LPVOID)ROMs1,
	Init,
	Exec,
	RenderScene,
	UpdateSound,
	Term
};

GAME_DESCRIPTOR _gmdLunarRescue = 
{
	TEXT("Lunar Rescue"),
	(LPVOID)ROMs2,
	Init,
	Exec,
	RenderScene,
	UpdateSound,
	Term
};

