/*---------------------------------------------------------------------------
	
	File:
		pacman.cpp
	
	Abstract:
		Implements the "Pacman" game driver functions.

	Revisions:
		05-Feb-97 EAM Created
		03-May-97 EAM Moved to IExecZ80 and added sound support

	Copyright (C) 1997 Edward Massey

	http://www.greenvillenc.com/home/edward

	Hardware information adopted from MAME, by Nicola Salmoria.

	BUG:  sprite transparency isn't quite correct
	CONSIDER:  update a cached version of sound reg's on memory write, use
	      these precomputed values in UpdateSound
	TODO:  optimizations (eliminate _rgCharMap)
	BUG:  sometimes, Ms. Pacman game starts out w/blue characters
	TODO:  maintain sound channel position for smoother sound

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "pacman.h"
#include "game.h"
#include "game1.h"
#include "../z80/execz80.h"

void PacmanUpdateSound(LPBYTE rgbBuffer, LONG cbBuffer, const BYTE *RAM,
	BOOL &fSoundChanged);

static BYTE M_RDMEM(WORD A);
static void M_WRMEM(WORD A, BYTE);
static void PacmanOut(BYTE A, BYTE V);

//
//  Machine state...
//

static int _bInterruptVector = 0;
static BYTE _rgbDSW[] = { 0xE9, 0x00 };
static IExecZ80 *_pZ80;
static IMAGE *_pimgChars;
static IMAGE *_pimgSprites;
static const BYTE &_fInterruptEnable = RAM[INTERRUPT_ENABLE];
static const int _cCyclesPerFrame = 50000;
static BOOL _fSoundChanged = TRUE; // NOTE:  force calculation first time

/*---------------------------------------------------------------------------

	Static data

	from MAME...
	
	Convert the color PROMs into a more useable format.

	Pac Man has a 16 bytes palette PROM and a 128 bytes color lookup table PROM.
	The palette PROM is connected to the RGB output this way:

	bit 7 -- 220 ohm resistor  -- BLUE
		-- 470 ohm resistor  -- BLUE
		-- 220 ohm resistor  -- GREEN
		-- 470 ohm resistor  -- GREEN
		-- 1  kohm resistor  -- GREEN
		-- 220 ohm resistor  -- RED
		-- 470 ohm resistor  -- RED
	bit 0 -- 1  kohm resistor  -- RED

---------------------------------------------------------------------------*/

static const BYTE _rgbColorPROM[] =
{
	/* palette */
	0x00,0x07,0x66,0xEF,0x00,0xF8,0xEA,0x6F,0x00,0x3F,0x00,0xC9,0x38,0xAA,0xAF,0xF6,

	/* color lookup table */
	0x00,0x00,0x00,0x00,0x00,0x0F,0x0B,0x01,0x00,0x00,0x00,0x00,0x00,0x0F,0x0B,0x03,
	0x00,0x00,0x00,0x00,0x00,0x0F,0x0B,0x05,0x00,0x00,0x00,0x00,0x00,0x0F,0x0B,0x07,
	0x00,0x00,0x00,0x00,0x00,0x0B,0x01,0x09,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0F,0x00,0x0E,0x00,0x01,0x0C,0x0F,
	0x00,0x0E,0x00,0x0B,0x00,0x0C,0x0B,0x0E,0x00,0x0C,0x0F,0x01,0x00,0x00,0x00,0x00,
	0x00,0x01,0x02,0x0F,0x00,0x07,0x0C,0x02,0x00,0x09,0x06,0x0F,0x00,0x0D,0x0C,0x0F,
	0x00,0x05,0x03,0x09,0x00,0x0F,0x0B,0x00,0x00,0x0E,0x00,0x0B,0x00,0x0E,0x00,0x0B,
	0x00,0x00,0x00,0x00,0x00,0x0F,0x0E,0x01,0x00,0x0F,0x0B,0x0E,0x00,0x0E,0x00,0x0F
};

static const TCHAR _szPath[] = "pacman";

RAGE_ROMTABLE_BEGIN(_rgRoms)
	RAGE_ROMTABLE_ENTRY("pacman.6e", 0x0000, 0x1000)
	RAGE_ROMTABLE_ENTRY("pacman.6f", 0x1000, 0x1000)
	RAGE_ROMTABLE_ENTRY("pacman.6h", 0x2000, 0x1000)
	RAGE_ROMTABLE_ENTRY("pacman.6j", 0x3000, 0x1000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(_rgGfxRoms)
	RAGE_ROMTABLE_ENTRY("pacman.5e", 0x0000, 0x1000) // characters
	RAGE_ROMTABLE_ENTRY("pacman.5f", 0x1000, 0x1000) // sprites
RAGE_ROMTABLE_END

static const TCHAR _szPathMs[] = "mspacman";

RAGE_ROMTABLE_BEGIN(_rgRomsMs)
	RAGE_ROMTABLE_ENTRY("boot1", 0x00000, 0x1000)
	RAGE_ROMTABLE_ENTRY("boot2", 0x01000, 0x1000)
	RAGE_ROMTABLE_ENTRY("boot3", 0x02000, 0x1000)
	RAGE_ROMTABLE_ENTRY("boot4", 0x03000, 0x1000)
	RAGE_ROMTABLE_ENTRY("boot5", 0x08000, 0x1000)
	RAGE_ROMTABLE_ENTRY("boot6", 0x09000, 0x1000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(_rgGfxRomsMs)
	RAGE_ROMTABLE_ENTRY("5e", 0x0000, 0x1000) // characters
	RAGE_ROMTABLE_ENTRY("5f", 0x1000, 0x1000) // sprites
RAGE_ROMTABLE_END

static const DECODE_GRAPHICS dgsChars =
{
	0x00000000,
	{
		{ 8, 8 },
		256,
		16 * 8,
		2,
		{ 0, 4 },
		2,
		{ 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 }, /* characters are rotated 90 degrees */
		{ 8*8+0, 8*8+1, 8*8+2, 8*8+3, 0, 1, 2, 3 },	/* bits are packed in groups of four */
	},

	1,
	
	1, 
	
	{
		0
	},
	0, 
	NULL
};

static const DECODE_GRAPHICS dgsSprites =
{
	0x00000001,
	{
		{ 16, 16 },	/* 16*16 sprites */
		64,	/* 64 sprites */
		64*8,	/* every sprite takes 64 bytes */
		2,	/* 2 bits per pixel */
		{ 0, 4 },	/* the two bitplanes for 4 pixels are packed into one byte */
		2,	/* 2 bits per pixel */
		{ 39 * 8, 38 * 8, 37 * 8, 36 * 8, 35 * 8, 34 * 8, 33 * 8, 32 * 8,
				7 * 8, 6 * 8, 5 * 8, 4 * 8, 3 * 8, 2 * 8, 1 * 8, 0 * 8 },
		{ 8*8, 8*8+1, 8*8+2, 8*8+3, 16*8+0, 16*8+1, 16*8+2, 16*8+3,
				24*8+0, 24*8+1, 24*8+2, 24*8+3, 0, 1, 2, 3 }
	},
	4,
	4, 
	{
		0, 1, 2, 3
	},
	0, 
	NULL
};

static const SIZE _sizBuffer = { 256 + 16, 256 + 16 + 16};
static const RECT _rcDisplay = { 16, 0, 28 * 8 + 16, 256 + 32 };

static const EXECZ80_MEM_DESCRIPTOR _rgmd[] = 
{
#if 0
	{ 0x0000, 0xFFFF, 0, M_RDMEM, M_WRMEM },
#else
	{ 0x0000, 0xFFFF, 0, NULL, NULL },
	{ 0x8000, 0xFFFF, 0, NULL, M_WRMEM },
	{ IN0_PORT, IN0_PORT, 0, M_RDMEM, M_WRMEM },
	{ IN1_PORT, IN1_PORT, 0, M_RDMEM, M_WRMEM },
	{ DSW1_PORT, DSW1_PORT, 0, M_RDMEM, M_WRMEM },
	{ 0x5040, 0x5060, 0, M_RDMEM, M_WRMEM },
#endif
	{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
};

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static POINT _rgCharMap[32 * 32];

static void InitCharMap()
{
	for (int i = 0; i < 32 * 32; i++)
	{
		int x = 31 - (i / 32);
		int y = i % 32;

		_rgCharMap[i].x = 8 * x;
		_rgCharMap[i].y = 8 * y;

		if (x < 2)
		{
			if (y < 28)
			{
				_rgCharMap[i].x = 16 + 8 * (27 - y);
				_rgCharMap[i].y = 8 * (1 - x);
			}
		}
		else if (x >= 30)
		{
			if (y < 28)
			{
				_rgCharMap[i].x = 16 + 8 * (27 - y);
				_rgCharMap[i].y = 16 + 8 * (63 - x);
			}
		}
		else
		{
			_rgCharMap[i].x = 8 * x;
			_rgCharMap[i].y = 16 + 8 * y;
		}
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void BuildCompositePalette(PALETTEENTRY *ppeOut, UINT &cpeOut)
{
	for (cpeOut = 0; cpeOut < DIM(_rgbColorPROM) - 16; cpeOut++)
	{
		int i = _rgbColorPROM[cpeOut + 16];

		int bit0,bit1,bit2;

		bit0 = (_rgbColorPROM[i] >> 0) & 0x01;
		bit1 = (_rgbColorPROM[i] >> 1) & 0x01;
		bit2 = (_rgbColorPROM[i] >> 2) & 0x01;

		ppeOut[cpeOut].peRed = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;

		bit0 = (_rgbColorPROM[i] >> 3) & 0x01;
		bit1 = (_rgbColorPROM[i] >> 4) & 0x01;
		bit2 = (_rgbColorPROM[i] >> 5) & 0x01;

		ppeOut[cpeOut].peGreen = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;

		bit0 = 0;
		bit1 = (_rgbColorPROM[i] >> 6) & 0x01;
		bit2 = (_rgbColorPROM[i] >> 7) & 0x01;

		ppeOut[cpeOut].peBlue  = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;
	}
}

HRESULT PacmanInit(GAME_LOAD_INFO &gli)
{
	InitCharMap();

	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	BuildCompositePalette(gli.ppe, gli.cpe);

	BYTE tmpstorage[0x2000];

	if (gli.pvContext == 0x00000000)
	{
		if (readroms(RAM, _rgRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
			return E_FAIL;

		if (readroms(tmpstorage, _rgGfxRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
			return E_FAIL;
	}
	else
	{
		if (readroms(RAM, _rgRomsMs, _szPathMs, gli.pszMsg, gli.cchMsgMax) != 0)
			return E_FAIL;

		if (readroms(tmpstorage, _rgGfxRomsMs, _szPathMs, gli.pszMsg, gli.cchMsgMax) != 0)
			return E_FAIL;
	}

	//
	//
	//

	MyGfxDecode3(dgsChars,   tmpstorage,          _pimgChars);
	MyGfxDecode3(dgsSprites, tmpstorage + 0x1000, _pimgSprites);

	//
	//
	//
	
	EXECZ80_CREATE cs;

	ZERO_INIT(cs);

	cs.rgbMemory    = RAM;
	cs.pfnMemRead   = M_RDMEM;
	cs.pfnMemWrite  = M_WRMEM;
	cs.pfnPortWrite = PacmanOut;
	cs.pmd          = _rgmd;
	
	HRESULT hr = EclCreateZ80(cs, _pZ80);

	if (FAILED(hr))
		return hr;

	return S_OK;
}

void PacmanTerm()
{
	if (_pZ80 != NULL)
		_pZ80->Release();

	FreeImage(_pimgChars);
	FreeImage(_pimgSprites);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE M_RDMEM(WORD A)
{
	switch (A)
	{
		case IN0_PORT:
		{
			BYTE b = 0xff;

			if (osd_key_pressed(OSD_KEY_3))
				b ^= IN0_CREDIT;

			if (osd_key_pressed(OSD_KEY_F1))
				b ^= IN0_RACK_TEST;

			if (osd_key_pressed(OSD_KEY_DOWN))
				b ^= IN0_DOWN;

			if (osd_key_pressed(OSD_KEY_RIGHT))
				b ^= IN0_RIGHT;

			if (osd_key_pressed(OSD_KEY_LEFT))
				b ^= IN0_LEFT;

			if (osd_key_pressed(OSD_KEY_UP))
				b ^= IN0_UP;

			return b;
		}

		case IN1_PORT:
		{
			BYTE b = 0xff;

			if (osd_key_pressed(OSD_KEY_2))
				b ^= IN1_START2;

			if (osd_key_pressed(OSD_KEY_1))
				b ^= IN1_START1;

			if (osd_key_pressed(OSD_KEY_F2))
				b ^= IN1_TEST;

			return b;
		}

		case DSW1_PORT:
			return _rgbDSW[0];
	}

	return RAM[A];
}

static BYTE MspacmanM_RDMEM(WORD A)
{
	switch (A)
	{
		case IN0_PORT:
		{
			BYTE b = 0xff;

			if (osd_key_pressed(OSD_KEY_3))
				b ^= IN0_CREDIT;

			if (osd_key_pressed(OSD_KEY_F1))
				b ^= IN0_RACK_TEST;

			if (osd_key_pressed(OSD_KEY_DOWN))
				b ^= IN0_DOWN;

			if (osd_key_pressed(OSD_KEY_RIGHT))
				b ^= IN0_RIGHT;

			if (osd_key_pressed(OSD_KEY_LEFT))
				b ^= IN0_LEFT;

			if (osd_key_pressed(OSD_KEY_UP))
				b ^= IN0_UP;

			return b;
		}

		case IN1_PORT:
		{
			BYTE b = 0xff;

			if (osd_key_pressed(OSD_KEY_2))
				b ^= IN1_START2;

			if (osd_key_pressed(OSD_KEY_1))
				b ^= IN1_START1;

			if (osd_key_pressed(OSD_KEY_F2))
				b ^= IN1_TEST;

			return b;
		}

		case DSW1_PORT:
			return _rgbDSW[0];
	}

	return RAM[A];
}

static void M_WRMEM(WORD A, BYTE V)
{
	A &= 0x7fff;	/* It looks like the Pac Man arcade machine has a 15 bit */
					/* path to RAM: we must clear the high bit otherwise CREDIT */
					/* and HIGHSCORE will not be displayed on the screen. */

	if (A <= ROM_END)
		return;

	if (A >= SOUND_BASE && A < SOUND_BASE + 0x20)
	{
		V &= 0x0F;
		_fSoundChanged = TRUE;
	}

	RAM[A] = V;
}

static void PacmanOut(BYTE bPort, BYTE bValue)
{
	switch (bPort)
	{
		case 0:
			_bInterruptVector = bValue;
			break;

		default:
			TRACE(__FILE__ ": unknown port write (%02X,%02X)\n", bPort, bValue);
			break;
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

int PacmanExec()
{
	int cCycles;
	
	cCycles = _pZ80->Exec(_cCyclesPerFrame);
	
	if (_fInterruptEnable)
		cCycles += _pZ80->Irq(_bInterruptVector);

	return cCycles;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

void PacmanRenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	int i;

	//
	//  Draw 32x32 characters
	//

	// CONSIDER:  might want to touch host video ram sequentially

	for (i = 0; i < 32 * 32; i++)
	{
		const int offset_value = RAM[VIDEO_RAM_START + i];
		const int color = RAM[COLOR_RAM_START + i] & 0x1F;
		const int x = _rgCharMap[i].x;
		const int y = _rgCharMap[i].y;
//		const int x = 8 * (31 - (i / 32));
//		const int y = 8 * (i % 32);


		_pimgChars->DrawChar(rgbBuffer, lPitch, x, y,
			offset_value, color << 2);
	}

	//
	//  Draw the 8 sprites
	//
	
	for (i = 7; i >= 0; i--)
	{
		if (RAM[SPRITE_COORDS_BASE + 2*i] == 0)
			continue;

		if (i <= 2)
		{
			_pimgSprites->DrawSprite(rgbBuffer, lPitch,
				254 - RAM[SPRITE_COORDS_BASE + 2*i],
				17 + 255 - RAM[SPRITE_COORDS_BASE+1 + 2*i],
				RAM[SPRITE_CODES_BASE + 2*i],
				RAM[SPRITE_CODES_BASE+1 + 2*i] << 2);
		}
		else
		{
			_pimgSprites->DrawSprite(rgbBuffer, lPitch,
				255 - RAM[SPRITE_COORDS_BASE + 2*i],
				17 + 255 - RAM[SPRITE_COORDS_BASE+1 + 2*i],
				RAM[SPRITE_CODES_BASE + 2*i],
				RAM[SPRITE_CODES_BASE+1 + 2*i] << 2);
		}
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

void PacmanUpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
	if (RAM[SOUND_ENABLE])
	{
		PacmanUpdateSound(rgbBuffer, cbBuffer, &RAM[SOUND_BASE],
			_fSoundChanged);
	}
	else
	{
		while (cbBuffer--)
			*rgbBuffer++ = 128;
	}
}

/*---------------------------------------------------------------------------

	Define game descriptors

---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdPacman = 
{
	TEXT("Pacman"),
	(LPVOID)0x00000000,
	PacmanInit,
	PacmanExec,
	PacmanRenderScene,
	PacmanUpdateSound,
	PacmanTerm
};

GAME_DESCRIPTOR _gmdMsPacman = 
{
	TEXT("Ms. Pacman"),
	(LPVOID)0x00000001,
	PacmanInit,
	PacmanExec,
	PacmanRenderScene,
	PacmanUpdateSound,
	PacmanTerm
};

