/*---------------------------------------------------------------------------
	
	File:
		pengo.cpp
	
	Abstract:
		Implements the "Pengo" game driver functions.  Emulates the game's
		hardware (except CPU).

	Revisions:
		10-Feb-97 EAM Created

	Copyright (C) 1997 Edward Massey

---------------------------------------------------------------------------*/

#include "pch.h"
#include <stdio.h>
#include "misc.h"
#include "../inc/z80api.h"
#include "pengo.h"
#include "game.h"
#include "game1.h"

#define CPU_SPEED 3072000	/* 3.072 Mhz. Is this correct for Pac Man? */
#define FRAMES_PER_SECOND 60

#define SND_CLOCK 3072000	/* 3.072 Mhz */

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static const PALETTEENTRY _rgpe[] =
{
/* first bank */
	{ 0x00, 0x00, 0x00 }, // BLACK
	{ 0xdb, 0xdb, 0xdb }, // WHITE
	{ 0xff, 0x00, 0x00 }, // RED
	{ 0x00, 0xff, 0x00 }, // GREEN
	{ 0x24, 0x24, 0xdb }, // BLUE
	{ 0x00, 0xff, 0xdb }, // CYAN,
	{ 0xff, 0xff, 0x00 }, // YELLOW,
	{ 0xff, 0xb6, 0xdb }, // PINK
	{ 0xff, 0xb6, 0x49 }, // ORANGE
	{ 0xdb, 0x24, 0x00 }, // DKRED
	{ 0xff, 0xb6, 0x00 }, // DKORANGE
	{ 0xff, 0xff, 0x49 }, // YELLOW2
	{ 0x00, 0xdb, 0xdb }, // DKCYAN
	{ 0xdb, 0xdb, 0x00 }, // DKYELLOW
	{ 0x6d, 0x6d, 0xdb }, // BLUISH
	{ 0xdb, 0x00, 0xdb }, // PURPLE

/* second bank */
	{ 0x00, 0x00, 0x00 },
	{ 0xdb, 0xdb, 0xdb },
	{ 0x00, 0x6d, 0xdb },
	{ 0x00, 0xdb, 0xdb },
	{ 0x00, 0xff, 0xdb },
	{ 0xdb, 0x24, 0x00 },
	{ 0xff, 0x00, 0x00 },
	{ 0xff, 0xb6, 0x00 },
	{ 0xdb, 0xdb, 0x00 },
	{ 0xff, 0xff, 0x00 },
	{ 0xff, 0xff, 0x49 },
	{ 0x00, 0xb6, 0x00 },
	{ 0x24, 0xdb, 0x00 },
	{ 0x00, 0xff, 0x00 },
	{ 0xff, 0xb6, 0xdb },
	{ 0xdb ,0x00 ,0xdb },
};

enum {BLACK,WHITE,RED,GREEN,BLUE,CYAN,YELLOW,PINK,ORANGE,DKRED,DKORANGE,
		YELLOW2,DKCYAN,DKYELLOW,BLUISH,PURPLE};

#define UNUSED BLACK

/* The color of each character/sprite is determined by a 5 bit value, which */
/* selects the entry to use in this lookup table, and by the 2 bits of the */
/* chracter/sprite data. Additionally, bit 7 of the byte stored in Color RAM */
/* seems to be set for text characters; the reason is yet unknown. */

#define COLOR_CODES 64

static const BYTE _rgbColorTable[4 * COLOR_CODES] =
{
	/* first bank */
	BLACK,BLACK,BLACK,BLACK,
	BLACK,CYAN,GREEN,WHITE,
	BLACK,CYAN,RED,WHITE,
	BLACK,CYAN,YELLOW,WHITE,	/* dancing penguin #6 */
	BLACK,CYAN,PINK,WHITE,		/* dancing penguin #5 */
	BLACK,CYAN,DKORANGE,WHITE,	/* dancing penguin #4 */
	BLACK,CYAN,YELLOW2,WHITE,	/* dancing penguin #3 */
	BLACK,CYAN,DKCYAN,WHITE,	/* dancing penguin #2 */
	BLACK,CYAN,DKYELLOW,WHITE,	/* dancing penguin #1 */
	BLACK,CYAN,BLUE,WHITE,		/* ice cubes */
	BLACK,GREEN,YELLOW,WHITE,	/* title sno-bee 3 */
	BLACK,GREEN,RED,WHITE,		/* title sno-bee 1 */
	BLACK,GREEN,PINK,WHITE,		/* title sno-bee 4 */
	BLACK,GREEN,CYAN,WHITE,		/* title sno-bee 2 */
	BLACK,RED,GREEN,WHITE,		/* title */
	UNUSED,UNUSED,UNUSED,UNUSED,
	BLACK,ORANGE,GREEN,WHITE,
	BLACK,DKRED,RED,CYAN,
	BLACK,ORANGE,CYAN,DKYELLOW,	/* desk (intermission) */
	BLUE,BLUE,BLUE,BLUE,
	UNUSED,UNUSED,UNUSED,UNUSED,
	UNUSED,UNUSED,UNUSED,UNUSED,
	BLACK,RED,RED,RED,
	BLACK,GREEN,GREEN,GREEN,
	BLACK,YELLOW,YELLOW,YELLOW,
	BLACK,PINK,PINK,PINK,
	BLACK,DKORANGE,DKORANGE,DKORANGE,
	BLACK,YELLOW2,YELLOW2,YELLOW2,
	BLACK,WHITE,WHITE,WHITE,
	BLACK,CYAN,CYAN,CYAN,
	ORANGE,DKRED,DKORANGE,YELLOW2,
	DKCYAN,DKYELLOW,BLUISH,PURPLE,

	/* second bank */
	0x10,0x10,0x10,0x10,
	0x10,0x13,0x17,0x1D,
	0x10,0x1C,0x1F,0x1B,
	0x10,0x1C,0x1E,0x1B,
	0x10,0x1C,0x16,0x1B,
	0x10,0x1C,0x17,0x1B,
	0x10,0x1C,0x13,0x1B,
	0x10,0x1C,0x18,0x1B,
	0x10,0x1C,0x1D,0x1B,
	0x10,0x1C,0x14,0x1B,
	0x10,0x1C,0x19,0x1B,
	0x10,0x1C,0x15,0x1B,
	0x10,0x1C,0x12,0x1B,
	0x10,0x1C,0x1B,0x12,
	0x10,0x18,0x1C,0x12,
	0x10,0x18,0x1F,0x12,
	0x10,0x13,0x12,0x11,
	0x10,0x12,0x1F,0x13,
	0x10,0x1F,0x1E,0x12,
	0x10,0x1E,0x17,0x1F,
	0x10,0x17,0x16,0x1E,
	0x10,0x16,0x15,0x17,
	0x10,0x15,0x10,0x16,
	0x10,0x10,0x1B,0x15,
	0x10,0x1B,0x1C,0x10,
	0x10,0x1C,0x1D,0x1B,
	0x10,0x1D,0x18,0x1C,
	0x10,0x18,0x19,0x1D,
	0x10,0x19,0x1A,0x18,
	0x10,0x1A,0x11,0x19,
	0x10,0x11,0x14,0x1A,
	0x10,0x14,0x13,0x11
};

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

RAGE_ROMTABLE_BEGIN(_rgRoms)
	RAGE_ROMTABLE_ENTRY("pengo.u8",  0x0000, 0x1000)
	RAGE_ROMTABLE_ENTRY("pengo.u7",  0x1000, 0x1000)
	RAGE_ROMTABLE_ENTRY("pengo.u15", 0x2000, 0x1000)
	RAGE_ROMTABLE_ENTRY("pengo.u14", 0x3000, 0x1000)
	RAGE_ROMTABLE_ENTRY("pengo.u21", 0x4000, 0x1000)
	RAGE_ROMTABLE_ENTRY("pengo.u20", 0x5000, 0x1000)
	RAGE_ROMTABLE_ENTRY("pengo.u32", 0x6000, 0x1000)
	RAGE_ROMTABLE_ENTRY("pengo.u31", 0x7000, 0x1000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(_rgGfxRoms)
	RAGE_ROMTABLE_ENTRY("pengo.u92", 0x0000, 0x2000)
	RAGE_ROMTABLE_ENTRY("pengo.105", 0x2000, 0x2000)
RAGE_ROMTABLE_END

static BYTE _rgbDSW[] = { 0xB0 };

static unsigned char _rgbChars[256][8][8];
static unsigned char _rgbSprites[256][16][16];

static const TCHAR _szPath[] = "pengo";

static HRESULT Init(GAME_LOAD_INFO &gli)
{
	static const SIZE _sizBuffer = { 256 + 16, 256 + 16};
	static const RECT _rcDisplay = { 0, 0, 256, 256 };

	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	BuildCompositePalette(_rgpe, DIM(_rgpe), _rgbColorTable,
		COLOR_CODES * 4, gli.ppe, 256, gli.cpe);

	unsigned char tmpstorage[0x4000];

	if (readroms(RAM, _rgRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	if (readroms(tmpstorage, _rgGfxRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//
	//
	
	static const MY_GFX_DECODE mgdChar = {
		{ 8, 8 },
		256,
		16 * 8,
		2,
		4,
		2,
		{ 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 }, /* characters are rotated 90 degrees */
		{ 8*8+0, 8*8+1, 8*8+2, 8*8+3, 0, 1, 2, 3 },	/* bits are packed in groups of four */
	};

	// CONSIDER:  alloc _rgbChar/_rgbSprites dynamically based on MY_GFX_DECODE
	MyGfxDecode(tmpstorage, mgdChar, (LPBYTE)_rgbChars);

	//
	//
	//
	
	static const MY_GFX_DECODE mgdSprites = {
		{ 16, 16 },	/* 16*16 sprites */
		64,	/* 64 sprites */
		64*8,	/* every sprite takes 64 bytes */
		2,	/* 2 bits per pixel */
		4,	/* the two bitplanes for 4 pixels are packed into one byte */
		2,	/* 2 bits per pixel */
		{ 39 * 8, 38 * 8, 37 * 8, 36 * 8, 35 * 8, 34 * 8, 33 * 8, 32 * 8,
				7 * 8, 6 * 8, 5 * 8, 4 * 8, 3 * 8, 2 * 8, 1 * 8, 0 * 8 },
		{ 8*8, 8*8+1, 8*8+2, 8*8+3, 16*8+0, 16*8+1, 16*8+2, 16*8+3,
				24*8+0, 24*8+1, 24*8+2, 24*8+3, 0, 1, 2, 3 }
	};

	// CONSIDER:  alloc _rgbChar/_rgbSprites dynamically based on MY_GFX_DECODE
	MyGfxDecode(tmpstorage + 0x1000, mgdSprites, (LPBYTE)_rgbSprites);

	return S_OK;
}

/***************************************************************************

  Perform a memory read. This function is called by the CPU emulation.

***************************************************************************/

static BYTE M_RDMEM(WORD A)
{
	/* handle input ports (see memmap.h for details) */
	switch (A)
	{
		case IN0_PORT:
		{
			BYTE res = 0xff;

//			osd_poll_joystick();
			if (osd_key_pressed(OSD_KEY_3)) res &= ~IN0_CREDIT;
//			if (osd_key_pressed(OSD_KEY_F1)) res &= ~IN0_RACK_TEST;
			if (osd_key_pressed(OSD_KEY_DOWN)) res &= ~IN0_DOWN;
			if (osd_key_pressed(OSD_KEY_RIGHT)) res &= ~IN0_RIGHT;
			if (osd_key_pressed(OSD_KEY_LEFT)) res &= ~IN0_LEFT;
			if (osd_key_pressed(OSD_KEY_UP)) res &= ~IN0_UP;
			return res;
			break;
		}
		case IN1_PORT:
		{
			BYTE res = 0xff;

			if (osd_key_pressed(OSD_KEY_2)) res &= ~IN1_START2;
			if (osd_key_pressed(OSD_KEY_1)) res &= ~IN1_START1;
			if (osd_key_pressed(OSD_KEY_F2)) res &= ~IN1_TEST;
			return res;
			break;
		}

		case DSW1_PORT:
			return _rgbDSW[0];
			break;

		default:
			return RAM[A];
			break;
	}
}


/***************************************************************************

  Perform a memory read. This function is called by the CPU emulation.

***************************************************************************/

static unsigned char samples[8][32] =
{
	{ 0xff,0x11,0x22,0x33,0x44,0x55,0x55,0x66,0x66,0x66,0x55,0x55,0x44,0x33,0x22,0x11,
	0xff,0xdd,0xcc,0xbb,0xaa,0x99,0x99,0x88,0x88,0x88,0x99,0x99,0xaa,0xbb,0xcc,0xdd },
	{ 0xff,0x44,0x66,0x66,0x55,0x33,0x11,0x22,0x33,0x33,0x22,0x11,0xee,0xcc,0xbb,0xdd,
	0xff,0x11,0x33,0x22,0x00,0xdd,0xcc,0xbb,0xbb,0xcc,0xdd,0xbb,0x99,0x88,0x88,0xaa },
	{ 0xff,0x22,0x44,0x55,0x66,0x55,0x44,0x22,0xff,0xcc,0xaa,0x99,0x88,0x99,0xaa,0xcc,
	0xff,0x33,0x55,0x66,0x55,0x33,0xff,0xbb,0x99,0x88,0x99,0xbb,0xff,0x66,0xff,0x88 },
	{ 0xff,0x55,0x33,0x00,0x33,0x55,0x11,0xee,0x33,0x66,0x44,0xff,0x11,0x22,0xee,0xaa,
	0xff,0x44,0x00,0xcc,0xdd,0xff,0xaa,0x88,0xbb,0x00,0xdd,0x99,0xbb,0xee,0xbb,0x99 },
	{ 0x88,0x00,0x77,0xff,0x99,0x00,0x66,0xff,0xaa,0x00,0x55,0xff,0xbb,0x00,0x44,0xff,
	0xcc,0x00,0x33,0xff,0xdd,0x00,0x22,0xff,0xee,0x00,0x11,0xff,0xff,0x00,0x00,0xff },
	{ 0xff,0x00,0xee,0x11,0xdd,0x22,0xcc,0x33,0xbb,0x44,0xaa,0x55,0x99,0x66,0x88,0x77,
	0x88,0x77,0x99,0x66,0xaa,0x55,0xbb,0x44,0xcc,0x33,0xdd,0x22,0xee,0x11,0xff,0x00 },
	{ 0x88,0x99,0xaa,0xbb,0xcc,0xdd,0xee,0xff,0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,
	0x77,0x66,0x55,0x44,0x33,0x22,0x11,0x00,0xff,0xee,0xdd,0xcc,0xbb,0xaa,0x99,0x88 },
	{ 0x88,0x99,0xaa,0xbb,0xcc,0xdd,0xee,0xff,0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,
	0x88,0x99,0xaa,0xbb,0xcc,0xdd,0xee,0xff,0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77 }
};

static int soundchanged;

static inline int sh_wrmem(WORD A, BYTE V)
{
	return 0;

	if (RAM[SOUND_ENABLE] == 0)
	{
#if 0
		memset(_rgwPcmBuf, 0, sizeof(_rgwPcmBuf));
#endif
	}

	if (A >= SOUND_BASE && A < SOUND_BASE + 0x20)
	{
		if (RAM[A] != (V & 0x0f))
			soundchanged = 1;

		RAM[A] = V & 0x0f;

		/////////////////////////////////////////

		for (int voice = 0; voice < 3; voice++)
		{
			int freq,volume,wave;

			freq = RAM[SOUND_BASE + 0x14 + 5 * voice];	/* always 0 */
			freq = freq * 16 + RAM[SOUND_BASE + 0x13 + 5 * voice];
			freq = freq * 16 + RAM[SOUND_BASE + 0x12 + 5 * voice];
			freq = freq * 16 + RAM[SOUND_BASE + 0x11 + 5 * voice];

			if (voice == 0)
				freq = freq * 16 + RAM[SOUND_BASE + 0x10 + 5 * voice];
			else
				freq = freq * 16;

			freq = (SND_CLOCK / 2048) * freq / 512;
			volume = RAM[SOUND_BASE + 0x15 + 5 * voice];
			volume = (volume << 4) | volume;

			wave = RAM[SOUND_BASE + 0x05 + 5 * voice] & 7;

/*
			if (wave != currwave[voice])
			{
				osd_play_sample(voice,samples[wave],32,freq,volume,1);
				currwave[voice] = wave;
			}
			else osd_adjust_sample(voice,freq,volume);
*/
		
#if 0
			freq /= 80;

			memcpy(_rgwPcmBuf, _rgPcmBufs[(freq / 80) & 0xFF],
				sizeof(_rgPcmBufs[0]));
#endif

/*
			for (int i = 0; i < DIM(_rgwPcmBuf); i++)
			{
				if (voice == 0)
					_rgwPcmBuf[i] = samples[wave][(i * freq / 256) % 32] / 4;
				else
					_rgwPcmBuf[i] += samples[wave][(i * freq / 256) % 32] / 4;
			}
*/
		}

		return 1;
	}
	else return 0;
}

static inline int vh_wrmem(WORD A, BYTE V)
{
	if (A >= VIDEO_RAM_START && A < VIDEO_RAM_START + VIDEO_RAM_SIZE)
	{
			RAM[A] = V;

		return 1;
	}
	else if (A >= COLOR_RAM_START && A < COLOR_RAM_START + VIDEO_RAM_SIZE)
	{
			RAM[A] = V;

		return 1;
	}
	else return 0;
}

static void M_WRMEM(WORD A, BYTE V)
{
	A &= 0x7fff;	/* It looks like the Pac Man arcade machine has a 15 bit */
					/* path to RAM: we must clear the high bit otherwise CREDIT */
					/* and HIGHSCORE will not be displayed on the screen. */

	if (A <= ROM_END) return;	/* Do nothing, it's ROM */
	else if (vh_wrmem(A,V)) return;	/* the video hardware handled the write */
	else if (sh_wrmem(A,V)) return;	/* the sound hardware handled the write */
	else RAM[A] = V;
}

static HRESULT UpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
	return S_OK;
}

/***************************************************************************

  Interrupt handler. This function is called at regular intervals
  (determined by IPeriod) by the CPU emulation.

***************************************************************************/

static int _IntVector = 0;	/* Here we store the interrupt vector, if the code performs */
					/* an OUT to port $0. Not all games do it: many use */
					/* Interrupt Mode 1, which doesn't use an interrupt vector */
					/* (see Z80.c for details). */

static int Interrupt(void)
{
	if (!CheckMessages())
		CPURunning = 0;

	Throttle(FRAMES_PER_SECOND);

	if (RAM[INTERRUPT_ENABLE] == 0) return IGNORE_INT;
	else return _IntVector;
}

static void PengoOut(BYTE A, BYTE V)
{
//	if (sh_doout(A,V)) return;		/* the sound hardware handled the operation */

	/* OUT to port $0 is used to set the interrupt vector */
	if (A == 0) _IntVector = V;
}

// TODO:  call higher-level Exec function that handles the common parts
//        of Interrupt, pass it our CPU speed and frame rate

static HRESULT Run(LPVOID pvContext)
{
	EXEC_Z80_STRUCT es;
	
	ZERO_INIT(es);
	
	es.nCallbackFreq = 50000;
	es.pfnInterrupt = Interrupt;
	es.pfnReadMemory = M_RDMEM;
	es.pfnWriteMemory = M_WRMEM;
	es.rgbMemory = RAM;
	es.pfnOut = PengoOut;
	
	return ExecZ80(es);
}

// TODO:  blit 32-bits at a time!

inline static void DrawChar(LPBYTE rgbBuffer, LONG lPitch, int X, int Y,
	int offset_value, int color)
{
	X = ((X * 8) % 256);
	Y *= 8;

	LPDWORD pDest = (LPDWORD)&rgbBuffer[lPitch * Y + X];
	LPCDWORD pSrc = (LPCDWORD)&_rgbChars[offset_value][0][0];

	color = (color & 0x0F) << 2;
	color |= (color << 8) | (color << 16) | (color << 24);

	lPitch /= 4;

	*pDest       = (color | *pSrc++);
	*(pDest + 1) = (color | *pSrc++);
	pDest += lPitch;
	*pDest       = (color | *pSrc++);
	*(pDest + 1) = (color | *pSrc++);
	pDest += lPitch;
	*pDest       = (color | *pSrc++);
	*(pDest + 1) = (color | *pSrc++);
	pDest += lPitch;
	*pDest       = (color | *pSrc++);
	*(pDest + 1) = (color | *pSrc++);
	pDest += lPitch;
	*pDest       = (color | *pSrc++);
	*(pDest + 1) = (color | *pSrc++);
	pDest += lPitch;
	*pDest       = (color | *pSrc++);
	*(pDest + 1) = (color | *pSrc++);
	pDest += lPitch;
	*pDest       = (color | *pSrc++);
	*(pDest + 1) = (color | *pSrc++);
	pDest += lPitch;
	*pDest       = (color | *pSrc++);
	*(pDest + 1) = (color | *pSrc++);
	pDest += lPitch;
}

inline static void DrawSprite(LPBYTE rgbBuffer, LONG lPitch, int x, int y,
	int n, int color)
{
	//
	//
	//
	
	color = (color & 0x0F) << 2;
	color |= (color << 8) | (color << 16) | (color << 24);

	//
	//
	//
	
	LPDWORD pDest = (LPDWORD)&rgbBuffer[lPitch * y + x];
	LPCDWORD pSrc = (LPCDWORD)&_rgbSprites[n][0][0];
	
	//
	//  Draw sprite, with transparency, and color scheme adjusted
	//  to offscreen buffer, 32-bits at a time
	//
	//  CONSIDER:  unroll the loop
	//
	
	lPitch /= 4;
	
	for (int i = 0; i < 16; i++)
	{
		*(pDest + 0) = color | *pSrc++;
		*(pDest + 1) = color | *pSrc++;
		*(pDest + 2) = color | *pSrc++;
		*(pDest + 3) = color | *pSrc++;

		pDest += lPitch;
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static HRESULT RenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	LPBYTE pBase = &RAM[VIDEO_RAM_START];

	int i;

//	for (i = 0; i < 16 * 16; i++)
//		DrawChar(rgbBuffer, lPitch, i % 16, i / 16, i, 1);

//	return S_OK;

	//
	//  Draw 32x32 characters
	//

	for (i = 0; i < 32 * 32; i++)
	{
		int offset_value = *pBase;
		int color = 1; // pBase[0x400] & 0x1f;

		DrawChar(rgbBuffer, lPitch, 31 - (i / 32), i % 32, offset_value, color);

		pBase++;
	}

	/* Draw the sprites. Note that it is important to draw them exactly in this */
	/* order, to have the correct priorities. */
	for (i = 5;i >= 0;i--)
	{
		DrawSprite(rgbBuffer, lPitch,
			255 - RAM[SPRITE_COORDS_BASE + 2*i],
			255 - RAM[SPRITE_COORDS_BASE+1 + 2*i],
			RAM[SPRITE_CODES_BASE + 2*i] >> 2,
			RAM[SPRITE_CODES_BASE+1 + 2*i]);
	}

	return S_OK;
}

/*---------------------------------------------------------------------------

	Define game descriptors

---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdPengo = 
{
	TEXT("Pengo"),
	Init,
	Run,
	RenderScene,
	NULL,
	NULL,
	NULL,
	UpdateSound
};

