/*---------------------------------------------------------------------------

	File:
		scramble.cpp
	
	Abstract:
		Implements the "Scramble" game driver functions.

	Revisions:
		02-Mar-97 EAM Created

	Copyright (C) 1997 Edward Massey

	Hardware information was derived from the source code for the excellent
	classic arcade emulator MAME, by Nicola Salmoria.
	
---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "game.h"
#include "game1.h"
#include "../z80/execz80.h"
#include "../main/ay8910.h"

static AY8910 _rgAy8910[2];

extern void SoundUpdateMixer(LPBYTE rgbBuffer, LONG cbBuffer);
extern void SoundSetEnvelope(int iChnl, const char *pData, int cSamples, BOOL fLoop);

extern void sound_command_w(int offset,int data);
extern int pending_commands;
extern int sound_command_r();
extern void SoundUpdateMixer(LPBYTE rgbBuffer, LONG cbBuffer);

struct LINE_INFO
{
	BYTE bShift;
	BYTE bColor;
};

struct SPRITE_INFO
{
	BYTE x, i, c, y;
};

struct BULLET_INFO
{
	BYTE d1, x, d2, y;
};

struct SCRAMBLE_GAME
{
	static IMAGE *pimgChars;
	static IMAGE *pimgSprites;
	
	static const DECODE_GRAPHICS dgsChars;
	static const DECODE_GRAPHICS dgsSprites;
	static const TCHAR szPath[];
	static const RomModule GFXROMs[];
	static const RomModule ROMs[];
	static const RomModule AudioROMs[];
	static const SIZE _sizBuffer;
	static const RECT _rcDisplay;
	static IExecZ80 *_pZ80;
	static IExecZ80 *_pZ80Audio;
	static const SPRITE_INFO *psi;

	static HRESULT Init(GAME_LOAD_INFO &gli);
	static int Exec();
	static void Term();
	static BYTE M_RDMEM(WORD A);

	static void M_WRMEM(WORD A, BYTE V);
	static void RenderScene(LPBYTE rgbBuffer, LONG lPitch);
	static void UpdateSound(LPBYTE rgbBuffer, LONG cbBuffer);

	static BYTE AudioM_RDMEM(WORD A);
	static void AudioM_WRMEM(WORD A, BYTE V);

	static BYTE In(BYTE);
	static void Out(BYTE,BYTE);

	static BYTE rgbAudioRAM[0x10000];
};

IMAGE *SCRAMBLE_GAME::pimgChars;
IMAGE *SCRAMBLE_GAME::pimgSprites;
IExecZ80 *SCRAMBLE_GAME::_pZ80;
IExecZ80 *SCRAMBLE_GAME::_pZ80Audio;
const SPRITE_INFO *SCRAMBLE_GAME::psi = (SPRITE_INFO*)&RAM[0x5040];
BYTE SCRAMBLE_GAME::rgbAudioRAM[0x10000];

const DECODE_GRAPHICS SCRAMBLE_GAME::dgsChars =
{
	0x00000000,
	{
		{ 8, 8 },
		256,
		8 * 8,
		2,
		{ 0, 8 * 8 * 256 },
		2,
		{ 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
		{ 0, 1, 2, 3, 4, 5, 6, 7 },
	},
	1,
	1, 
	{
		0,
	},
	0, 
	NULL
};

const DECODE_GRAPHICS SCRAMBLE_GAME::dgsSprites =
{
	0x00000001,
	{
		{ 16, 16 },
		64,
		32 * 8,
		2,
		{ 0, 64 * 16 * 16 },
		2,
		{ 23*8, 22*8, 21*8, 20*8, 19*8, 18*8, 17*8, 16*8,
			7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
		{ 0, 1, 2, 3, 4, 5, 6, 7,
			8*8+0, 8*8+1, 8*8+2, 8*8+3, 8*8+4, 8*8+5, 8*8+6, 8*8+7 },
	},
	1,
	4, 
	{  0, 64, 128, 3 * 64 },
	0, 
	NULL
};

const TCHAR SCRAMBLE_GAME::szPath[] = "scramble";

const RomModule SCRAMBLE_GAME::GFXROMs[] = {
	RAGE_ROMTABLE_ENTRY("5f", 0x0000, 0x0800)
	RAGE_ROMTABLE_ENTRY("5h", 0x0800, 0x0800)
RAGE_ROMTABLE_END

const RomModule SCRAMBLE_GAME::ROMs[] = {
	RAGE_ROMTABLE_ENTRY("2c", 0x0000, 0x0800)
	RAGE_ROMTABLE_ENTRY("2e", 0x0800, 0x0800)
	RAGE_ROMTABLE_ENTRY("2f", 0x1000, 0x0800)
	RAGE_ROMTABLE_ENTRY("2h", 0x1800, 0x0800)
	RAGE_ROMTABLE_ENTRY("2j", 0x2000, 0x0800)
	RAGE_ROMTABLE_ENTRY("2l", 0x2800, 0x0800)
	RAGE_ROMTABLE_ENTRY("2m", 0x3000, 0x0800)
	RAGE_ROMTABLE_ENTRY("2p", 0x3800, 0x0800)
RAGE_ROMTABLE_END

const RomModule SCRAMBLE_GAME::AudioROMs[] = {
	RAGE_ROMTABLE_ENTRY("5c", 0x0000, 0x0800)
	RAGE_ROMTABLE_ENTRY("5d", 0x0800, 0x0800)
	RAGE_ROMTABLE_ENTRY("5e", 0x1000, 0x0800)
RAGE_ROMTABLE_END

const SIZE SCRAMBLE_GAME::_sizBuffer = { 256 + 16, 256 + 16};
const RECT SCRAMBLE_GAME::_rcDisplay = { 16, 0, 256 - 16, 256 };

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static const BYTE _rgbColorPROM[] =
{
	/* palette */
	0x00,0x17,0xC7,0xF6,
	0x00,0x17,0xC0,0x3F,
	0x00,0x07,0xC0,0x3F,
	0x00,0xC0,0xC4,0x07,
	0x00,0xC7,0x31,0x17,
	0x00,0x31,0xC7,0x3F,
	0x00,0xF6,0x07,0xF0,
	0x00,0x3F,0x07,0xC4
};

static void BuildCompositePalette(PALETTEENTRY *ppeOut, UINT &cpeOut)
{
	for (cpeOut = 0; cpeOut < DIM(_rgbColorPROM); cpeOut++)
	{
		int i = cpeOut;

		int bit0,bit1,bit2;

		bit0 = (_rgbColorPROM[i] >> 0) & 0x01;
		bit1 = (_rgbColorPROM[i] >> 1) & 0x01;
		bit2 = (_rgbColorPROM[i] >> 2) & 0x01;

		ppeOut[cpeOut].peRed = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;

		bit0 = (_rgbColorPROM[i] >> 3) & 0x01;
		bit1 = (_rgbColorPROM[i] >> 4) & 0x01;
		bit2 = (_rgbColorPROM[i] >> 5) & 0x01;

		ppeOut[cpeOut].peGreen = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;

		bit0 = 0;
		bit1 = (_rgbColorPROM[i] >> 6) & 0x01;
		bit2 = (_rgbColorPROM[i] >> 7) & 0x01;

		ppeOut[cpeOut].peBlue  = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;
	}
}

static int scramble_portB_r()
{
	int clockticks,clock;

#define TIMER_RATE (512*2)

/*
	EXECZ80_CONTEXT ctx;

	SCRAMBLE_GAME::_pZ80Audio->GetContext(ctx);

	clockticks = (ctx.IPeriod - ctx.ICount);
*/

	LARGE_INTEGER x, y;
	
	if (QueryPerformanceCounter(&x) && QueryPerformanceFrequency(&y))
	{
//		clockticks = 1572750L * (x.QuadPart - _nPerfTime.QuadPart) / y.QuadPart;
//		clockticks = 2 * 1572750L * (x.QuadPart - _nPerfTime.QuadPart) / y.QuadPart;
		clockticks = (int)(x.QuadPart * 2L * 1572750L / y.QuadPart);
	}

	clock = clockticks / TIMER_RATE;

	clock = ((clock & 0x01) << 4) | ((clock & 0x02) << 6) |
			((clock & 0x08) << 2) | ((clock & 0x10) << 2);

	return clock;
}

BYTE ScrambleAudioIn(BYTE A)
{
	switch (A)
	{
		case 0x20:
			return _rgAy8910[0].ReadRegLatch();

/*
		{
			switch (lastreg0)
			{
				case AY_PORTA:
					return sound_command_r(0);

				case AY_PORTB:
					return scramble_portB_r(1);
			
				default:
					return _rgAy8910[0].ReadReg(lastreg0);
			}
		}
*/

		case 0x80:
			return _rgAy8910[1].ReadRegLatch();
	}

	return 0;
}

void ScrambleAudioOut(BYTE A, BYTE data)
{
	switch (A)
	{
		case 0x10:
			_rgAy8910[0].WriteLatch(data);
			break;

		case 0x20:
			_rgAy8910[0].WriteRegLatch(data);
			break;

		case 0x40:
			_rgAy8910[1].WriteLatch(data);
			break;

		case 0x80:
			_rgAy8910[1].WriteRegLatch(data);
			break;
	}
}

/*
static unsigned char porthandler0(AY8910 *chip, int port, int iswrite, unsigned char val)
{
	int num = 0;

	if (iswrite)
	{
	}
	else
	{
		chip->Regs[port] = 0;

		if (port == 0x0e)
			chip->Regs[port] = sound_command_r(0);
		else
			chip->Regs[port] = scramble_portB_r(1);
	}

	return 0;
}
*/

HRESULT SCRAMBLE_GAME::Init(GAME_LOAD_INFO &gli)
{
	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	BuildCompositePalette(gli.ppe, gli.cpe);

	BYTE tmpstorage[0x4000];

	if (readroms(RAM, ROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	if (readroms(rgbAudioRAM, AudioROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	if (readroms(tmpstorage, GFXROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//  Decode sprites
	//
	
	MyGfxDecode3(dgsChars, tmpstorage, pimgChars);
	MyGfxDecode3(dgsSprites, tmpstorage, pimgSprites);

	//
	//
	//
	
	static const EXECZ80_MEM_DESCRIPTOR rgmd[] = 
	{
		{ 0x0000, 0xFFFF, 0, M_RDMEM, M_WRMEM },
		{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
	};

	EXECZ80_CREATE cs;

	ZERO_INIT(cs);

	cs.rgbMemory   = RAM;
	cs.pfnMemRead  = M_RDMEM;
	cs.pfnMemWrite = M_WRMEM;
	cs.pmd         = rgmd;
	
	HRESULT hr = EclCreateZ80(cs, _pZ80);

	if (FAILED(hr))
		return hr;

	//
	//
	//
	
	static const EXECZ80_MEM_DESCRIPTOR rgmdAudio[] = 
	{
		{ 0x0000, 0xFFFF, 0, AudioM_RDMEM, AudioM_WRMEM },
		{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
	};

	ZERO_INIT(cs);

	cs.rgbMemory    = rgbAudioRAM;
	cs.pfnMemRead   = AudioM_RDMEM;
	cs.pfnMemWrite  = AudioM_WRMEM;
	cs.pfnPortRead  = ScrambleAudioIn;
	cs.pfnPortWrite = ScrambleAudioOut;
	cs.pmd          = rgmdAudio;

	hr = EclCreateZ80(cs, _pZ80Audio);

//	if (AYInit(2, 1789750000, 11025, 512, 0) == 0)
	{
//		AYSetPortHandler(0,AY_PORTA,porthandler0);
//		AYSetPortHandler(0,AY_PORTB,porthandler0);
	}

	_rgAy8910[0].m_pfnReadPortA = sound_command_r;
	_rgAy8910[0].m_pfnReadPortB = scramble_portB_r;

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

BYTE SCRAMBLE_GAME::M_RDMEM(WORD A)
{
	switch (A)
	{
		case 0x8100:
		{
			BYTE b = 0xFF;

			if (osd_key_pressed(OSD_KEY_ALT))
				b ^= 0x02;

			if (osd_key_pressed(OSD_KEY_3))
				b ^= 0x04;

			if (osd_key_pressed(OSD_KEY_CONTROL))
				b ^= 0x08;

			if (osd_key_pressed(OSD_KEY_RIGHT))
				b ^= 0x10;

			if (osd_key_pressed(OSD_KEY_LEFT))
				b ^= 0x20;

			return b;
		}
			
		case 0x8101:
		{
			BYTE b = 0xFC;

			if (osd_key_pressed(OSD_KEY_1))
				b ^= 0x80;

			if (osd_key_pressed(OSD_KEY_2))
				b ^= 0x40;

			return b;
		}

		case 0x8102:
		{
			BYTE b = 0xF1;

			if (osd_key_pressed(OSD_KEY_UP))
				b ^= 0x10;

			if (osd_key_pressed(OSD_KEY_DOWN))
				b ^= 0x40;

			/* avoid protection */

			EXECZ80_CONTEXT ctx;

			_pZ80->GetContext(ctx);

			if (ctx.PC == 0x00e4)
				b &= 0x7f;

			return b;
		}

		case 0x8202:
			return 0x6F;

		default:
			return RAM[A];
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

void SCRAMBLE_GAME::M_WRMEM(WORD A, BYTE V)
{
	if (A < 0x4000)
		return;

	switch (A)
	{
		case 0x8200:
			sound_command_w(A, V);
			break;
	}

	RAM[A] = V;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE rgbBufs[5][512];

void SCRAMBLE_GAME::UpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
	_rgAy8910[0].UpdateSound(rgbBufs[0], DIM(rgbBufs[0]));
	_rgAy8910[1].UpdateSound(rgbBufs[1], DIM(rgbBufs[0]));

	SoundSetEnvelope(1, (char*)rgbBufs[0], DIM(rgbBufs[0]), TRUE);
	SoundSetEnvelope(2, (char*)rgbBufs[1], DIM(rgbBufs[0]), TRUE);

	SoundUpdateMixer(rgbBuffer, cbBuffer);
}

int SCRAMBLE_GAME::Exec()
{
	int cCycles;
	
	cCycles = _pZ80->Exec(50000);

//	if (RAM[0x6801] == 0)
		_pZ80->Nmi();

	BOOL fSoundEnable = FALSE; // TRUE;

	if (fSoundEnable)
	{
		_pZ80Audio->Exec(1572750 / 60);

		if (pending_commands > 0)
			_pZ80Audio->Irq(0xFF);
	}

	return cCycles;
}

void SCRAMBLE_GAME::Term()
{
	_pZ80->Release();

	FreeImage(pimgChars);
	FreeImage(pimgSprites);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

void SCRAMBLE_GAME::RenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	int i, j;

	//
	//  Draw 32x32 characters, shifted into correct place
	//

	const LINE_INFO *pli = (LINE_INFO*)&RAM[0x5000];

	for (i = 0; i < 32; i++)
	{
		const int nShift = pli[i].bShift;
		const int color = pli[i].bColor;
		const int sy = 8 * i;

		for (j = 0; j < 32; j++)
		{
			const int sx = (8 * (31 - j) + nShift) % 256;

			DrawChar((LPBYTE)pimgChars->rgbData, RAM[0x4800 + i + j * 32],
				color << 2, sx, sy, rgbBuffer, lPitch);
		}
	}

	//
	//
	//

	const BULLET_INFO *pbi = (BULLET_INFO*)&RAM[0x5060];
	
	for (i = 0; i < 8; i++)
	{
		const int y = 256 - pbi[i].y - 8;

		if (y >= 0)
			rgbBuffer[pbi[i].x + y * lPitch] = 3;
	}

	//
	//  Draw sprites
	//

	for (i = 0; i < 8; i++)
	{
		if (psi[i].y > 8)
		{
			// NOTE:  use of prerotated sprite images means we can pass i unmodified

			DrawSprite(pimgSprites->rgbData, pimgSprites->rgbMask, 
				psi[i].i, psi[i].c << 2, psi[i].x, psi[i].y, rgbBuffer, lPitch);
		}
	}
}

BYTE SCRAMBLE_GAME::AudioM_RDMEM(WORD A)
{
	return SCRAMBLE_GAME::rgbAudioRAM[A];
}

void SCRAMBLE_GAME::AudioM_WRMEM(WORD A, BYTE B)
{
	if (A < 0x1800)
		return;

	SCRAMBLE_GAME::rgbAudioRAM[A] = B;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdScramble = 
{
	TEXT("Scramble"),
	NULL,
	SCRAMBLE_GAME::Init,
	SCRAMBLE_GAME::Exec,
	SCRAMBLE_GAME::RenderScene,
	SCRAMBLE_GAME::UpdateSound,
	SCRAMBLE_GAME::Term
};

