/*---------------------------------------------------------------------------

	File:
		zaxxon.cpp
	
	Abstract:
		Implements the "Zaxxon" game driver.

	Revisions:
		08-Mar-97 EAM Created

	Copyright (C) 1997 Edward Massey

	Hardware information for this driver was based on work done by Mirko
	Buffoni in his Zaxxon driver for MAME.

  	TODO:  Don't bother drawing chars that are in the first or last two cols.
	TODO:  Don't draw sprites that are too far off to the right.
	TODO:  Move to interleaved sprite data, should make a big difference here

	TODO:  Interleave sprite data and mask
	TODO:  Left shift sprite data during decode
	TODO:  Optimize Z80 performance

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "game.h"
#include "game1.h"
#include "../z80/execz80.h"

static IMAGE *_pimgChars1;
static IMAGE *_pimgChars2;
static IMAGE *_pimgSprites;
static IExecZ80 *_pZ80;

static BYTE _rgbBackground[0x8000];

static BYTE _rgbPrebitmap[256][4096];
static BYTE _rgbBackgroundbitmap[512][2303+32];

#define INTERRUPT_ENABLE 0xFFF0

struct SPRITE
{
	BYTE x, i, c, y;
};

static const SPRITE *rgs = (SPRITE*)&RAM[0xA000];

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static const PALETTEENTRY _rgpeMario[] =
{
	// 00 - enemy ship explosion

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 0,   0,   0   },
	{ 255, 255, 255 },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 01 - player ship explosion 1
	
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 02 - player ship explosion 1
	
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 06 - enumy plane fire
	
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 07 - satellite 1
	
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 08 - player's ship (after hitting wall)
	
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 09 - turret 1

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 0A - enemy plane

	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 0D - crosshairs
	
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 255 },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 10 - player ship/force field/explosions

	{ 0,   0,   0   },
	{ 255, 255, 0   },
	{ 255, 0,   0   },
	{ 0,   255, 0   },
	{ 255, 0,   255 },
	{ 255, 128, 0   },
	{ 0,   255, 255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 12 - turret 2
	
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 13 - barrier
	
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 14 - drum/radar
	
	{ 0,   0,   0   },
	{ 128, 0,   0   },
	{ 0,   0,   0   },
	{ 255, 128, 0   },
	{ 128, 128, 128 },
	{ 255, 64,  64  },
	{ 255, 0,   0   },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 17 - satellite 2
	
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// 1A - shadow
	
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 64,  64,  64  },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },
};

static const TCHAR szPath[] = "zaxxon";

RAGE_ROMTABLE_BEGIN(ROMs)
	RAGE_ROMTABLE_ENTRY("zaxxon.3",  0x0000, 0x2000)
	RAGE_ROMTABLE_ENTRY("zaxxon.2",  0x2000, 0x2000)
	RAGE_ROMTABLE_ENTRY("zaxxon.1",  0x4000, 0x1000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(GFXROMs)
	RAGE_ROMTABLE_ENTRY("zaxxon.14", 0x0000, 0x0800) // charset 1
	RAGE_ROMTABLE_ENTRY("zaxxon.15", 0x0800, 0x0800)
	RAGE_ROMTABLE_ENTRY("zaxxon.6",  0x1000, 0x2000) // charset 2
	RAGE_ROMTABLE_ENTRY("zaxxon.5",  0x3000, 0x2000)
	RAGE_ROMTABLE_ENTRY("zaxxon.4",  0x5000, 0x2000)
	RAGE_ROMTABLE_ENTRY("zaxxon.11", 0x7000, 0x2000) // sprites
	RAGE_ROMTABLE_ENTRY("zaxxon.12", 0x9000, 0x2000)
	RAGE_ROMTABLE_ENTRY("zaxxon.13", 0xb000, 0x2000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(BKGROMs)
	RAGE_ROMTABLE_ENTRY("zaxxon.8",  0x0000, 0x2000)
	RAGE_ROMTABLE_ENTRY("zaxxon.7",  0x2000, 0x2000)
	RAGE_ROMTABLE_ENTRY("zaxxon.10", 0x4000, 0x2000)
	RAGE_ROMTABLE_ENTRY("zaxxon.9",  0x6000, 0x2000)
RAGE_ROMTABLE_END

static const DECODE_GRAPHICS dgsChars1 =
{
	0x00000001,
	{
		{ 8, 8 },
		256,
		8 * 8,
		2,
		{ 0, 8 * 8 * 256 },
		2,
		{ 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
		{ 0, 1, 2, 3, 4, 5, 6, 7 },
	},

	1,
	
	1, 
	
	{
		0
	},
	0, 
	NULL
};

static const DECODE_GRAPHICS dgsChars2 =
{
	0x00000000,
	{
		{ 8, 8 },
		1024,
		8 * 8,
		3,
		{ 0, 8 * 8 * 1024, 2 * 8 * 8 * 1024 },
		3,
		{ 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8 },
		{ 0, 1, 2, 3, 4, 5, 6, 7 },
	},

	1,
	
	1, 
	
	{
		0
	},
	0, 
	NULL
};

static const DECODE_GRAPHICS dgsSprites =
{
	0x00000001,
	{
		{ 32, 32 },
		64,
		128 * 8,
		3,
		{ 0, 1024 * 8 * 8, 2 * 1024 * 8 * 8 },
		3,
		{ 103*8, 102*8, 101*8, 100*8, 99*8, 98*8, 97*8, 96*8,
			71*8, 70*8, 69*8, 68*8, 67*8, 66*8, 65*8, 64*8,
			39*8, 38*8, 37*8, 36*8, 35*8, 34*8, 33*8, 32*8,
			7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
		{ 0, 1, 2, 3, 4, 5, 6, 7,
			8*8+0, 8*8+1, 8*8+2, 8*8+3, 8*8+4, 8*8+5, 8*8+6, 8*8+7,
			16*8+0, 16*8+1, 16*8+2, 16*8+3, 16*8+4, 16*8+5, 16*8+6, 16*8+7,
			24*8+0, 24*8+1, 24*8+2, 24*8+3, 24*8+4, 24*8+5, 24*8+6, 24*8+7 },
	},

	1,
	
	4, 
	
	{
		0, 64, 128, 64 + 128
	},
	0, 
	NULL
};

static const SIZE _sizBuffer = { 256 + 32, 256 + 32};
static const RECT _rcDisplay = { 16, 0, 256 - 16, 256};

static void SetupBkgnd();
static BYTE M_RDMEM(WORD A);
static void M_WRMEM(WORD A, BYTE);

static const EXECZ80_MEM_DESCRIPTOR rgmd[] = 
{
	{ 0x0000, 0xFFFF, 0, M_RDMEM, M_WRMEM },
	{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
};

HRESULT ZaxxonInit(GAME_LOAD_INFO &gli)
{
	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	for (int i = 0; i < DIM(_rgpeMario); i++)
		gli.ppe[i] = _rgpeMario[i];

	gli.cpe = DIM(_rgpeMario);

	//
	//
	//
	
	if (readroms(RAM, ROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//
	//
	
	BYTE tmpstorage[0xD000];

	if (readroms(tmpstorage, GFXROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	MyGfxDecode3(dgsChars1,  tmpstorage,          _pimgChars1);
	MyGfxDecode3(dgsChars2,  tmpstorage + 0x1000, _pimgChars2);
	MyGfxDecode3(dgsSprites, tmpstorage + 0x7000, _pimgSprites);

	if (readroms(_rgbBackground, BKGROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	SetupBkgnd();

	//
	//
	//
	
	EXECZ80_CREATE cs;

	ZERO_INIT(cs);

	cs.rgbMemory    = RAM;
	cs.pfnMemRead   = M_RDMEM;
	cs.pfnMemWrite  = M_WRMEM;
	cs.pmd          = rgmd;
	
	HRESULT hr = EclCreateZ80(cs, _pZ80);

	if (FAILED(hr))
		return hr;

	return S_OK;
}

void ZaxxonTerm()
{
	if (_pZ80 != NULL)
		_pZ80->Release();
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE M_RDMEM(WORD A)
{
	switch (A)
	{
		case 0xC000:
		{
			BYTE b = 0x00;

			if (osd_key_pressed(OSD_KEY_RIGHT))
				b ^= 0x01;

			if (osd_key_pressed(OSD_KEY_LEFT))
				b ^= 0x02;

			if (osd_key_pressed(OSD_KEY_DOWN))
				b ^= 0x04;

			if (osd_key_pressed(OSD_KEY_UP))
				b ^= 0x08;

			if (osd_key_pressed(OSD_KEY_CONTROL))
				b ^= 0x10;

			return b;
		}

		case 0xC001:
		{
			BYTE b = 0x00;

			return b;
		}

		case 0xC002:
		{
			BYTE b = 0x7F;

			return b;
		}

		case 0xC003:
		{
			BYTE b = 0x33;

			return b;
		}

		case 0xC100:
		{
			BYTE b = 0x00;

			if (osd_key_pressed(OSD_KEY_1))
				b ^= 0x04;

			if (osd_key_pressed(OSD_KEY_2))
				b ^= 0x08;

			if (osd_key_pressed(OSD_KEY_3))
				b ^= 0x80;

			return b;
		}

		default:
			return RAM[A];
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void M_WRMEM(WORD A, BYTE V)
{
	if (A < 0x5000)
		return;

	RAM[A] = V;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

int ZaxxonExec()
{
	int cCycles;
	
	cCycles = _pZ80->Exec(50000);
	
	if (RAM[INTERRUPT_ENABLE])
		cCycles += _pZ80->Irq(0xFF);

	return cCycles;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void SetupBkgnd()
{
	int offs, sx, sy;
	
	for (offs = 0;offs < 0x4000;offs++)
	{
		sx = 8 * (511 - offs / 32);
		sy = 8 * (offs % 32);

/*
		DrawChar((LPBYTE)_rgbChars2, _rgbBackground[offs] + 256 * ((_rgbBackground[offs + 0x4000]) & 3),
			0, //(_rgbBackground[0x4000 + offs] >> 2) << 3,
			sx, sy, _rgbPrebitmap[0], 4096);
*/

//		_pimgChars2->DrawChar(rgbBuffer, lPitch, x, y,
//			offset_value, color << 2);

	}

	for (offs = 0; offs < 2335; offs++)
	{
		for (sx = 0;sx < 512;sx += 2)
		{
			if (offs < 256)
				_rgbBackgroundbitmap[offs][sx+1] = _rgbPrebitmap[offs][sx];
		}
	}

	return;

	for (offs = -510;offs < 4096;offs += 2)
	{
		sy = (2302-510/2) - offs/2;

		for (sx = 0;sx < 512;sx += 2)
		{
			if (offs + sx < 0 || offs + sx >= 4096)
			{
				_rgbBackgroundbitmap[sy][sx] = 0;
				_rgbBackgroundbitmap[sy][sx] = 0;
			}
			else
			{
				_rgbBackgroundbitmap[sy][sx]   = _rgbPrebitmap[sx/2][offs + sx];
				_rgbBackgroundbitmap[sy][sx+1] = _rgbPrebitmap[sx/2][offs + sx+1];
			}
		}
	}
}

static const int _rgiHex[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 17, 18, 19, 20, 21, 22 };

void ZaxxonRenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	int i, j;

/*
	int skew,scroll;

	scroll = 2048+63 - (RAM[0xfff8] + 256 * RAM[0xfff9]);
	skew = 128;

	static int x = 0, y = 0;
	
	if (GetAsyncKeyState(VK_UP) < 0 && y > 0)
		y--;
	else if (GetAsyncKeyState(VK_DOWN) < 0)
		y++;
		
	if (GetAsyncKeyState(VK_LEFT) < 0 && x > 0)
		x--;
	else if (GetAsyncKeyState(VK_RIGHT) < 0)
		x++;

	for (i = 0;i < 256-4*8;i++)
	{
		for (j = 0; j < 255; j++)
		{
//			rgbBuffer[i * lPitch + j] = _rgbBackgroundbitmap[scroll][j + skew];
			rgbBuffer[i * lPitch + j] = _rgbBackgroundbitmap[i + y][j + x];
		}

//		skew -= 2;
	}
*/

/*
	for (i = 0; i < 32; i++)
	{
		for (j = 0; j < 32; j++)
		{
			int k = i * 32 + j + base;
			
			DrawChar2(rgbBuffer, lPitch, 8 * i, 8 * j,
				_rgbBackground[k] + 256 * ((_rgbBackground[k + 0x4000]) & 3), 0);
		}
	}

	DrawChar(rgbBuffer, lPitch, 16, 0, _rgiHex[(base & 0xf000) >> 12], 0);
	DrawChar(rgbBuffer, lPitch, 24, 0, _rgiHex[(base & 0x0f00) >> 8], 0);
	DrawChar(rgbBuffer, lPitch, 32, 0, _rgiHex[(base & 0x00f0) >> 4], 0);
	DrawChar(rgbBuffer, lPitch, 40, 0, _rgiHex[(base & 0x000f) >> 0], 0);
*/

	for (i = 0; i < 32; i++)
	{
		for (j = 2; j < 30; j++)
		{
			const int c = RAM[0x8000 + i + j * 32];

			// skip blanks in playfield
//			if (c != 0x60)
//				DrawChar((LPBYTE)_rgbChars, c, 0, 8 * (31 - j), 8 * i, rgbBuffer, lPitch);

			_pimgChars1->DrawChar(rgbBuffer, lPitch, 8 * (31 - j), 8 * i,
				c, 0);
		}
	}

/*
	DrawChar(rgbBuffer, lPitch, 16, 0, _rgiHex[(RAM[0xFFF8] & 0xF0) >> 4], 0);
	DrawChar(rgbBuffer, lPitch, 24, 0, _rgiHex[(RAM[0xFFF8] & 0x0F) >> 0], 0);
	DrawChar(rgbBuffer, lPitch, 32, 0, _rgiHex[(RAM[0xFFF9] & 0xF0) >> 4], 0);
	DrawChar(rgbBuffer, lPitch, 40, 0, _rgiHex[(RAM[0xFFF9] & 0x0F) >> 0], 0);
*/

	for (i = 63; i >= 0; i--)
	{
		if (rgs[i].x != 0xFF && rgs[i].x != 0x00)
		{
			_pimgSprites->DrawSprite2(rgbBuffer, lPitch,
				rgs[i].x, rgs[i].y, rgs[i].i, (rgs[i].c & 0x1F) << 3);
		}
	}
}

void ZaxxonUpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdZaxxon = 
{
	TEXT("Zaxxon"),
	(LPVOID)0x00000001,
	ZaxxonInit,
	ZaxxonExec,
	ZaxxonRenderScene,
	ZaxxonUpdateSound,
	ZaxxonTerm
};

