/*---------------------------------------------------------------------------

	File:
		exec6502.cpp

	- use a switch statement (instead of a table of function pointers) so
	  compiler will create a jump table

	- all execution is done within a single function so compiler can do more
	  register optimization

	- flags aren't necessarily stored in 6502 format internally to allow 
	  flexibility

	- there is no _rgcCycles table, cCycles is decremented with immediate
	  value

	Abstract:
		Provides emulation of the Motorola 6502 processor.

	Revisions:
		18-Feb-97 EAM Created
		07-Apr-97 EAM Began a real implementation
		14-Apr-97 EAM Optimizations (no m_regPC, m_regS, Cycles[])
		15-Apr-97 EAM Fixing bugs

	Copyright (C) 1997 Edward Massey

	References:
		M6502: portable 6502 emulator - Marat Fayzullin, Alex Krasivsky
		M6502: Neil Bradley
		6502.c: author unknown

	Performance stats....
		Centipede (1 MHz 6502), K5/133, Marat's: ~530 fps, mine: ~700 fps

	Consider:  un-class-ify it; use macros or inline by hand; see if we can
		do faster arithmetic/flags

	CONSIDER:  store a flag that tells the loop to end prematurely so we can
		immediately exit the loop when a memory semaphore is being polled.
		Ex:  this flag could be set with SetContext during a memory read.

	Why another 6502?
	 * none of the ones I've seen are written very cleanly
	 * I like this method of handling interrupts better
	 * I like COM interfaces
	 * better performance that other C implementations

	This library is free software; you can redistribute it and/or modify it
	under the terms of version 2 of the GNU Library General Public License
	as published by the Free Software Foundation.

	This library is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library
	General Public License for more details.
	To obtain a copy of the GNU Library General Public License, write to the
	Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	Any permitted reproduction of these routines, in whole or in part, must
	bear this legend.

---------------------------------------------------------------------------*/

#include "pch.h"
#include "ex6502i.h"

#pragma inline_recursion(on)

int CExec6502::ZNTable[256];
int CExec6502::NoOvTable[256];

extern "C" void TRACE(LPCTSTR pszFormat, ...)
{
	va_list marker;
	va_start(marker, pszFormat);
	TCHAR szBuf[128];
	_vsnprintf(szBuf, 128, pszFormat, marker);

	static FILE *pfTrace;

	if (pfTrace == NULL)
		pfTrace = fopen("debug.txt", "w");

	if (pfTrace != NULL)
	{
		fputs(szBuf, pfTrace);
		fflush(pfTrace);
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

EXEC6502_API HRESULT EclCreate6502(EXEC6502_CREATE &cs, IExec6502 *&pExec)
{
	pExec = new CExec6502(cs);

	if (pExec == NULL)
		return E_OUTOFMEMORY;

	pExec->AddRef();

	return S_OK;
}

static void WriteIllegal(WORD wAddr, BYTE bVal)
{
	TRACE(__FILE__ ": illegal memory write [%04X]->%02X\n", wAddr, bVal);
}

static BYTE ReadIllegal(WORD wAddr)
{
	TRACE(__FILE__ ": illegal memory read [%04X]\n", wAddr);

	return 0;
}

static BYTE _rgbAdcFlags[2][256][256];
static BYTE _rgbAdcValues[2][256][256];

CExec6502::CExec6502(EXEC6502_CREATE &cs) : m_rgbMemory(cs.rgbMemory)
{
	//
	//  Initialize reference count
	//
	
	m_cRef = 0;

	//
	//  Initialize flags tables
	//
	
	if (ZNTable[0] != Z_FLAG)
	{
		ZNTable[0] = Z_FLAG;

		for (int i = 128; i < 256; i++)
			ZNTable[i] = N_FLAG;

		NoOvTable[0] = Z_FLAG;

		//
		//
		//

		int regA, bArg, regP;

		_ASSERT(C_FLAG == 1);

		for (regP = 0; regP <= 1; regP++)
		{
			for (regA = 0; regA < 256; regA++)
			{
				for (bArg = 0; bArg < 256; bArg++)
				{
					WORD w = regA + bArg + (regP & C_FLAG);

					BYTE res = (~(regA ^ bArg) & (regA ^ LOBYTE(w)) & 0x80 ? V_FLAG : 0) |
						(HIBYTE(w) ? C_FLAG : 0) | ZNTable[LOBYTE(w)];

					_rgbAdcFlags[regP][regA][bArg] = res;
					_rgbAdcValues[regP][regA][bArg] = LOBYTE(w);
				}
			}
		}
	}

	//
	//  Initialize memory address descriptor tables
	//

	for (DWORD dwAddr = 0; dwAddr < 0x10000; dwAddr++)
	{
		//
		//  Defaults...
		//
		
		m_rgMemDesc[dwAddr]  = 0;
		m_rgpfnRead[dwAddr]  = ReadIllegal;
		m_rgpfnWrite[dwAddr] = WriteIllegal;

		//
		//
		//
		
		const EXEC6502_MEM_DESCRIPTOR *pmd;

		for (pmd = cs.pmd; pmd->pfnRead != (EXEC6502_MEMREAD_PROC)0xFFFFFFFF; pmd++)
		{
			if (dwAddr < pmd->dwAddrMin || dwAddr > pmd->dwAddrMax)
				continue;

			//
			//
			//
			
			if (pmd->pfnRead == NULL)
			{
				m_rgMemDesc[dwAddr] |= MF_READ_DIRECT;
				m_rgpfnRead[dwAddr] = NULL;
			}
			else
			{
				m_rgMemDesc[dwAddr] &= ~MF_READ_DIRECT;
				m_rgpfnRead[dwAddr] = pmd->pfnRead;

				if (pmd->dwFlags & MF_ADJUST_ADDRESS)
				{
					m_rgMemDesc[dwAddr] |= MF_ADJUST_ADDRESS;
//					_ASSERT(FALSE); // not quite that simple
				}
			}
				
			//
			//
			//
			
			if (pmd->pfnWrite == NULL)
			{
				m_rgMemDesc[dwAddr] |= MF_WRITE_DIRECT;
				m_rgpfnWrite[dwAddr] = NULL;
			}
			else
			{
				m_rgMemDesc[dwAddr] &= ~MF_WRITE_DIRECT;
				m_rgpfnWrite[dwAddr] = pmd->pfnWrite;

				if (pmd->dwFlags & MF_ADJUST_ADDRESS)
				{
					m_rgMemDesc[dwAddr] |= MF_ADJUST_ADDRESS;
//					_ASSERT(FALSE); // not quite that simple
				}
			}
		}
	}

#ifdef _DEBUG
	DWORD dwAddrMin = 0;

	TRACE("6502 memory descriptor dump (this=%p)\n", this);
	TRACE("\n");
	TRACE("addresses   read       write\n");
	TRACE("---------   ---------- ----------\n");

	for (dwAddr = 0; dwAddr < 0x10000; dwAddr++)
	{
		if ((m_rgMemDesc[dwAddr]  != m_rgMemDesc[dwAddrMin]) ||
			(m_rgpfnRead[dwAddr]  != m_rgpfnRead[dwAddrMin]) ||
			(m_rgpfnWrite[dwAddr] != m_rgpfnWrite[dwAddrMin]) ||
			dwAddr == 0xFFFF)
		{
			static const TCHAR szFn[][16] = {
				"null      ",
				"illegal   ",
				"callback  ",
				"direct    "
			};
			
			LPCTSTR pszRead, pszWrite;
			
			if (m_rgMemDesc[dwAddrMin] & MF_READ_DIRECT)
			{
					pszRead = szFn[3];
			}
			else
			{
				if (m_rgpfnRead[dwAddrMin] == NULL)
					pszRead = szFn[0];
				else if (m_rgpfnRead[dwAddrMin] == ReadIllegal)
					pszRead = szFn[1];
				else
					pszRead = szFn[2];
			}
				
			if (m_rgMemDesc[dwAddrMin] & MF_WRITE_DIRECT)
			{
					pszWrite = szFn[3];
			}
			else
			{
				if (m_rgpfnWrite[dwAddrMin] == NULL)
					pszWrite = szFn[0];
				else if (m_rgpfnWrite[dwAddrMin] == WriteIllegal)
					pszWrite = szFn[1];
				else
					pszWrite = szFn[2];
			}

			TRACE("%04X-%04X   %s %s\n", dwAddrMin, dwAddr - 1,
				pszRead, pszWrite);
				
			dwAddrMin = dwAddr;
		}
	}

	TRACE("\n");
#endif

	//
	//  Start with a freshly reset 6502
	//

	Reset();
}

CExec6502::~CExec6502()
{
	_ASSERT(m_cRef == 0);
}

HRESULT CExec6502::QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppvObject)
{
	if (riid == IID_IUnknown)
		*ppvObject = this;
	else
		return E_NOINTERFACE;
	
	AddRef();
	
	return S_OK;
}

ULONG CExec6502::AddRef()
{
	return ++m_cRef;
}

ULONG CExec6502::Release()
{
	_ASSERT(m_cRef > 0);

	if (--m_cRef != 0)
		return m_cRef;

	delete this;

	return 0;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

void CExec6502::SetContext(const EXEC6502_CONTEXT &ctx)
{
	m_regA = ctx.regA;
	m_regX = ctx.regX;
	m_regY = ctx.regY;
	m_regP = ctx.regP;
	m_rgbStack = &m_rgbMemory[ctx.regS];
	SetPC(ctx.regPC);
}

void CExec6502::GetContext(EXEC6502_CONTEXT &ctx)
{
	ctx.regA  = m_regA;
	ctx.regX  = m_regX;
	ctx.regY  = m_regY;
	ctx.regP  = m_regP;
	ctx.regS  = (BYTE)(m_rgbStack - m_rgbMemory);
	ctx.regPC = GetPC();
}

void CExec6502::Reset()
{
	m_regA = 0;
	m_regX = 0;
	m_regY = 0;
	m_regP = Z_FLAG | R_FLAG;

	m_rgbStack = &m_rgbMemory[0x01FF];
	SetPC(ReadWord(M6502_RST_VECTOR));
}

int CExec6502::Irq()
{
	if (m_regP & I_FLAG)
	{
//		m_cMissedInt++;
		return 0;
	}

	PushWord(GetPC());
	PushByte(m_regP);

	m_regP = (m_regP | I_FLAG) & ~D_FLAG;

	SetPC(ReadWord(M6502_IRQ_VECTOR));

	return 7;
}

int CExec6502::Nmi()
{
	PushWord(GetPC());
	PushByte(m_regP);
	m_regP = m_regP & ~D_FLAG;

	SetPC(ReadWord(M6502_NMI_VECTOR));

	return 7;
}

void CExec6502::Break()
{
	PushWord(GetPC());
	PushByte(m_regP | B_FLAG);
	m_regP = (m_regP | I_FLAG) & ~D_FLAG;

	SetPC(ReadWord(M6502_IRQ_VECTOR));
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

//
//  Fetch immediate data from instruction stream
//

inline BYTE CExec6502::ImmedByte()
{
	return *m_rgbOpcode++;
}

inline WORD CExec6502::ImmedWord()
{
	return MAKEWORD(*m_rgbOpcode++, *m_rgbOpcode++);
}

//
//
//

inline BYTE CExec6502::ReadByte(WORD wAddr)
{
	if (m_rgMemDesc[wAddr] & MF_READ_DIRECT)
		return m_rgbMemory[wAddr];

	return CallbackReadByte(wAddr);
}

//#define ReadByte(wAddr) (m_pfnRead(wAddr))

inline WORD CExec6502::ReadWord(WORD wAddr)
{
	if (m_rgMemDesc[wAddr] & MF_READ_DIRECT)
		return MAKEWORD(m_rgbMemory[wAddr++], m_rgbMemory[wAddr]);

	return MAKEWORD(CallbackReadByte(wAddr++), CallbackReadByte(wAddr));
}

//#define ReadWord(wAddr) (MAKEWORD(m_pfnRead(wAddr++), m_pfnRead(wAddr)))

//
//
//

inline void CExec6502::MemWrite(WORD wAddr, BYTE bVal)
{
	if (m_rgMemDesc[wAddr] & MF_WRITE_DIRECT)
		m_rgbMemory[wAddr] = bVal;
	else
		CallbackMemWrite(wAddr, bVal);
}

//
//
//

inline void CExec6502::PushByte(BYTE bVal)
{
	*m_rgbStack-- = bVal;
}

inline void CExec6502::PushWord(WORD wVal)
{
	*m_rgbStack-- = HIBYTE(wVal);
	*m_rgbStack-- = LOBYTE(wVal);
}

inline BYTE CExec6502::PullByte()
{
	return *++m_rgbStack;
}

inline WORD CExec6502::PullWord()
{
	BYTE bl = *++m_rgbStack;
	BYTE bh = *++m_rgbStack;

	return MAKEWORD(bl, bh);
}

inline void CExec6502::SetPC(WORD wAddr)
{
	m_rgbOpcode = &m_rgbMemory[wAddr];
}

inline WORD CExec6502::GetPC()
{
	return (WORD)(m_rgbOpcode - m_rgbMemory);
}

BYTE CExec6502::CallbackReadByte(WORD wAddr)
{
	return m_rgpfnRead[wAddr](wAddr);
}

void CExec6502::CallbackMemWrite(WORD wAddr, BYTE bVal)
{
//	if (m_rgbMemDesc[wAddr] & MF_ADJUST_ADDRESS)
//		m_rgpfnWrite[wAddr](wAddr - , bVal);
	m_rgpfnWrite[wAddr](wAddr, bVal);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

inline BYTE CExec6502::MR_Zx()
{
	return ReadByte(ImmedByte() + m_regX);

	WORD wAddr = ImmedByte() + m_regX;

	if (m_rgMemDesc[wAddr] & MF_READ_DIRECT)
		return m_rgbMemory[wAddr];

	return CallbackReadByte(wAddr);
}

inline BYTE CExec6502::MR_Zy()
{
	return ReadByte(ImmedByte() + m_regY);

	WORD wAddr = ImmedByte() + m_regY;

	if (m_rgMemDesc[wAddr] & MF_READ_DIRECT)
		return m_rgbMemory[wAddr];

	return CallbackReadByte(wAddr);
}

inline BYTE CExec6502::MR_Ax()
{
	return ReadByte(ImmedWord() + m_regX);

	WORD wAddr = ImmedWord() + m_regX;

	if (m_rgMemDesc[wAddr] & MF_READ_DIRECT)
		return m_rgbMemory[wAddr];

	return CallbackReadByte(wAddr);
}

inline BYTE CExec6502::MR_Ay()
{
	return ReadByte(ImmedWord() + m_regY);

	WORD wAddr = ImmedWord() + m_regY;

	if (m_rgMemDesc[wAddr] & MF_READ_DIRECT)
		return m_rgbMemory[wAddr];

	return CallbackReadByte(wAddr);
}

inline BYTE CExec6502::MR_Ix()
{
	return ReadByte(ReadWord(ImmedByte() + m_regX));

	WORD wAddr = ImmedByte() + m_regX;

	if (m_rgMemDesc[wAddr] & MF_READ_DIRECT)
		wAddr = MAKEWORD(m_rgbMemory[wAddr++], m_rgbMemory[wAddr++]);
	else
		wAddr = MAKEWORD(CallbackReadByte(wAddr++), CallbackReadByte(wAddr++));

	if (m_rgMemDesc[wAddr] & MF_READ_DIRECT)
		return m_rgbMemory[wAddr];

	return CallbackReadByte(wAddr);
}

inline BYTE CExec6502::MR_Iy()
{
	return ReadByte(ReadWord(ImmedByte()) + m_regY);

	WORD wAddr = ImmedByte();

	if (m_rgMemDesc[wAddr] & MF_READ_DIRECT)
		wAddr = MAKEWORD(m_rgbMemory[wAddr++], m_rgbMemory[wAddr]);
	else
		wAddr = MAKEWORD(CallbackReadByte(wAddr++), CallbackReadByte(wAddr++));

	wAddr += m_regY;
	
	if (m_rgMemDesc[wAddr] & MF_READ_DIRECT)
		return m_rgbMemory[wAddr];

	return CallbackReadByte(wAddr);
}

inline BYTE CExec6502::MR_Zp()
{
	return ReadByte((WORD)ImmedByte());

	WORD wAddr = ImmedByte();
	
	if (m_rgMemDesc[wAddr] & MF_READ_DIRECT)
		return m_rgbMemory[wAddr];

	return CallbackReadByte(wAddr);
}

inline BYTE CExec6502::MR_Ab()
{
	return ReadByte(ImmedWord());

	WORD wAddr = ImmedWord();
	
	if (m_rgMemDesc[wAddr] & MF_READ_DIRECT)
		return m_rgbMemory[wAddr];

	return CallbackReadByte(wAddr);
}

inline void CExec6502::MW_Zx(BYTE bVal)
{
	MemWrite(ImmedByte() + m_regX, bVal);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

// adjust flags based on A, X, or Y value...
inline void CExec6502::Flags(BYTE b)
{
	m_regP = (m_regP & ~(Z_FLAG | N_FLAG)) | ZNTable[b];
}

#define Flags(b) (m_regP = (m_regP & ~(Z_FLAG | N_FLAG)) | ZNTable[b])

inline void CExec6502::Adc(BYTE bArg)
{
	if (m_regP & D_FLAG)
	{
		BYTE b1 = (m_regA & 0x0F) + (bArg & 0x0F) + (m_regP & C_FLAG);
		BYTE b2 = (m_regA >> 4) + (bArg >> 4) + (b1 > 15 ? 1 : 0);
		
		if(b1 > 9)
		{
			b1 += 6;
			b2++;
		}
		
		if(b2 > 9)
			b2 += 6;
		
		m_regA = (b1 & 0x0F) | (b2 << 4);
		m_regP = (m_regP & ~C_FLAG) | (b2 > 15 ? C_FLAG:0);
	}
	else
	{
#if 0
		BYTE regP = m_regP;
		m_regP &= ~(N_FLAG|V_FLAG|Z_FLAG|C_FLAG);
		m_regP |= _rgbAdcFlags[regP & 1][m_regA][bArg];
		m_regA = _rgbAdcValues[regP & 1][m_regA][bArg];
/*
adc...

flags affected: c,z,n,v
flags unaffected i,d,b,r

       adc sbc cmp bit and ora eor inc dec asl lsr rol ror
c_flag  *   *   *   ?   ?   ?   ?   ?   ?   ?   ?   ?   ?
z_flag  *   *   *   ?   ?   ?   ?   *   *   ?   ?   ?   ?
n_flag  *   *   *   ?   ?   ?   ?   *   *   ?   ?   ?   ?
v_flag  *   *   ?   ?   ?   ?   ?   ?   ?   ?   ?   ?   ?

i_flag - cli, sei,           brk  irq, 
d_flag - cld, sed, adc, sbc, brk, irq, nmi
b_flag -                     brk
r_flag - reset, plp, rti

Flags()...

	m_regP = (m_regP & ~(Z_FLAG | N_FLAG)) | ZNTable[b];

*/		

#else
		WORD w = m_regA + bArg + (m_regP & C_FLAG);

		m_regP &= ~(N_FLAG|V_FLAG|Z_FLAG|C_FLAG);

		m_regP |= (~(m_regA ^ bArg) & (m_regA ^ LOBYTE(w)) & 0x80 ? V_FLAG : 0) |
			(HIBYTE(w) ? C_FLAG : 0) | ZNTable[LOBYTE(w)];

		m_regA = LOBYTE(w);
#endif
	}
}

inline void CExec6502::Sbc(BYTE bArg)
{
/* Warning! C_FLAG is inverted before SBC and after it */
	if(m_regP & D_FLAG)
	{
		BYTE b1 = (m_regA & 0x0F) - (bArg & 0x0F) - (~m_regP & C_FLAG);
		
		if (b1 & 0x10)
			b1 -= 6;
		
		BYTE b2 = (m_regA >> 4) - (bArg >> 4) - (b1 & 0x10);
		
		if (b2 & 0x10)
			b2 -= 6;
		
		m_regA = (b1 & 0x0F) | (b2 << 4);
		m_regP = (m_regP & ~C_FLAG) | (b2 > 15 ? 0 : C_FLAG);
	}
	else
	{
		WORD w = m_regA - bArg - (~m_regP & C_FLAG);
		
		m_regP &= ~(N_FLAG|V_FLAG|Z_FLAG|C_FLAG);
		
		m_regP |= ((m_regA ^ bArg) & (m_regA ^ LOBYTE(w)) & 0x80 ? V_FLAG : 0) |
			(HIBYTE(w) ? 0:C_FLAG) | ZNTable[LOBYTE(w)];

		m_regA = LOBYTE(w);
	}
}

inline void CExec6502::Cmp(BYTE bArg1, BYTE bArg2)
{
	WORD w = bArg1 - bArg2;
	m_regP &= ~(N_FLAG|Z_FLAG|C_FLAG);
	m_regP |= ZNTable[LOBYTE(w)] | (HIBYTE(w) ? 0 : C_FLAG);
}

inline void CExec6502::Bit(BYTE bArg)
{
	m_regP &= ~(N_FLAG | V_FLAG | Z_FLAG);
	m_regP |= (bArg & (N_FLAG | V_FLAG)) | (bArg & m_regA ? 0 : Z_FLAG);
}

inline void CExec6502::And(BYTE bArg) { Flags(m_regA &= bArg); }
inline void CExec6502::Ora(BYTE bArg) { Flags(m_regA |= bArg); }
inline void CExec6502::Eor(BYTE bArg) { Flags(m_regA ^= bArg); }

inline BYTE CExec6502::Asl(BYTE bArg)
{
	m_regP &= ~C_FLAG;
	m_regP |= bArg >> 7;
	Flags(bArg <<= 1);
	return bArg;
}

inline BYTE CExec6502::Lsr(BYTE bArg)
{
	m_regP &= ~C_FLAG;
	m_regP |= bArg & C_FLAG;
	Flags(bArg >>= 1);
	return bArg;
}

inline BYTE CExec6502::Rol(BYTE bArg)
{
	BYTE b = (bArg << 1 ) | (m_regP & C_FLAG);
	m_regP &= ~C_FLAG;
	m_regP |= bArg >> 7;
	bArg = b;
	Flags(bArg);
	return bArg;
}

inline BYTE CExec6502::Ror(BYTE bArg)
{
	BYTE b = (bArg >> 1) | (m_regP << 7);
	m_regP &= ~C_FLAG;
	m_regP |= bArg & C_FLAG;
	bArg = b;
	Flags(bArg);
	return bArg;
}

inline int CExec6502::Jr0(BOOL f)
{
	if (f)
	{
		m_rgbOpcode += ((signed char)ImmedByte());
		return 1;
	}

	m_rgbOpcode++;
	return 0;
}

inline int CExec6502::Jr1(BOOL f)
{
	if (!f)
	{
		m_rgbOpcode += ((signed char)ImmedByte());
		return 1;
	}

	m_rgbOpcode++;
	return 0;
}

inline BYTE CExec6502::Inc(BYTE bArg)
{
	Flags(++bArg);
	return bArg;
}

inline BYTE CExec6502::Dec(BYTE bArg)
{
	Flags(--bArg);
	return bArg;
}

void Illegal()
{
	TRACE("illegal instruction\n");
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

int Exec6502(int cCyclesArg) // non-C++ class version
{
	return 0;
}

/*

// can this be done quickly?...
#define MF_READ_BREAKPOINT  0x00000001
#define MF_WRITE_BREAKPOINT 0x00000002

tracing memory reads and writes...
this is done internally by setting the callback address to an internal
function, this dosen't sacrifice speed when this option is off!

tracing memory access statistics...
it would be nice to get some stats on memory usage, especially during
development work (also, % direct/callback might help performance tuning)

how to use breakpoints...

void foo()
{
	cCycle = 16667;

	while (cCycles > 0)
	{
		int cCycleExec = p6502->Exec(cCycle);

		if (cCycleExec < 0)
		{
			// we hit a breakpoint
			p6502->GetContext(ctx);
		}

		cCycle -= cCycleExec;
	}
}

CExec6502::ExecWithBreakpoints(int *prgi, int c, inc cCycles)
{
	int cCyclesArg = cCycles;
	
	while (cCycles > 0)
	{
		if (GetPC() == ...)
			break;

		cCycles -= Exec(1);
	}

	return (cCyclesArg - cCycles);
}

*/

static const _rgcIns[] = { 0, 100 };

int CExec6502::Exec(int cCyclesArg)
{
	int cCycles = cCyclesArg;

	WORD wAddr;

	while (cCycles > 0) // && m_fExec // set to FALSE by interface calls to break out
	{
#if 0//def _DEBUG
		static cins = 0;

		cins++;

		if (cins >= _rgcIns[0] && cins <= _rgcIns[1]) 
		{
			TCHAR szBuf[64];
			EclDasm6502(szBuf, m_rgbOpcode, GetPC());
			TRACE("%04X %02X %02X %02X %02X %04X %s\n", GetPC(), m_regP, m_regA, m_regX, m_regY, cCycles, szBuf);
		}

		if (GetAsyncKeyState(VK_SHIFT) < 0)
			TRACE("%i\n", cins);
#endif

		const BYTE bOpcode = ImmedByte();
		
		switch (bOpcode)
		{
			case 0x00: // brk
				cCycles -= 7;
				Break();
				break;

			case 0x01: // ora ($ss, x)
				cCycles -= 6;
				Ora(MR_Ix());
				break;

			case 0x02:
				cCycles -= 2;
				Illegal();
				break;

			case 0x03:
				cCycles -= 8;
				Illegal();
				break;

			case 0x04:
				cCycles -= 3;
				Illegal();
				break;

			case 0x05: // ora $ss
				cCycles -= 3;
				Ora(MR_Zp());
				break;

			case 0x06: // asl $ss
				cCycles -= 5;
				wAddr = (WORD)ImmedByte();
				MemWrite(wAddr, Asl(ReadByte(wAddr)));
				break;

			case 0x07:
				cCycles -= 5;
				Illegal();
				break;

			case 0x08: // php
				cCycles -= 3;
				PushByte(m_regP);
				break;

			case 0x09: //  ora #$ss
				cCycles -= 2;
				Ora(ImmedByte());
				break;

			case 0x0A: // asl
				cCycles -= 2;
				m_regA = Asl(m_regA);
				break;

			case 0x0B:
				cCycles -= 2;
				Illegal();
				break;

			case 0x0C:
				cCycles -= 4;
				Illegal();
				break;

			case 0x0D: // ora $ssss
				cCycles -= 4;
				Ora(MR_Ab());
				break;

			case 0x0E: // asl $ssss
				cCycles -= 6;
				wAddr = ImmedWord();
				MemWrite(wAddr, Asl(ReadByte(wAddr)));
				break;

			case 0x0F:
				cCycles -= 6;
				Illegal();
				break;

			case 0x10: // bpl
				cCycles -= 2;
				cCycles -= Jr1(m_regP & N_FLAG);
				break;

			case 0x11: // ora ($ss), y
				cCycles -= 5;
				Ora(MR_Iy());
				break;

			case 0x12:
				cCycles -= 2;
				Illegal();
				break;

			case 0x13:
				cCycles -= 8;
				Illegal();
				break;

			case 0x14:
				cCycles -= 4;
				Illegal();
				break;

			case 0x15: // ora $ss, x
				cCycles -= 4;
				Ora(MR_Zx());
				break;

			case 0x16: // asl $ss, x
				cCycles -= 6;
				wAddr = ImmedByte() + m_regX;
				MemWrite(wAddr, Asl(ReadByte(wAddr)));
				break;

			case 0x17:
				cCycles -= 6;
				Illegal();
				break;

			case 0x18: // clc
				cCycles -= 2;
				m_regP &= ~C_FLAG;
				break;

			case 0x19: // ora $ssss, y
				cCycles -= 4;
				Ora(MR_Ay());
				break;

			case 0x1A:
				cCycles -= 2;
				Illegal();
				break;

			case 0x1B:
				cCycles -= 7;
				Illegal();
				break;

			case 0x1C:
				cCycles -= 5;
				Illegal();
				break;

			case 0x1D: // ora $ssss, x
				cCycles -= 5;
				Ora(MR_Ax());
				break;

			case 0x1E: // asl $ssss, x
				cCycles -= 7;
				wAddr = ImmedWord() + m_regX;
				MemWrite(wAddr, Asl(ReadByte(wAddr)));
				break;

			case 0x1F:
				cCycles -= 7;
				Illegal();
				break;

			case 0x20: // jsr $ssss
				cCycles -= 6;
				wAddr = ImmedWord();
				PushWord(GetPC() - 1);
				m_rgbOpcode = &m_rgbMemory[wAddr];
				break;
				
			case 0x21: // and ($ss, x)
				cCycles -= 6;
				And(MR_Ix());
				break;

			case 0x22:
				cCycles -= 2;
				Illegal();
				break;

			case 0x23:
				cCycles -= 8;
				Illegal();
				break;

			case 0x24: // bit $ss
				cCycles -= 3;
				Bit(MR_Zp());
				break;

			case 0x25: // and $ss
				cCycles -= 3;
				And(MR_Zp());
				break;

			case 0x26: // rol $ss
				cCycles -= 5;
				wAddr = (WORD)ImmedByte();
				MemWrite(wAddr, Rol(ReadByte(wAddr)));
				break;

			case 0x27:
				cCycles -= 5;
				Illegal();
				break;

			case 0x28: // plp
				cCycles -= 4;
				m_regP = PullByte() | R_FLAG;
				break;

			case 0x29: // and #$ss
				cCycles -= 2;
				And(ImmedByte());
				break;

			case 0x2A: // rol
				cCycles -= 2;
				Rol(m_regA);
				break;

			case 0x2B:
				cCycles -= 2;
				Illegal();
				break;

			case 0x2C: // bit $ssss
				cCycles -= 4;
				Bit(MR_Ab());
				break;

			case 0x2D: // and $ssss
				cCycles -= 4;
				And(MR_Ab());
				break;

			case 0x2E: // rol $ssss
				cCycles -= 6;
				wAddr = ImmedWord();
				MemWrite(wAddr, Rol(ReadByte(wAddr)));
				break;

			case 0x2F:
				cCycles -= 6;
				Illegal();
				break;

			case 0x30: // bmi
				cCycles -= 2;
				cCycles -= Jr0(m_regP & N_FLAG);
				break;

			case 0x31: // and ($ss), y
				cCycles -= 5;
				And(MR_Iy());
				break;

			case 0x32:
				cCycles -= 2;
				Illegal();
				break;

			case 0x33:
				cCycles -= 8;
				Illegal();
				break;

			case 0x34:
				cCycles -= 4;
				Illegal();
				break;

			case 0x35: // and $ss, x
				cCycles -= 4;
				And(MR_Zx());
				break;

			case 0x36: // rol $ss, x
				cCycles -= 6;
				wAddr = ImmedByte() + m_regX;
				MemWrite(wAddr, Rol(ReadByte(wAddr)));
				break;

			case 0x37:
				cCycles -= 6;
				Illegal();
				break;

			case 0x38: // sec
				cCycles -= 2;
				m_regP |= C_FLAG;
				break;

			case 0x39: // and $ssss, y
				cCycles -= 4;
				And(MR_Ay());
				break;

			case 0x3A:
				cCycles -= 2;
				Illegal();
				break;

			case 0x3B:
				cCycles -= 7;
				Illegal();
				break;

			case 0x3C:
				cCycles -= 5;
				Illegal();
				break;

			case 0x3D: // and $ssss, x
				cCycles -= 5;
				And(MR_Ax());
				break;				

			case 0x3E: // rol $ssss, x
				cCycles -= 7;
				wAddr = ImmedWord() + m_regX;
				MemWrite(wAddr, Rol(ReadByte(wAddr)));
				break;

			case 0x3F:
				cCycles -= 7;
				Illegal();
				break;

			case 0x40: // rti
				cCycles -= 6;
				m_regP = PullByte() | R_FLAG;
				m_rgbOpcode = &m_rgbMemory[PullWord()];
				break;

			case 0x41: // eor ($ss, x)
				cCycles -= 6;
				Eor(MR_Ix());
				break;

			case 0x42:
				cCycles -= 2;
				Illegal();
				break;

			case 0x43:
				cCycles -= 8;
				Illegal();
				break;

			case 0x44:
				cCycles -= 3;
				Illegal();
				break;

			case 0x45: // eor $ss
				cCycles -= 3;
				Eor(MR_Zp());
				break;

			case 0x46: // lsr $ss
				cCycles -= 5;
				wAddr = (WORD)ImmedByte();
				MemWrite(wAddr, Lsr(ReadByte(wAddr)));
				break;

			case 0x47:
				cCycles -= 5;
				Illegal();
				break;

			case 0x48: // pha
				cCycles -= 3;
				PushByte(m_regA);
				break;

			case 0x49: // eor #$ss
				cCycles -= 2;
				Eor(ImmedByte());
				break;

			case 0x4A: // lsr
				cCycles -= 2;
				m_regA = Lsr(m_regA);
				break;

			case 0x4B:
				cCycles -= 2;
				Illegal();
				break;

			case 0x4C: // jmp $ssss
				cCycles -= 3;
				m_rgbOpcode = &m_rgbMemory[ImmedWord()];
				break;

			case 0x4D: // eor $ssss
				cCycles -= 4;
				Eor(MR_Ab());
				break;

			case 0x4E: // lsr $ssss
				cCycles -= 6;
				wAddr = ImmedWord();
				MemWrite(wAddr, Lsr(ReadByte(wAddr)));
				break;

			case 0x4F:
				cCycles -= 6;
				Illegal();
				break;

			case 0x50: // bvc
				cCycles -= 2;
				cCycles -= Jr1(m_regP & V_FLAG);
				break;

			case 0x51: // eor ($ss), y
				cCycles -= 5;
				Eor(MR_Iy());
				break;

			case 0x52:
				cCycles -= 2;
				Illegal();
				break;

			case 0x53:
				cCycles -= 8;
				Illegal();
				break;

			case 0x54:
				cCycles -= 4;
				Illegal();
				break;

			case 0x55: // eor $ss, x
				cCycles -= 4;
				Eor(MR_Zx());
				break;

			case 0x56: // lsr $ss, x
				cCycles -= 6;
				wAddr = ImmedByte() + m_regX;
				MemWrite(wAddr, Lsr(ReadByte(wAddr)));
				break;

			case 0x57:
				cCycles -= 6;
				Illegal();
				break;

			case 0x58: // cli
				cCycles -= 2;
				m_regP &= ~I_FLAG;
				break;

			case 0x59: // eor $ssss, y
				cCycles -= 4;
				Eor(MR_Ay());
				break;

			case 0x5A:
				cCycles -= 2;
				Illegal();
				break;

			case 0x5B:
				cCycles -= 7;
				Illegal();
				break;

			case 0x5C:
				cCycles -= 5;
				Illegal();
				break;

			case 0x5D: // eor $ssss, x
				cCycles -= 5;
				Eor(MR_Ax());
				break;

			case 0x5E: // asl $ssss, x
				cCycles -= 7;
				wAddr = ImmedWord() + m_regX;
				MemWrite(wAddr, Lsr(ReadByte(wAddr)));
				break;

			case 0x5F:
				cCycles -= 7;
				Illegal();
				break;

			case 0x60: // rts
				cCycles -= 6;
				m_rgbOpcode = &m_rgbMemory[PullWord() + 1];
				break;

			case 0x61: // adc ($ss, x)
				cCycles -= 6;
				Adc(MR_Ix());
				break;

			case 0x62:
				cCycles -= 2;
				Illegal();
				break;

			case 0x63:
				cCycles -= 8;
				Illegal();
				break;

			case 0x64:
				cCycles -= 3;
				Illegal();
				break;

			case 0x65: // adc $ss
				cCycles -= 3;
				Adc(MR_Zp());
				break;

			case 0x66: // ror $ss
				cCycles -= 5;
				wAddr = (WORD)ImmedByte();
				MemWrite(wAddr, Ror(ReadByte(wAddr)));
				break;

			case 0x67:
				cCycles -= 5;
				Illegal();
				break;

			case 0x68: // pla
				cCycles -= 4;
				Flags(m_regA = PullByte());
				break;

			case 0x69: // adc #$ss
				cCycles -= 2;
				Adc(ImmedByte());
				break;

			case 0x6A:
				cCycles -= 2;
				Ror(m_regA);
				break;

			case 0x6B:
				cCycles -= 2;
				Illegal();
				break;

			case 0x6C: // jmp ($ssss)
				cCycles -= 5;
				m_rgbOpcode = &m_rgbMemory[ReadWord(ImmedWord())];
				break;

			case 0x6D: // adc $ssss
				cCycles -= 4;
				Adc(MR_Ab());
				break;

			case 0x6E: // ror $ssss
				cCycles -= 6;
				wAddr = ImmedWord();
				MemWrite(wAddr, Ror(ReadByte(wAddr)));
				break;

			case 0x6F:
				cCycles -= 6;
				Illegal();
				break;

			case 0x70: // bvs
				cCycles -= 2;
				cCycles -= Jr0(m_regP & V_FLAG);
				break;

			case 0x71: // adc ($ss), y
				cCycles -= 5;
				Adc(MR_Iy());
				break;

			case 0x72:
				cCycles -= 2;
				Illegal();
				break;

			case 0x73:
				cCycles -= 8;
				Illegal();
				break;

			case 0x74:
				cCycles -= 4;
				Illegal();
				break;

			case 0x75: // adc $ss, x
				cCycles -= 4;
				Adc(MR_Zx());
				break;

			case 0x76: // ror $ss, x
				cCycles -= 6;
				wAddr = ImmedByte() + m_regX;
				MemWrite(wAddr, Ror(ReadByte(wAddr)));
				break;

			case 0x77:
				cCycles -= 6;
				Illegal();
				break;

			case 0x78: // sei
				cCycles -= 2;
				m_regP |= I_FLAG;
				break;

			case 0x79: // adc $ssss, y
				cCycles -= 4;
				Adc(MR_Ay());
				break;

			case 0x7A:
				cCycles -= 2;
				Illegal();
				break;

			case 0x7B:
				cCycles -= 7;
				Illegal();
				break;

			case 0x7C:
				cCycles -= 5;
				Illegal();
				break;

			case 0x7D: // adc $ssss, x
				cCycles -= 5;
				Adc(MR_Ax());
				break;

			case 0x7E: // asl $ssss, x
				cCycles -= 7;
				wAddr = ImmedWord() + m_regX;
				MemWrite(wAddr, Ror(ReadByte(wAddr)));
				break;

			case 0x7F:
				cCycles -= 7;
				Illegal();
				break;

			case 0x80:
				cCycles -= 2;
				Illegal();
				break;

			case 0x81: // sta ($ss, x)
				cCycles -= 6;
				MemWrite(ReadWord(ImmedByte() + m_regX), m_regA);
				break;

			case 0x82:
				cCycles -= 2;
				Illegal();
				break;

			case 0x83:
				cCycles -= 6;
				Illegal();
				break;

			case 0x84: // sty $ss
				cCycles -= 3;
				MemWrite((WORD)ImmedByte(), m_regY);
				break;

			case 0x85: // sta $ss
				cCycles -= 3;
				MemWrite((WORD)ImmedByte(), m_regA);
				break;

			case 0x86: // stx $ss
				cCycles -= 3;
				MemWrite((WORD)ImmedByte(), m_regX);
				break;

			case 0x87:
				cCycles -= 3;
				Illegal();
				break;

			case 0x88: // dey
				cCycles -= 2;
				Flags(--m_regY);
				break;

			case 0x89:
				cCycles -= 2;
				Illegal();
				break;

			case 0x8A: // txa
				cCycles -= 2;
				Flags(m_regA = m_regX);
				break;

			case 0x8B:
				cCycles -= 2;
				Illegal();
				break;

			case 0x8C: // sty $ssss
				cCycles -= 4;
				MemWrite(ImmedWord(), m_regY);
				break;

			case 0x8D: // sta $ssss
				cCycles -= 4;
				MemWrite(ImmedWord(), m_regA);
				break;

			case 0x8E: // stc $ssss
				cCycles -= 4;
				MemWrite(ImmedWord(), m_regX);
				break;

			case 0x8F:
				cCycles -= 4;
				Illegal();
				break;

			case 0x90: // bcc
				cCycles -= 2;
				cCycles -= Jr1(m_regP & C_FLAG);
				break;

			case 0x91: // sta ($ss), y
				cCycles -= 6;
				MemWrite(ReadWord((WORD)ImmedByte()) + m_regY, m_regA);
				break;

			case 0x92:
				cCycles -= 2;
				Illegal();
				break;

			case 0x93:
				cCycles -= 6;
				Illegal();
				break;

			case 0x94: // sty $ss, x
				cCycles -= 4;
				MW_Zx(m_regY);
				break;

			case 0x95: // sta $ss, x
				cCycles -= 4;
				MW_Zx(m_regA);
				break;

			case 0x96: // stx $ss, y
				cCycles -= 4;
				MemWrite(ImmedByte() + m_regY, m_regX);
				break;

			case 0x97:
				cCycles -= 4;
				Illegal();
				break;

			case 0x98:
				cCycles -= 2;
				Flags(m_regA = m_regY);
				break;

			case 0x99: // sta $ssss, y
				cCycles -= 5;
				MemWrite(ImmedWord() + m_regY, m_regA);
				break;

			case 0x9A: // txs
				cCycles -= 2;
				m_rgbStack = &m_rgbMemory[0x0100 | m_regX];
				break;

			case 0x9B:
				cCycles -= 5;
				Illegal();
				break;

			case 0x9C:
				cCycles -= 5;
				Illegal();
				break;

			case 0x9D: // sta $ssss, x
				cCycles -= 5;
				MemWrite(ImmedWord() + m_regX, m_regA);
				break;

			case 0x9E:
				cCycles -= 5;
				Illegal();
				break;

			case 0x9F:
				cCycles -= 5;
				Illegal();
				break;

			case 0xA0: // ldy #$ss
				cCycles -= 2;
				Flags(m_regY = ImmedByte());
				break;

			case 0xA1: // lda ($ss, x)
				cCycles -= 6;
				Flags(m_regA = MR_Ix());
				break;

			case 0xA2: // ldx #$ss
				cCycles -= 2;
				Flags(m_regX = ImmedByte());
				break;

			case 0xA3:
				cCycles -= 6;
				Illegal();
				break;

			case 0xA4: // ldy $ss
				cCycles -= 3;
				Flags(m_regY = MR_Zp());
				break;

			case 0xA5: // lda $ss
				cCycles -= 3;
				Flags(m_regA = MR_Zp());
				break;

			case 0xA6: // ldx $ss
				cCycles -= 3;
				Flags(m_regX = MR_Zp());
				break;

			case 0xA7:
				cCycles -= 3;
				Illegal();
				break;

			case 0xA8: // tay
				cCycles -= 2;
				Flags(m_regY = m_regA);
				break;

			case 0xA9: // lda #$ss
				cCycles -= 2;
				Flags(m_regA = ImmedByte());
				break;

			case 0xAA: // tax
				cCycles -= 2;
				Flags(m_regX = m_regA);
				break;

			case 0xAB:
				cCycles -= 2;
				Illegal();
				break;

			case 0xAC: // ldy $ssss
				cCycles -= 4;
				Flags(m_regY = MR_Ab());
				break;

			case 0xAD: // lda $ssss
				cCycles -= 4;
				Flags(m_regA = MR_Ab());
				break;

			case 0xAE: // ldx $ssss
				cCycles -= 4;
				Flags(m_regX = MR_Ab());
				break;

			case 0xAF:
				cCycles -= 4;
				Illegal();
				break;

			case 0xB0: // bcs
				cCycles -= 2;
				cCycles -= Jr0(m_regP & C_FLAG);
				break;

			case 0xB1: // lda ($ss), y
				cCycles -= 5;
				Flags(m_regA = MR_Iy());
				break;

			case 0xB2:
				cCycles -= 2;
				Illegal();
				break;

			case 0xB3:
				cCycles -= 5;
				Illegal();
				break;

			case 0xB4: // ldy $ss, x
				cCycles -= 4;
				Flags(m_regY = MR_Zx());
				break;

			case 0xB5: // lda $ss, x
				cCycles -= 4;
				Flags(m_regA = MR_Zx());
				break;

			case 0xB6: // ldx $ss, y
				cCycles -= 4;
				Flags(m_regX = MR_Zx());
				break;

			case 0xB7:
				cCycles -= 4;
				Illegal();
				break;

			case 0xB8: // clv
				cCycles -= 2;
				m_regP &= ~V_FLAG;
				break;

			case 0xB9: // lda $ssss, y
				cCycles -= 4;
				Flags(m_regA = MR_Ay());
				break;
					  
			case 0xBA: // tsx
				cCycles -= 2;
				m_regX = BYTE(m_rgbStack - m_rgbMemory[0x0100]);
				break;

			case 0xBB:
				cCycles -= 5;
				Illegal();
				break;

			case 0xBC: // ldy $ssss, x
				cCycles -= 4;
				Flags(m_regY = MR_Ax());
				break;

			case 0xBD: // lda $ssss, x
				cCycles -= 4;
				Flags(m_regA = MR_Ax());
				break;

			case 0xBE: // ldx $ssss, y
				cCycles -= 4;
				Flags(m_regX = MR_Ay());
				break;

			case 0xBF:
				cCycles -= 4;
				Illegal();
				break;

			case 0xC0: // cpy #$ss
				cCycles -= 2;
				Cmp(m_regY, ImmedByte());
				break;

			case 0xC1: // cmp ($ss, x)
				cCycles -= 6;
				Cmp(m_regA, MR_Ix());
				break;

			case 0xC2:
				cCycles -= 2;
				Illegal();
				break;

			case 0xC3:
				cCycles -= 8;
				Illegal();
				break;

			case 0xC4: // cpy $ss
				cCycles -= 3;
				Cmp(m_regY, MR_Zp());
				break;

			case 0xC5: // cmp $ss
				cCycles -= 3;
				Cmp(m_regA, MR_Zp());
				break;

			case 0xC6: // dec $ss
				cCycles -= 5;
				wAddr = (WORD)ImmedByte();
				MemWrite(wAddr, Dec(ReadByte(wAddr)));
				break;

			case 0xC7:
				cCycles -= 5;
				Illegal();
				break;

			case 0xC8: // iny
				cCycles -= 2;
				m_regY = Inc(m_regY);
				break;

			case 0xC9: // cmp #$ss
				cCycles -= 2;
				Cmp(m_regA, ImmedByte());
				break;

			case 0xCA: // dex
				cCycles -= 2;
				m_regX = Dec(m_regX);
				break;

			case 0xCB:
				cCycles -= 2;
				Illegal();
				break;

			case 0xCC: // cpy $ssss
				cCycles -= 4;
				Cmp(m_regY, MR_Ab());
				break;

			case 0xCD: // cmp $ssss
				cCycles -= 4;
				Cmp(m_regA, MR_Ab());
				break;

			case 0xCE: // ror $ssss
				cCycles -= 6;
				wAddr = ImmedWord();
				MemWrite(wAddr, Dec(ReadByte(wAddr)));
				break;

			case 0xCF:
				cCycles -= 6;
				Illegal();
				break;

			case 0xD0: // bne
				cCycles -= 2;
				cCycles -= Jr1(m_regP & Z_FLAG);
				break;

			case 0xD1: // cmp ($ss), y
				cCycles -= 5;
				Cmp(m_regA, MR_Iy());
				break;

			case 0xD2:
				cCycles -= 2;
				Illegal();
				break;

			case 0xD3:
				cCycles -= 8;
				Illegal();
				break;

			case 0xD4:
				cCycles -= 4;
				Illegal();
				break;

			case 0xD5: // cmp $ss, x
				cCycles -= 4;
				Cmp(m_regA, MR_Zx());
				break;

			case 0xD6: // dec $ss, x
				cCycles -= 6;
				wAddr = ImmedByte() + m_regX;
				MemWrite(wAddr, Dec(ReadByte(wAddr)));
				break;

			case 0xD7:
				cCycles -= 6;
				Illegal();
				break;

			case 0xD8: // cld
				cCycles -= 2;
				m_regP &= ~D_FLAG;
				break;
				
			case 0xD9: // cmp $ssss, y
				cCycles -= 4;
				Cmp(m_regA, MR_Ay());
				break;

			case 0xDA:
				cCycles -= 2;
				Illegal();
				break;

			case 0xDB:
				cCycles -= 7;
				Illegal();
				break;

			case 0xDC:
				cCycles -= 5;
				Illegal();
				break;

			case 0xDD: // cmp $ssss, x
				cCycles -= 5;
				Cmp(m_regA, MR_Ax());
				break;

			case 0xDE: // dec $ssss, x
				cCycles -= 7;
				wAddr = ImmedWord() + m_regX;
				MemWrite(wAddr, Dec(ReadByte(wAddr)));
				break;

			case 0xDF:
				cCycles -= 7;
				Illegal();
				break;

			case 0xE0: // cpx #$ss
				cCycles -= 2;
				Cmp(m_regX, ImmedByte());
				break;

			case 0xE1: // sbc ($ss, x)
				cCycles -= 6;
				Sbc(MR_Ix());
				break;

			case 0xE2:
				cCycles -= 2;
				Illegal();
				break;

			case 0xE3:
				cCycles -= 8;
				Illegal();
				break;

			case 0xE4: // cpx $ss
				cCycles -= 3;
				Cmp(m_regX, MR_Zp());
				break;

			case 0xE5: // sbc $ss
				cCycles -= 3;
				Sbc(MR_Zp());
				break;

			case 0xE6: // inc $ss
				cCycles -= 5;
				wAddr = (WORD)ImmedByte();
				MemWrite(wAddr, Inc(ReadByte(wAddr)));
				break;

			case 0xE7:
				cCycles -= 5;
				Illegal();
				break;

			case 0xE8: // inx
				cCycles -= 2;
				Flags(++m_regX);
				break;

			case 0xE9: // sbc #$ss
				cCycles -= 2;
				Sbc(ImmedByte());
				break;

			case 0xEA: // nop
				cCycles -= 2;
				break;

			case 0xEB:
				cCycles -= 2;
				Illegal();
				break;

			case 0xEC: // cpx $ssss
				cCycles -= 4;
				Cmp(m_regX, MR_Ab());
				break;

			case 0xED: // sbc $ssss
				cCycles -= 4;
				Sbc(MR_Ab());
				break;

			case 0xEE: // inc $ssss
				cCycles -= 6;
				wAddr = ImmedWord();
				MemWrite(wAddr, Inc(ReadByte(wAddr)));
				break;

			case 0xEF:
				cCycles -= 6;
				Illegal();
				break;

			case 0xF0: // beq
				cCycles -= 2;
				cCycles -= Jr0(m_regP & Z_FLAG);
				break;

			case 0xF1: // sbc ($ss), y
				cCycles -= 5;
				Sbc(MR_Iy());
				break;

			case 0xF2:
				cCycles -= 2;
				Illegal();
				break;

			case 0xF3:
				cCycles -= 8;
				Illegal();
				break;

			case 0xF4:
				cCycles -= 4;
				Illegal();
				break;

			case 0xF5: // sbc $ss, x
				cCycles -= 4;
				Sbc(MR_Zx());
				break;

			case 0xF6: // inc $ss, x
				cCycles -= 6;
				wAddr = ImmedByte() + m_regX;
				MemWrite(wAddr, Inc(ReadByte(wAddr)));
				break;

			case 0xF7:
				cCycles -= 6;
				Illegal();
				break;

			case 0xF8: // sed
				cCycles -= 2;
				m_regP |= D_FLAG;
				break;

			case 0xF9: // sbc $ssss, y
				cCycles -= 4;
				Sbc(MR_Ay());
				break;				

			case 0xFA:
				cCycles -= 2;
				Illegal();
				break;

			case 0xFB:
				cCycles -= 7;
				Illegal();
				break;

			case 0xFC:
				cCycles -= 5;
				Illegal();
				break;

			case 0xFD: // sbc $ssss, x
				cCycles -= 5;
				Sbc(MR_Ax());
				break;

			case 0xFE: // inc $ssss, x
				cCycles -= 7;
				wAddr = ImmedWord() + m_regX;
				MemWrite(wAddr, Inc(ReadByte(wAddr)));
				break;

			case 0xFF:
				cCycles -= 7;
				Illegal();
				break;

			//
			//  unhandled instruction
			//
			
			default:
			{
				_ASSERT(FALSE);
				break;
			}
		}
	}

	return (cCyclesArg - cCycles);
}

