/*---------------------------------------------------------------------------
	
	File:
		sound.cpp
	
	Abstract:
		Sound generation code for MageX

	Revisions:
		10-May-97 EAM Created

	Copyright (C) 1997 Edward Massey

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"

extern "C" void TRACE(LPCTSTR pszFormat, ...);

struct WAVE : WAVEHDR
{
	BYTE rgbData[512];
	int cbData;
};

struct MM_SOUND_STATE
{
	int cWave;
	int cbData;
	HWAVEOUT hWaveOut;
	WAVE rgWave[6]; // pool of wave out buffers
	WAVE *pWaveActive;
};

static MM_SOUND_STATE _mms;

HRESULT SoundActivate(HWND hwnd, BOOL fActivate)
{
	_mms.cWave = 6;
	_mms.cbData = 512;

	if (fActivate)
	{
		_ASSERT(_mms.hWaveOut == NULL);

		//
		//
		//

		MMRESULT nResult;
		WAVEFORMATEX wf;
		
		wf.wFormatTag      = WAVE_FORMAT_PCM;
		wf.nChannels       = 1;
		wf.nSamplesPerSec  = 11025; // 22050; // 44100;
		wf.wBitsPerSample  = 8;
		wf.nBlockAlign     = (wf.wBitsPerSample / 8) * wf.nChannels;
		wf.nAvgBytesPerSec = wf.nSamplesPerSec * wf.nBlockAlign;
		wf.cbSize          = sizeof(wf);

		nResult = waveOutOpen(&_mms.hWaveOut, WAVE_MAPPER,
			&wf, NULL, NULL, 0);

		if (nResult == MMSYSERR_NOERROR)
		{
			//
			//
			//

			_ASSERT(_mms.cWave <= DIM(_mms.rgWave));
			_ASSERT(_mms.cbData <= FIELD_SIZEOF(WAVE, rgbData));

			for (int i = 0; i < _mms.cWave; i++)
			{
				ZERO_INIT(_mms.rgWave[i]);

				_mms.rgWave[i].lpData = (LPSTR)_mms.rgWave[i].rgbData;
				_mms.rgWave[i].dwBufferLength = _mms.cbData;

				nResult = waveOutPrepareHeader(_mms.hWaveOut, &_mms.rgWave[i],
					sizeof(WAVEHDR));
					
				// mark header as available
				_mms.rgWave[i].dwFlags |= WHDR_DONE;

				_mms.rgWave[i].cbData = 0;

				if (nResult != MMSYSERR_NOERROR)
					break;
			}
		}

		//
		//
		//
		
		if (nResult != MMSYSERR_NOERROR)
		{
			TCHAR szMsg[256];
			
			waveOutGetErrorText(nResult, szMsg, DIM(szMsg));

			MessageBox(hwnd, szMsg, TEXT("MageX - waveOutErrorText"), MB_ICONINFORMATION);
		}
		else
		{
			fActivate = TRUE;
		}
	}

	//
	//
	//

	if (!fActivate)
	{
		_ASSERT(_mms.hWaveOut != NULL);

		waveOutReset(_mms.hWaveOut);

		for (int i = 0; i < _mms.cWave; i++)
			waveOutUnprepareHeader(_mms.hWaveOut, &_mms.rgWave[i], sizeof(WAVEHDR));

		waveOutClose(_mms.hWaveOut);

		_mms.hWaveOut = NULL;
	}

	return S_OK;
}

#ifdef _DEBUG

static void TestUpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
	static const BYTE samples[][32] =
	{
		0x00,0x00,0x00,0x00,0x22,0x22,0x22,0x22,0x44,0x44,0x44,0x44,0x22,0x22,0x22,0x22,
		0x00,0x00,0x00,0x00,0xdd,0xdd,0xdd,0xdd,0xbb,0xbb,0xbb,0xbb,0xdd,0xdd,0xdd,0xdd
	};

	static i = 0;

	while (cbBuffer--)
	{
		*rgbBuffer++ = samples[0][(i++ % 32)];
	}	
}

#endif

inline WAVE *FindAvailableWave()
{
	for (int i = 0; i < _mms.cWave; i++)
	{
		if (_mms.rgWave[i].dwFlags & WHDR_DONE)
			return &_mms.rgWave[i];
	}

	return NULL;
}

void SoundPump(void (*pfnUpdateSound)(LPBYTE rgbBuffer, LONG cbBuffer))
{
	if (_mms.hWaveOut == NULL)
		return;
	
	if (pfnUpdateSound == NULL)
		return;

	if (_mms.pWaveActive == NULL)
		_mms.pWaveActive = FindAvailableWave();

	if (_mms.pWaveActive != NULL)
	{
//		if (GetAsyncKeyState(VK_SHIFT) < 0)
//			pfnUpdateSound = TestUpdateSound;

		WAVE &wave = *_mms.pWaveActive;

		pfnUpdateSound(&wave.rgbData[wave.cbData], 256);
//		pfnUpdateSound(&wave.rgbData[wave.cbData], _mms.cbData);

		wave.cbData += 256;

		if (wave.cbData >= 512)
		{
			waveOutWrite(_mms.hWaveOut, &wave, sizeof(WAVEHDR));
		
			wave.cbData = 0;

			_mms.pWaveActive = NULL;
		}
	}

//	_mms.pWaveActive = FindAvailableWave();
}

void SoundDraw(LPBYTE rgbBuffer, LONG lPitch)
{
/*
	if (_mms.hWaveOut != NULL)
	{
		for (int i = 0; i < DIM(_mms.rgWave[0].rgbData); i++)
		{
			rgbBuffer[i/2 + lPitch * _mms.rgWave[0].rgbData[i]] = 255;
		}
	}
*/
}

struct CHANNEL
{
	void SetEnvelope(const char *pData, int cSamples, BOOL fLoop);

	DWORD m_dwFlags;
	int m_iCurrent;
	int m_cSamples;
	const char *m_pData; // points to begining of buffer
};

#define CHANNEL_ENABLED 0x00000001
#define CHANNEL_LOOP    0x00000002

static CHANNEL _rgChannel[8];

void SoundSetEnvelope(int iChnl, const char *pData, int cSamples, BOOL fLoop)
{
	_rgChannel[iChnl].SetEnvelope(pData, cSamples, fLoop);
}

void CHANNEL::SetEnvelope(const char *pData, int cSamples, BOOL fLoop)
{
	m_dwFlags = 0;

	if (fLoop)
		m_dwFlags |= CHANNEL_LOOP;

	m_pData = pData;

	if (m_pData != NULL)
		m_dwFlags |= CHANNEL_ENABLED;

	m_iCurrent = 0;

	m_cSamples = cSamples;
}

void SoundSetChannelVolume(int iChnl, int nVolume);
void SoundSetChannelPitch(int iChnl, int nPitch);
void SoundSetChannelEnable(int iChnl, BOOL fEnable);

// mix 8 bit mono sound...

// CONSIDER:  invert the loops might give performance boost??? (put channel on outer loop, buffer on inner)

void SoundUpdateMixer(LPBYTE rgbBuffer, LONG cbBuffer)
{
	while (cbBuffer--)
	{
		//
		//  initialize accumulator (destination buffer is unsigned)
		//
		
		int nAcc = 128;

		//
		//  loop through each channel
		//
		
		CHANNEL *pChnl = _rgChannel;
		CHANNEL *pChnlEnd = &_rgChannel[DIM(_rgChannel)];

		for (pChnl = _rgChannel; pChnl < pChnlEnd; pChnl++)
		{
			if ((pChnl->m_dwFlags & CHANNEL_ENABLED) == 0)
				continue;

			nAcc += pChnl->m_pData[pChnl->m_iCurrent];

			if (pChnl->m_iCurrent++ >= pChnl->m_cSamples)
			{
				if (pChnl->m_dwFlags & CHANNEL_LOOP)
					pChnl->m_iCurrent = 0;
				else
					pChnl->m_dwFlags &= ~CHANNEL_ENABLED;
			}


/*
			nAcc += ch.pData[ch.iPos >> 12] * ch.nVol;

			ch.iPos += ch.cInc;

			if (ch.iPos > (32 << 12))
				ch.iPos -= (32 << 12);
*/
		}

		//
		//  clip...
		//
		
		if (nAcc < 0)
			nAcc = 0;
		else if (nAcc > 255)
			nAcc = 255;

		//
		//
		//

		*rgbBuffer++ = nAcc;
	}	
}

#if 0

void SoundUpdateMixer2(LPBYTE rgbBuffer, LONG cbBuffer)
{
	int rgnBuffer[2048];

	_ASSERT(cbBuffer < DIM(rgnBuffer));

	ZERO_INIT(rgnBuffer);

	CHANNEL *pChnl = _rgChannel;
	CHANNEL *pChnlEnd = &_rgChannel[DIM(_rgChannel)];

	int *pnBuf = rgnBuffer;

	for (pChnl = _rgChannel; pChnl < pChnlEnd; pChnl++)
	{
		if ((pChnl->m_dwFlags & CHANNEL_ENABLED) == 0)
			continue;

		*pnBuf++ += pChnl->m_pData[pChnl->m_iCurrent];

		if (pChnl->m_iCurrent++ >= pChnl->m_cSamples)
		{
			if (pChnl->m_dwFlags & CHANNEL_LOOP)
				pChnl->m_iCurrent = 0;
			else
				pChnl->m_dwFlags &= ~CHANNEL_ENABLED;
		}
	}

	while (cbBuffer--)
	{
		*rgbBuffer++ = rgbMix[*pnBuf++];
	}
}

#endif

/////////////////////////////////////////////////////////////

/*

mixer code for MageX

supports unlimited number of channels.  Each channel has an envelope,
volume (l/r), and frequency.

This code is designed to support sound chip synthesis and pre-calculated
sample playback for the game drivers in MageX.

rate - sampling rate (Hz)
freq - target frequency (Hz)
incr - increment to step through sample
leng - number of samples in sample

incr = freq * (leng / rate)

The existing 8910, SN76496, (and POKEY?) sound code will be rewritten to make use
of these services.  While I don't expect any POKEY performance boost,
the 8910 code could be much more efficient.

Frequency may be viewed as...
  frequency of the tone the user hears (channel must have base freq specified w/sample)
  frequency that each sample in a channel is played
  something like 1000 = original speed, 2000 = double speed
  fixed-point index increment

Assuming a 16-bit fixed point for simplicity...

00010000 -> play sample at original speed
00020000 -> play sample at twice the original frequency
00008000 -> play sample at half the original frequency

*/

static ULONG _lRate = 11025;

#define FIXED_POINT_SHIFT 12

struct CHANNEL_NEW
{
	CHANNEL_NEW();

	void SetFrequency(int nFrequency);
	void SetEnvelope(const BYTE *rgbSamples, int cSamples, int nFreqBase);
	void SetVolume(int nVol);

	void Mix(signed long *rgbBuffer, int cSamples);

private:
	DWORD m_dwFlags;
	DWORD m_dwPos;
	DWORD m_dwLen;
	DWORD m_dwInc;
	DWORD m_dwVolL, m_dwVolR;
	DWORD m_dwFreqBase;
	DWORD m_dwFreq;
	DWORD m_dwMul;
	const BYTE *m_rgbSamples;
};

void CHANNEL_NEW::SetEnvelope(const BYTE *rgbSamples, int cSamples, int nFreqBase)
{
	m_dwLen = cSamples << FIXED_POINT_SHIFT;
	m_dwMul = m_dwLen / _lRate;
	m_rgbSamples = rgbSamples;
	m_dwFreqBase = nFreqBase;

	m_dwFlags = 0;

	BOOL fLoop = TRUE;

	if (fLoop)
		m_dwFlags |= CHANNEL_LOOP;

	if (m_rgbSamples != NULL)
		m_dwFlags |= CHANNEL_ENABLED;

	m_dwPos = 0;

	SetFrequency(0);
}

void CHANNEL_NEW::SetFrequency(int nFreq)
{
	m_dwInc = nFreq * m_dwMul;

	m_dwFlags &= ~CHANNEL_ENABLED;

	if (m_dwInc != 0)
		m_dwFlags |= CHANNEL_ENABLED;

	m_dwFreq = nFreq;
}

void CHANNEL_NEW::SetVolume(int nVol)
{
	m_dwVolL = m_dwVolR = nVol;

	m_dwFlags &= ~CHANNEL_ENABLED;

	if (m_dwVolL != 0)
		m_dwFlags |= CHANNEL_ENABLED;
}

void CHANNEL_NEW::Mix(signed long *rgbBuffer, int cSamples)
{
}

/*---------------------------------------------------------------------------

	Sound Synthesis for MageX

	Two strategies for mixing...

	1.  loop over output buffer in outer loop, channels in inner, render
		directly to output buffer

	2.  loop over channels in outer loop, render into secondary buffer, then
		clip into final buffer

	Optimizations:

	Clipping lookup table
	Volume lookup table

	Features:

	Multi-channel mixer
	Object oriented interface
	Master volume
	Each channel - volume, envelope, panning, freq
	Interpolation
	Up to XXX channels
	Noise generator

	Object-oriented emulation of AY-3-8910 and POKEY sound chips

	Goals for MageX sound:
	
	Make the mixer module generic enough to support Donkey Kong (long, fixed
	pitch, fixed volume samples), Pacman (short, variable pitch, variable
	volume, samples/envelopes), Time Pilot (AY-3-8910), Centipede (POKEY),
	Mr Do/Ladybug

	The 8910 and POKEY code (from MAME) will be rewritten to take advantage
	of the mixer services.  These sources currently do there own mixing.

	References:

	WaveMix by Microsoft
	Allegro 2.2 by Shawn ???
	PokeySound by Ron Fries
	AY-3-8910 programable sound generator by Ville Hallik

---------------------------------------------------------------------------*/

struct PAC_SOUND_CHANNEL
{
	int nFreq;
	int nVol;
	int nEnv;
	const BYTE *pData;
	int iPos;
	int cInc;
};

void PacmanUpdateSound(LPBYTE rgbBuffer, LONG cbBuffer, const BYTE *RAM,
	BOOL &fSoundChanged, const BYTE *rgbEnvelopes)
{
	//
	//
	//

	static PAC_SOUND_CHANNEL rgChannel[3];

	if (fSoundChanged)
	{
		fSoundChanged = FALSE;

		for (int iv = 0; iv < 3; iv++)
		{
			PAC_SOUND_CHANNEL &ch = rgChannel[iv];

			ch.nFreq =                 RAM[0x14 + 5 * iv];
			ch.nFreq = ch.nFreq * 16 + RAM[0x13 + 5 * iv];
			ch.nFreq = ch.nFreq * 16 + RAM[0x12 + 5 * iv];
			ch.nFreq = ch.nFreq * 16 + RAM[0x11 + 5 * iv];

			if (iv == 0)
				ch.nFreq = ch.nFreq * 16 + RAM[0x10 + 5 * iv];
			else
				ch.nFreq = ch.nFreq * 16;

			ch.nVol = RAM[0x15 + 5 * iv];

			ch.nEnv = RAM[0x05 + 5 * iv] & 0x07;

			ch.pData = &rgbEnvelopes[ch.nEnv * 32];

			ch.cInc = ch.nFreq % (32 << 12);
		}
	}

	//
	//
	//
	
	while (cbBuffer--)
	{
		//
		//  initialize accumulator (destination buffer is unsigned)
		//
		
		int nAcc = 128;

		//
		//  loop through each channel
		//

		for (int iv = 0; iv < 3; iv++)
		{
			PAC_SOUND_CHANNEL &ch = rgChannel[iv];
	
			if (ch.nVol != 0)
			{
				nAcc += ch.pData[ch.iPos >> 12] * ch.nVol;

				ch.iPos += ch.cInc;

				if (ch.iPos > (32 << 12))
					ch.iPos -= (32 << 12);
			}
		}

		//
		//  normalize and clip...
		//
		
		nAcc /= 16 * 3;
		
		if (nAcc < 0)
			nAcc = 0;
		else if (nAcc > 255)
			nAcc = 255;

		//
		//
		//

		*rgbBuffer++ = nAcc;
	}
}

static const BYTE _rgbPacmanEnvelopes[8][32] =
{
	{ 0xff, 0x11, 0x22, 0x33, 0x44, 0x55, 0x55, 0x66, 0x66, 0x66, 0x55, 0x55, 0x44, 0x33, 0x22, 0x11, 0xff, 0xdd, 0xcc, 0xbb, 0xaa, 0x99, 0x99, 0x88, 0x88, 0x88, 0x99, 0x99, 0xaa, 0xbb, 0xcc, 0xdd },
	{ 0xff, 0x44, 0x66, 0x66, 0x55, 0x33, 0x11, 0x22, 0x33, 0x33, 0x22, 0x11, 0xee, 0xcc, 0xbb, 0xdd, 0xff, 0x11, 0x33, 0x22, 0x00, 0xdd, 0xcc, 0xbb, 0xbb, 0xcc, 0xdd, 0xbb, 0x99, 0x88, 0x88, 0xaa },
	{ 0xff, 0x22, 0x44, 0x55, 0x66, 0x55, 0x44, 0x22, 0xff, 0xcc, 0xaa, 0x99, 0x88, 0x99, 0xaa, 0xcc, 0xff, 0x33, 0x55, 0x66, 0x55, 0x33, 0xff, 0xbb, 0x99, 0x88, 0x99, 0xbb, 0xff, 0x66, 0xff, 0x88 },
	{ 0xff, 0x55, 0x33, 0x00, 0x33, 0x55, 0x11, 0xee, 0x33, 0x66, 0x44, 0xff, 0x11, 0x22, 0xee, 0xaa, 0xff, 0x44, 0x00, 0xcc, 0xdd, 0xff, 0xaa, 0x88, 0xbb, 0x00, 0xdd, 0x99, 0xbb, 0xee, 0xbb, 0x99 },
	{ 0x88, 0x00, 0x77, 0xff, 0x99, 0x00, 0x66, 0xff, 0xaa, 0x00, 0x55, 0xff, 0xbb, 0x00, 0x44, 0xff, 0xcc, 0x00, 0x33, 0xff, 0xdd, 0x00, 0x22, 0xff, 0xee, 0x00, 0x11, 0xff, 0xff, 0x00, 0x00, 0xff },
	{ 0xff, 0x00, 0xee, 0x11, 0xdd, 0x22, 0xcc, 0x33, 0xbb, 0x44, 0xaa, 0x55, 0x99, 0x66, 0x88, 0x77, 0x88, 0x77, 0x99, 0x66, 0xaa, 0x55, 0xbb, 0x44, 0xcc, 0x33, 0xdd, 0x22, 0xee, 0x11, 0xff, 0x00 },
	{ 0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff, 0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x77, 0x66, 0x55, 0x44, 0x33, 0x22, 0x11, 0x00, 0xff, 0xee, 0xdd, 0xcc, 0xbb, 0xaa, 0x99, 0x88 },
	{ 0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff, 0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff, 0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77 }
};

void PacmanUpdateSound(LPBYTE rgbBuffer, LONG cbBuffer, const BYTE *RAM,
	BOOL &fSoundChanged)
{
	PacmanUpdateSound(rgbBuffer, cbBuffer, RAM, fSoundChanged, _rgbPacmanEnvelopes[0]);
}

static const BYTE _rgbEnvelopes[8][32] =
{
	{ 0x77, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xDD, 0xEE, 0xEE, 0xEE, 0xDD, 0xDD, 0xCC, 0xBB, 0xAA, 0x99, 0x77, 0x55, 0x44, 0x33, 0x22, 0x11, 0x11, 0x00, 0x00, 0x00, 0x11, 0x11, 0x22, 0x33, 0x44, 0x55 },
	{ 0x77, 0x99, 0xAA, 0xBB, 0x77, 0xDD, 0xDD, 0x77, 0xEE, 0x77, 0xDD, 0xDD, 0x77, 0xBB, 0xAA, 0x99, 0x77, 0x55, 0x77, 0x33, 0x77, 0x11, 0x77, 0x00, 0x77, 0x00, 0x77, 0x11, 0x77, 0x33, 0x77, 0x55 },
	{ 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xBB, 0xDD, 0xEE, 0xDD, 0xCC, 0xAA, 0x88, 0x88, 0x88, 0xAA, 0xCC, 0xDD, 0xEE, 0xDD, 0xBB, 0x88, 0x44, 0x22, 0x11, 0x22, 0x33, 0x55, 0x77, 0x77, 0x77, 0x55, 0x33, 0x22, 0x11, 0x22, 0x44, 0x77 },
	{ 0x77, 0xAA, 0xCC, 0xDD, 0xEE, 0xDD, 0xCC, 0xAA, 0x77, 0x44, 0x22, 0x11, 0x00, 0x11, 0x22, 0x44, 0x77, 0xBB, 0xDD, 0xEE, 0xDD, 0xBB, 0x77, 0x33, 0x11, 0x00, 0x11, 0x33, 0x77, 0xEE, 0x77, 0x00 },
	{ 0x77, 0xEE, 0xCC, 0x99, 0xCC, 0xEE, 0xAA, 0x77, 0xCC, 0xFF, 0xDD, 0x88, 0xAA, 0xBB, 0x77, 0x22, 0x88, 0xDD, 0x99, 0x44, 0x55, 0x77, 0x22, 0x00, 0x33, 0x88, 0x55, 0x11, 0x33, 0x66, 0x33, 0x11 },
	{ 0x77, 0x88, 0xAA, 0xCC, 0xEE, 0xDD, 0xCC, 0xCC, 0xBB, 0xAA, 0x88, 0x77, 0x55, 0x66, 0x77, 0x88, 0x88, 0x99, 0xAA, 0xBB, 0x99, 0x88, 0x66, 0x55, 0x44, 0x44, 0x33, 0x22, 0x44, 0x66, 0x88, 0x99 },
	{ 0xAA, 0xCC, 0xCC, 0xAA, 0x77, 0x77, 0x88, 0xBB, 0xDD, 0xEE, 0xDD, 0xAA, 0x66, 0x55, 0x55, 0x77, 0x99, 0x99, 0x88, 0x44, 0x11, 0x00, 0x11, 0x33, 0x66, 0x77, 0x77, 0x44, 0x22, 0x22, 0x44, 0x7 },
};

void GalagaUpdateSound(LPBYTE rgbBuffer, LONG cbBuffer, const BYTE *RAM,
	BOOL &fSoundChanged)
{
	PacmanUpdateSound(rgbBuffer, cbBuffer, RAM, fSoundChanged, _rgbEnvelopes[0]);
}

