/*---------------------------------------------------------------------------
	
	File:
		wndmgr.cpp
	
	Abstract:
		This file implements a very small and fast set of window management
		services that address some of the most common techniques used in 
		Win32 programming (including some of those found in the MFC).
		
		The library is layered like so...

		Lowest Level:
		
		Window creation, associating instance data, registering window class,
		reference counting event handler, default handling of messages (for
		all different varieties: MDI frame, MDI child, dialog, normal), 

		Generic message loop that can handle translation of accelerator keys
		including accelerators in dialog boxes (w/custom dialog modal loop).

		Command routing for update of UI elements.

		Utility Functions - Setup cool icons (large and small) for Windows 4.0.

		Highest Level:

		- don't have to register a window class
		- instance data is available
		- generic message loop with accelerators and idle processing
		- generic dialog loop (w/accel & idle)
		- message cracking
		- "automatic" default message handling (DefWindowProc,
		  DefMDIChildProc, DefFrameProc, DefDialogProc)
		- command routing/activation
		- property sheets
		- utilities (centering, persistance, ...)

	Revisions:
		Jan-21-96 EAM Created
		
	Copyright (C) 1997 Edward Massey

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "wndmgr.h"

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

#define THREAD

THREAD IWindow *_pwndCreate = NULL;
THREAD MSG _msgCurrent;

static const int _iWndData = 0; // GWL_USERDATA;

struct WINDOW_DATA
{
	IWindow *pwnd;
};

struct WINDOW_CLASS_DATA
{
};

static LPCTSTR _pszWindowClassName;

/*---------------------------------------------------------------------------
Generic Utilities
---------------------------------------------------------------------------*/

inline BOOL WndSetLong(HWND hwnd, int nIndex, LONG lVal)
{
	// NOTE:  See SDK docs for SetWindowLong limitations

	SetLastError(0);
	
	if (SetWindowLong(hwnd, nIndex, lVal) != 0L)
		return TRUE;

	DWORD dwErr = GetLastError();
	
	if (dwErr == 0L)
		return TRUE;

//	_RPT1(_CRT_ERROR, __FILE__ ": unable to set window long (err=%i)\n", dwErr);

	return FALSE;
}

inline BOOL WndGetLong(HWND hwnd, int nIndex, LONG &lVal)
{
	lVal = GetWindowLong(hwnd, nIndex);
	
	if (lVal != 0L)
		return TRUE;

	DWORD dwErr = GetLastError();
	
	if (dwErr == 0L)
		return TRUE;

//	_RPT1(_CRT_ERROR, __FILE__ ": unable to get window long (err=%i)\n", dwErr);

	return FALSE;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

inline BOOL GetWindowData(HWND hwnd, WINDOW_DATA &wd)
{
	return WndGetLong(hwnd, _iWndData, (LONG&)wd.pwnd);
}

inline BOOL SetWindowData(HWND hwnd, const WINDOW_DATA &wd)
{
	return WndSetLong(hwnd, _iWndData, (LONG)wd.pwnd);
}

static WNDPROC GetWndProc(HWND hwnd)
{
//	WNDPROC wndproc; WndGetLong(hwnd, GWL_WNDPROC, (LONG&)wndproc); return wndproc;
	return (WNDPROC)GetWindowLong(hwnd, GWL_WNDPROC);
}

static void SetWndProc(HWND hwnd, WNDPROC wndproc)
{
//	VERIFY(WndSetLong(hwnd, GWL_WNDPROC, (LONG)wndproc));
	WndSetLong(hwnd, GWL_WNDPROC, (LONG)wndproc);
}

/*---------------------------------------------------------------------------

	Register the single window class if not already registered.

	Register our own dialog class so we get first message (before controls
	are created so we can grab up the _pwndCreate pointer, otherwise, if a
	CWindow window is on the dialog, it gets the dialog object pointer)

	NOTE:  We are overriding the global dialog class here for our.  For
	instance, if a normal dialog (#32767) is created from a dialog resource,
	the resulting window will be of our dialog class, not the global dialog
	class.

---------------------------------------------------------------------------*/

LRESULT CALLBACK WndProcInit(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK WndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void OnInitMenuPopup(HMENU hmenu, IWindow *pwnd);

LPCTSTR WndRegisterWindowClass(LPCTSTR pszName, UINT style)
{
	WNDCLASS wndcls;

	wndcls.style         = style;
	wndcls.lpfnWndProc   = WndProcInit;
	wndcls.cbClsExtra    = sizeof(WINDOW_CLASS_DATA);
	wndcls.cbWndExtra    = sizeof(WINDOW_DATA);
	wndcls.hInstance     = GetModuleHandle(NULL);
	wndcls.hIcon         = NULL;
	wndcls.hCursor       = LoadCursor(NULL, IDC_ARROW);
	wndcls.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
	wndcls.lpszMenuName  = NULL;
	wndcls.lpszClassName = pszName;

	// NOTE: not documented...
	return MAKEINTRESOURCE(RegisterClass(&wndcls));
}

HWND WndCreateWindow(DWORD dwExStyle, LPCTSTR pszClassName,
	LPCTSTR pszWindowName, DWORD dwStyle, int x, int y, int cx, int cy,
	HWND hwndOwner, HMENU hmenu, IWindow *pwnd) /*HICON hicon, HBRUSH hbrBackground, HCURSOR hcsr,*/
{
	//
	//  Register window class if not done already
	//

	if (pszClassName == NULL)
	{
		if (_pszWindowClassName == NULL)
		{
			static const TCHAR szClsName[] = TEXT("EmWindow");

			_pszWindowClassName = WndRegisterWindowClass(szClsName,
				CS_DBLCLKS);

			if (_pszWindowClassName == NULL)
				return NULL;
		}

		pszClassName = _pszWindowClassName;
	}

	//
	//
	//

	IWindow *pwndCreatePrev = _pwndCreate;
	
	_pwndCreate = pwnd;

	HWND hwnd = CreateWindowEx(dwExStyle, pszClassName,
		pszWindowName, dwStyle, x, y, cx, cy, hwndOwner, hmenu,
		GetModuleHandle(NULL), NULL);

	_pwndCreate = pwndCreatePrev;

	return hwnd;
}
	     
/*---------------------------------------------------------------------------

	Handle the first message for a window.

	1.  Set the window data to point to the handler and add a reference to it
	2.  Notify each handler in the chain of the window's handle
	3.  Set the window procedure to "other" window procedure
	4.  Handle the message

---------------------------------------------------------------------------*/

LRESULT CALLBACK WndProcInit(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	//
	//  It is possible that the thread's event handler pointer (_pwndCreate)
	//  has not been set by this point.  This happens for windows created on
	//  dialog boxes since we're not in control of the window creation there.
	//
	//  CONSIDER:  Create our own CreateDialog function and parse the dialog
	//  resource ourselves, setting _pwndCreate for each control base on
	//  a dialog's yet-to-be-designed "control map".
	//
	
//	if (_pwndCreate == NULL)
//		return OnUnhandledMessage(hwnd, uMsg, wParam, lParam);

	//
	//  Store the [reference counted] event handler pointer in the window's
	//  32-bit user data storage area.
	//

#if 0
	TRACE("setting window %p event handler %p\n", hwnd, _pwndCreate);
#endif

	WINDOW_DATA wd;
	
	wd.pwnd = _pwndCreate;
	
	SetWindowData(hwnd, wd);

	_pwndCreate->AddRef();

	//
	//  Store the HWND member away
	//

//	SendNotifyCreate(_pwndCreate, hwnd);

	_pwndCreate = NULL; // HACK???

	//
	//  Set the window procedure to our "normal" (non-initial) procedure
	//
	
//	SetWndProc(hwnd, WndProc);
//	VERIFY(WndSetLong(hwnd, GWL_WNDPROC, (LONG)WndProc));
	WndSetLong(hwnd, GWL_WNDPROC, (LONG)WndProc);

	//
	//  Finally, handle this initial message.
	//

	return WndProc(hwnd, uMsg, wParam, lParam);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

LRESULT CALLBACK WndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	WINDOW_DATA wd;
	
	if (GetWindowData(hwnd, wd))
	{
		//
		//  Save current message for nesting
		//
		
		MSG msgPrev = _msgCurrent;

		//
		//
		//

		if (uMsg == WM_NCCREATE)
		{
			LPCREATESTRUCT lpcs = (LPCREATESTRUCT)lParam;
		}

		//
		//  Set current message with incoming message info
		//
		
		_msgCurrent.hwnd    = hwnd;
		_msgCurrent.message = uMsg;
		_msgCurrent.wParam  = wParam;
		_msgCurrent.lParam  = lParam;
		_msgCurrent.time    = GetMessageTime();
		_msgCurrent.pt.x    = LOWORD(GetMessagePos());
		_msgCurrent.pt.y    = HIWORD(GetMessagePos());

		LRESULT lResult = wd.pwnd->OnMessage(_msgCurrent);

		if (uMsg == WM_INITMENUPOPUP)
		{
			BOOL fSystemMenu = BOOL(HIWORD(lParam));
			
			if (!fSystemMenu)
			{
				OnInitMenuPopup(HMENU(wParam), wd.pwnd);
			}
		}

		//
		//  Release reference to event handler on last message
		//
		
		if (uMsg == WM_NCDESTROY)
			wd.pwnd->Release();

		//
		//  Restore previous current message for nesting
		//

		_msgCurrent = msgPrev;

		return lResult;
	}

	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK WndProcMdiFrame(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK WndProcMdiChild(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK WndProcDialog(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

LRESULT WndFwdCurMsg1()
{
	return DefWindowProc(_msgCurrent.hwnd, _msgCurrent.message,
		_msgCurrent.wParam, _msgCurrent.lParam);
}

LRESULT WndFwdCurMsg2()
{
	return DefMDIChildProc(_msgCurrent.hwnd, _msgCurrent.message,
		_msgCurrent.wParam, _msgCurrent.lParam);
}

/*---------------------------------------------------------------------------

	XlatKeyboardMsg - Translate keyboard message
	--------------------------------------------

	This routine provides the mechanism that allows us to have a single
	generic messsage loop that correctly handles modeless dialogs.

	Handle keyboard message translation.  This includes accelerator
	keys (incl. MDI) and dialog control navigation (mnemonics and
	tab).  Give the target window and ancestors a crack at
	translating the message.  Finally, give the application the
	oppurtunity to translate.

	GetEventHandler...

	Find the object associated with the given window handle.  A pointer to
	the object is stored in the window extra bytes at offset GWL_USERDATA for
	our window class and in a window property named "CWindow*" for other
	window classes (such as dialog boxes, MDI client, or other subclasses).

	NULL is acceptable input (returns NULL)

---------------------------------------------------------------------------*/

inline IWindow *GetEventHandler(HWND hwnd)
{
	WINDOW_DATA wd;

	if (GetClassLong(hwnd, GCW_ATOM) == LOWORD(_pszWindowClassName) &&
		GetWindowData(hwnd, wd))
	{
		return wd.pwnd;
	}

	return NULL;
}

inline BOOL XlatKeyboardMsg(MSG &msg)
{
	if (msg.message < WM_KEYFIRST || msg.message > WM_KEYLAST)
		return FALSE;

	//
	//  Walk window hierarchy chain
	//
	
	for (HWND hwnd = msg.hwnd; hwnd != NULL; hwnd = GetParent(hwnd))
	{
		//
		//  Walk handler chain for each window
		//

		IWindow *pwnd = GetEventHandler(hwnd);

		if (pwnd != NULL)
		{
			if (pwnd->PreTranslateMsg(msg))
				return TRUE;
		}
	}

	return FALSE;
}

/*---------------------------------------------------------------------------
	
	WndMsgLoop - generic message loop implementation
	------------------------------------------------
	
	Reference:  Knowledge base article Q74042, MFC Source CWinThread::Run

	Concearns:
	 - Good user input response.  Process messages before idle callbacks.
	 - Application idles when there is no more background processing.

	Algorithm:
	 - While there are no messages, service any idle callbacks that are
	   marked.  If there are no more idle service requests, wait for a message
	 - Exit message loop if the quit message is received.
	 - Translate keyboard messages.  If unhandled, do system message dispatch.
	 - Reset idle callback state so they are called again after we do idle.

UINT MessageLoop()
{
	MSG msg;

	for (;;)
	{
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if (msg.message == WM_QUIT)
				break;
			
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		else
		{
			if (!DoIdle())
				WaitMessage();
		}
	}

	return (msg.wParam);
}

---------------------------------------------------------------------------*/

int WndMsgLoop()
{
	MSG msg;
	POINT ptMseLast = { -1, -1 };

	for (;;)
	{
		//
		//  While there are no messages, do idle processing.  If no more idle
		//  processing is required, wait for a message.
		//
		
		while (!PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
//			if (!DoIdleProcessing(_cIdle++))
			{
				GetMessage(&msg, NULL, 0, 0);
				break;
			}
		}

		//
		//
		//
		
		if (msg.message == WM_QUIT)
			break;

		//
		//
		//
		
		if (!XlatKeyboardMsg(msg))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		//
		//
		//
		
		if (msg.message == WM_PAINT)
			continue;

/*
		if (msg.message >= WM_MOUSEFIRST && msg.message <= WM_MOUSELAST)
		{
			if (ptMseLast.x == msg.pt.x && ptMseLast.y == msg.pt.y)
				continue;

			ptMseLast.x = msg.pt.x;
			ptMseLast.y = msg.pt.y;
		}
*/
		
//		ResetIdleState();
	}

	return (int)msg.wParam;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

BOOL CALLBACK AbortProc(HDC hdc, int nError)
{
	MSG msg;
	
	if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
	{
		if (!XlatKeyboardMsg(msg))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	
	return TRUE;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

void OnInitMenuPopup(HMENU hmenu, IWindow *pwnd)
{
	const int cItem = GetMenuItemCount(hmenu);
	MENUITEMINFO mii;
	mii.cbSize = sizeof(mii);
	int iItem;

	for (iItem = 0; iItem < cItem; iItem++)
	{
		//
		//  We'll need the item identifier, submenu handle, and state
		//
		
		mii.fMask = MIIM_ID | MIIM_SUBMENU | MIIM_STATE;

		if (!GetMenuItemInfo(hmenu, iItem, TRUE, &mii))
			break;

		//
		//  Skip past separators and popups
		//
		
		if (mii.hSubMenu != NULL || mii.wID == 0)
			continue;

		//
		//  Enable the item, clear any check mark
		//

		mii.fState &= ~MFS_DISABLED;
		mii.fState &= ~MFS_CHECKED;

		//
		//  Route command.  Disable if not handled.
		//

		pwnd->UpdateUI(mii.wID, mii.fState);

		//
		//
		//

		mii.fMask = MIIM_STATE;

		SetMenuItemInfo(hmenu, iItem, TRUE, &mii);
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

void WndSetIcon(HWND hwnd, HICON hicon)
{
	DefWindowProc(hwnd, WM_SETICON, TRUE, (LPARAM)hicon);
}

BOOL WndCenter(HWND hwndChild, HWND hwndParent)
{
	//
	//
	//

	RECT rcChild;
	SIZE sizChild;

	if (!GetWindowRect(hwndChild, &rcChild))
		return FALSE;

	sizChild.cx = rcChild.right - rcChild.left;
	sizChild.cy = rcChild.bottom - rcChild.top;

	//
	//
	//

	RECT rcParent;
	SIZE sizParent;

	if (!GetWindowRect(hwndParent, &rcParent))
		return FALSE;

	sizParent.cx = rcParent.right - rcParent.left;
	sizParent.cy = rcParent.bottom - rcParent.top;

	//
	//
	//
	
	RECT rcWorkArea;
	
	if (!SystemParametersInfo(SPI_GETWORKAREA, sizeof(rcWorkArea), &rcWorkArea, 0))
	{
		rcWorkArea.left   = 0;
		rcWorkArea.top    = 0;
		rcWorkArea.right  = GetSystemMetrics(SM_CXSCREEN);
		rcWorkArea.bottom = GetSystemMetrics(SM_CYSCREEN);
	}

	//
	//
	//

	POINT ptOrg;

	ptOrg.x = rcParent.left + ((sizParent.cx - sizChild.cx) /2);

	if (ptOrg.x < rcWorkArea.left)
		ptOrg.x = rcWorkArea.left;
	else if ((ptOrg.x + sizChild.cx) > rcWorkArea.right)
		ptOrg.x = rcWorkArea.right - sizChild.cx;

	ptOrg.y = rcParent.top + ((sizParent.cy - sizChild.cy) /2);
	
	if (ptOrg.y < rcWorkArea.top)
		ptOrg.y = rcWorkArea.top;
	else if ((ptOrg.y + sizChild.cy) > rcWorkArea.bottom)
		ptOrg.y = rcWorkArea.bottom - sizChild.cy;

	return SetWindowPos(hwndChild, NULL, ptOrg.x, ptOrg.y, 0, 0,
		SWP_NOSIZE | SWP_NOZORDER);
}

